% THO2_TEST An independent check for ThO2.dat (finds optimal model
% parameters using MATLAB Optimization Toolbox and MATLAB Statistics and
% Machine Learning Toolbox).
%
% See also THO2_CPDATA

% (C) 2017 Alexey Voskov
% Laboratory of Chemical Thermodynamics, Department of Chemistry,
% Lomonosov Moscow State University, Moscow, Russia, 119991
%
% This file is the part of the CpFit program
%
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 2 of the License, or
% (at your option) any later version.
%
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
% GNU General Public License for more details.
% You should have received a copy of the GNU General Public License
% along with this program.  If not, see http://www.gnu.org/licenses/);
function ThO2_test
% Load data
[CpDat, HDat] = ThO2_CpData();
% Objective function
    function d = minfunc(param)
        nterms = numel(param) / 2;
        a = param(1:nterms);
        t = param((nterms+1):(2*nterms));
        Ccalc = Cpcalc(a, t, CpDat(:,1));
        Hcalc = HCalc(a, t, HDat(:,2)) - HCalc(a, t, HDat(:,1));
        d = [(Ccalc - CpDat(:,2))./CpDat(:,2); ...
            (Hcalc - HDat(:,3))./HDat(:,3)];
    end
% Initial approximation and settings
numOfTerms = 5;
param0 = [0.1*ones(1,numOfTerms), ones(1,numOfTerms)];
opt = optimset('Display', 'iter', 'MaxFunEvals', 100000, 'TolFun', 1e-10, 'TolX', 1e-12);
% Run nonlinear regression and find parameters confidence intervals
[param,~,res,~,~,~,J] = lsqnonlin(@minfunc, param0, zeros(size(param0)), [], opt);
param_ci = nlparci(param,res,'jacobian',J);
dparam = (param_ci(:,2) - param_ci(:,1))' / 2;
% Show the result
alpha = param(1:numOfTerms); dalpha = dparam(1:numOfTerms);
theta = param((numOfTerms+1):end); dtheta = dparam((numOfTerms+1):end);
fprintf('%15s %15s %15s %15s\n', 'ALPHA', 'DALPHA', 'THETA', 'DTHETA');
fprintf('%15g %15g %15g %15g\n', [alpha;dalpha;theta;dtheta]);
fprintf('\n');
end

function f = Cpcalc(alfa,teta,T)
f = zeros(size(T));
for i = 1:numel(alfa)
    x = teta(i)./T; ex = exp(x);
    ii = x < 500;
    f(ii) = f(ii) + 3*alfa(i)*ex(ii).*(x(ii)./(ex(ii)-1)).^2;
    f(~ii) = f(~ii) + 3*alfa(i)*(x(~ii).^2./ex(~ii));        
end
f = f * 8.3144598;
end

function f = HCalc(alfa,teta,T)
f = zeros(size(T));
for i = 1:numel(alfa)
    x = teta(i)./T;
    f = f + alfa(i) * 3*x./(exp(x) - 1);
end
f = f .* 8.3144598 .* T;
end