/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */
#include <stdio.h>
#include <stdlib.h>
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/bookctrl.h"
#include "wx/textfile.h"
#include "wx/tokenzr.h"
#include "wx/aboutdlg.h"
#include "wx/filefn.h"
#include "datafit.h"
#include "plotcanvas.h"
#include "customgrids.h"
#include "cp.h"
#include "graph.h"
#include "misc.h"

#include "cp_ids.h"



////////////////////////////////////////////////
wxBEGIN_EVENT_TABLE(MetadataWindow, wxPanel)
	EVT_TEXT(ID_TXT_FILEINFO0, MetadataWindow::OnTextEvent)
	EVT_TEXT(ID_TXT_FILEINFO1, MetadataWindow::OnTextEvent)
	EVT_TEXT(ID_TXT_FILEINFO2, MetadataWindow::OnTextEvent)
wxEND_EVENT_TABLE()

MetadataWindow::MetadataWindow(wxWindow *parent, std::function <void()> cbfFileChanged, const wxString &label,
	const wxPoint &pos, const wxSize &size)
	: wxPanel(parent, wxID_ANY, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;

	box = new wxStaticBox(this, wxID_ANY, label, pos, size);
	lblFileInfo[0] = new wxStaticText(box, wxID_ANY, "File description 1"); // Field 1 (lbl + txt)
	txtFileInfo[0] = new wxTextCtrl(box, ID_TXT_FILEINFO0, "Data set description");

	lblFileInfo[1] = new wxStaticText(box, wxID_ANY, "File description 2"); // Field 2 (lbl + txt)
	txtFileInfo[1] = new wxTextCtrl(box, ID_TXT_FILEINFO1, "T, K");

	lblFileInfo[2] = new wxStaticText(box, wxID_ANY, "File description 3"); // Field 3 (lbl + txt)
	txtFileInfo[2] = new wxTextCtrl(box, ID_TXT_FILEINFO2, "Cp, J/mol/K");

	wxStaticBoxSizer *boxSizer = new wxStaticBoxSizer(box, wxVERTICAL);
	for (int i = 0; i < 3; i++) {
		wxBoxSizer *lineSizer = new wxBoxSizer(wxHORIZONTAL);
		lineSizer->Add(lblFileInfo[i], 0, wxALL | wxEXPAND | wxALIGN_LEFT | wxALIGN_TOP, 3);
		lineSizer->SetItemMinSize(lblFileInfo[i], 190, 25);
		lineSizer->Add(txtFileInfo[i], 1, wxALL | wxEXPAND | wxALIGN_LEFT | wxALIGN_TOP, 3);
		boxSizer->Add(lineSizer, 1, wxALL | wxEXPAND | wxALIGN_LEFT | wxALIGN_TOP, 3);
	}

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	sizer->Add(boxSizer, 1, wxALL | wxEXPAND, 0);

	this->SetSizer(sizer);
	this->Layout();
}

bool MetadataWindow::SetFileInfo(std::vector <wxString> &fileInfo)
{
	if (fileInfo.size() != 3) {
		return false;
	}
	for (int i = 0; i < 3; i++) {
		txtFileInfo[i]->ChangeValue(fileInfo[i]);
	}
	return true;
}

std::vector <wxString> MetadataWindow::GetFileInfo()
{
	std::vector <wxString> fileInfo;
	for (int i = 0; i < 3; i++) {
		fileInfo.push_back(txtFileInfo[i]->GetValue());
	}
	return fileInfo;
}

void MetadataWindow::OnTextEvent(wxCommandEvent& event)
{
	cbfFileChanged();
}

////////////////////////////////////////
wxBEGIN_EVENT_TABLE(MainWindow, wxFrame)
	EVT_MENU(ID_MNU_NEW, MainWindow::OnNew)
	EVT_MENU(ID_MNU_OPEN, MainWindow::OnOpenFile)
	EVT_MENU(ID_MNU_SAVEAS, MainWindow::OnSaveAsFile)
	EVT_MENU(ID_MNU_LOADPARAM, MainWindow::OnLoadParameters)
	EVT_MENU(ID_MNU_EXIT, MainWindow::OnExit)
	EVT_CLOSE(MainWindow::OnClose)
	EVT_MENU(ID_MNU_ABOUT, MainWindow::OnAbout)
	EVT_MENU(ID_MNU_CUSTOMTERM, MainWindow::OnAddCustomTerm)

	EVT_MENU(ID_MNU_CALC, MainWindow::OnEvalPar)
	EVT_MENU(ID_MNU_AUTOCALC, MainWindow::OnAutoEvalPar)
	EVT_MENU(ID_MNU_CALC_ADDTERMS, MainWindow::OnAddTermsEvalPar)
	EVT_MENU(ID_MNU_CALC_EXTERMSONLY, MainWindow::OnEvalExTermsOnly)
	EVT_MENU(ID_MNU_CALC_NOOPT, MainWindow::OnNoOptEvalPar)
	
	EVT_MENU(ID_MNU_DATA_SORT, MainWindow::OnDataSort)
	EVT_MENU(ID_MNU_DATA_EXCLUDEW0, MainWindow::OnExcludeW0Points)
wxEND_EVENT_TABLE()


/**
 * \brief  Creates all elements in the main window form
 */
MainWindow::MainWindow() : wxFrame(NULL, wxID_ANY, wxString("CpFit"))
{
	strInitApprox = wxString::FromUTF8("Initial approximation");
	strInputData = wxString::FromUTF8("Input data");
	strNumOfTerms = wxString::FromUTF8("Number of terms");
	strResetParams = wxString::FromUTF8("Reset parameters");

	strNotNumeric = wxString::FromUTF8("Not numeric parameter");

	wxGridCellBoolEditor::UseStringValues("1", "0");

	wxPanel *panel = new wxPanel(this, wxID_ANY);
	isFileChanged = false;
	SetIcon(wxICON(sample));
	//---------- Some callbacks ----------
	auto cbfFileChanged = [this] () { this->SetIsFileChanged(true); }; // Callback for DataFileEditor
	auto cbfEvalParam = [this] () { wxCommandEvent evt; this->OnEvalPar(evt); }; // Callback for parameters evaluation
	//---------- Static box with file metadata ----------
	boxInfo = new MetadataWindow(panel, cbfFileChanged, "File metadata", wxPoint(10, 10), wxSize(740, 150));
	//---------- Static box with initial parameters ----------
	box = new ParametersWindow(panel, cbfEvalParam);
	//---------- Static box with initial data ----------
	ctlData = new DataFileEditor(panel, cbfFileChanged, wxPoint(20, 30), wxSize(380, 230));
	//---------- Sizers ----------	
	wxBoxSizer *horzSizer = new wxBoxSizer(wxHORIZONTAL);
	horzSizer->Add(box, 4, wxALL | wxEXPAND, 5);
	horzSizer->Add(ctlData, 6, wxALL | wxEXPAND, 5);

	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(boxInfo, 0, wxALL | wxEXPAND, 5);
	mainSizer->Add(horzSizer, 1, wxALL | wxEXPAND, 0);
	panel->SetSizer(mainSizer);
	// Window menu
	CreateMenu();
	// Resize and refresh title
	SetSize(wxSize(900, 600));
	SetMinSize(wxSize(900, 600));
	RefreshTitle();
}

void MainWindow::CreateMenu()
{
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_MNU_NEW, "&New\tALT-N", "New data set");
	menuFile->AppendSeparator();
	menuFile->Append(ID_MNU_OPEN, "&Open\tALT-O", "Open data file");
	menuFile->Append(ID_MNU_SAVEAS, "Save as", "Save current data as...");
	menuFile->AppendSeparator();
	menuFile->Append(ID_MNU_LOADPARAM, "Load parameters", "Load alpha and theta parameters from .csv file");
	menuFile->Append(ID_MNU_CUSTOMTERM, "Load custom extra term", "Load custom extra Cp term (lua script)");
	menuFile->AppendSeparator();
	menuFile->Append(ID_MNU_EXIT, "E&xit\tALT-X", "Quit this program");
	wxMenu *menuEval = new wxMenu();
	menuEval->Append(ID_MNU_CALC, "Parameters values", "Parameters values");
	menuEval->AppendSeparator();
	menuEval->Append(ID_MNU_AUTOCALC, "Parameters set (don't use init.approx.)", "Parameters number and values");
	menuEval->Append(ID_MNU_CALC_ADDTERMS, "Parameters set (use init.approx.)", "Parameters numbers and values (use init.approx.)");
	menuEval->Append(ID_MNU_CALC_EXTERMSONLY, "Parameters set: extra terms only", "Parameters numbers and values: extra terms only");
	menuEval->AppendSeparator();
	menuEval->Append(ID_MNU_CALC_NOOPT, "Evaluate without optimization", "Evaluate without optimization");
	wxMenu *menuData = new wxMenu();
	menuData->Append(ID_MNU_DATA_SORT, "Sort data", "Sort data");
	menuData->Append(ID_MNU_DATA_EXCLUDEW0, "Exclude points with w=0", "Exclude points with w=0");
	wxMenu *menuHelp = new wxMenu();
	menuHelp->Append(ID_MNU_ABOUT, "About", "About the program");
	wxMenuBar *menuBar = new wxMenuBar();
	menuBar->Append(menuFile, "&File");
	menuBar->Append(menuEval, "&Evaluate");
	menuBar->Append(menuData, "&Data");
	menuBar->Append(menuHelp, "&Help");
	SetMenuBar(menuBar);
}

/**
 * \brief  Addition of custom terms (lua scripts)
 */
void MainWindow::OnAddCustomTerm(wxCommandEvent &event)
{
	// Create open file dialog (for .lua files)
	wxFileDialog openFileDialog(this, wxT("Open custom term script file"), "", "",
		"lua script files (*.lua)|*.lua", wxFD_OPEN|wxFD_FILE_MUST_EXIST);
	// Run open file dialog and load the custom term from it
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		// Term loading process
		ExtraTermContainer::InitTermsTypes();
		wxString fileName = openFileDialog.GetPath();
		LuaTerm *term = new LuaTerm(std::string(fileName.c_str()));
		wxString title = wxT("Custom term loader");
		if (term->isValid) {
			// Try to call the function
			double Cp = term->CpFuncEval((term->Tmin + term->Tmax) / 2);
			if (Cp != Cp) {
				wxMessageBox(term->GetErrorMessage(), title, wxICON_ERROR);
				return; // Cancel addition of term
			}
			// Add term
			if (ExtraTermContainer::AddExtraTermType(term)) {
				wxString msg = wxT("Term ") + term->GetType() + wxT(" successfully loaded");
				ctlData->RefreshExtraTermsTypes();
				wxMessageBox(msg, wxT("Custom term loader"), wxICON_INFORMATION);
			} else {
				wxMessageBox(wxT("Error during loading ") + term->GetErrorMessage(), title, wxICON_ERROR);
			}
		} else {
			wxMessageBox(term->GetErrorMessage(), title, wxICON_ERROR);
		}
	}
}


/**
 * \brief  Creation of new data set (reset everyting)
 */
void MainWindow::OnNew(wxCommandEvent &event)
{
	box->ResetParameters();
	ctlData->Reset();
	dataFile.Reset();
	std::vector <wxString> fileInfo = {"Data set description", "T, K", "Cp, J/mol/K"};
	boxInfo->SetFileInfo(fileInfo);
	SetIsFileChanged(false);
}

/**
 * \brief Open the file and load data set from it
 */
void MainWindow::OnOpenFile(wxCommandEvent &event)
{    
	// Check if file has been changed (to prevent data loss)
	if (isFileChanged) {
		int ans = wxMessageBox("Current file is not saved. All changes will be lost. Continue?", "CpFit",
			wxYES_NO | wxNO_DEFAULT | wxICON_EXCLAMATION);
		if (ans == wxNO) {
			return; // The user cancelled opening the file: stop it!
		}
	}
	// Create open file dialog (for .dat files)
	wxFileDialog openFileDialog(this, wxT("Open dat file"), "", "",
                       "dat files (*.dat;*.txt)|*.dat;*.txt", wxFD_OPEN|wxFD_FILE_MUST_EXIST);
	// Run open file dialog and open the file
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		if (dataFile.LoadFile(fileName)) {
			RefreshDataGrid();
			SetIsFileChanged(false);
		} else {
			wxMessageBox("Cannot load the file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
}

/**
 * \brief Save current data set to file
 */
void MainWindow::OnSaveAsFile(wxCommandEvent &event)
{
	// Synchronize window with data
	SaveToDataFile();
	// Get file name
	wxFileDialog openFileDialog(this, wxT("Save dat file"), "", "",
                       "dat files (*.dat;*.txt)|*.dat;*.txt", wxFD_SAVE);
	// Save file on disk
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		// Checks if the file exists (if exists -- ask if we need 
		// to overwrite it)
		wxTextFile file;
		bool writeToFile = true;
		if (wxFileExists(fileName)) {
			wxString msg = wxString::Format("File %s already exists. Overwrite it?", fileName);
			if (wxMessageBox(msg, "CpFit", wxICON_EXCLAMATION | wxYES_NO | wxNO_DEFAULT) == wxNO) {
				writeToFile = false;
			}
		}
		// Save to the file (if allowed by the user)
		if (writeToFile) {
			if (dataFile.SaveFile(fileName)) {
				SetIsFileChanged(false);
			} else {
				wxMessageBox("Cannot save the file", "CpFit", wxICON_ERROR | wxOK);
			}
		}
	}
}

/**
 * \brief Loads model parameters from CSV file
 *
 * The first row contains alpha and theta strings,
 * the rest of rows are parameters. Dot is decimal point,
 * semicolon is separator.
 */
void MainWindow::OnLoadParameters(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Load parameters values"), "", "",
                       "csv files (*.csv)|*.csv", wxFD_OPEN);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		// Open file
		wxString fileName = openFileDialog.GetPath();
		wxTextFile file;
		if (!file.Open(fileName)) {
			wxString msg = wxString::Format("Cannot open file %s", fileName);
			wxMessageBox(msg, "CpFit", wxICON_ERROR);
			return;
		}
		// Read file header
		// a) Check header presence
		if (file.Eof()) {
			file.Close();
			wxMessageBox("File is empty!", "CpFit", wxICON_ERROR);
			return;
		}
		// b) find alpha and theta in the header (its indices)
		int alpha_ind = -1, theta_ind = -1, ind = 0;
		wxString str = file.GetFirstLine();
		wxStringTokenizer tokenizer(str, ";");
		while (tokenizer.HasMoreTokens()) {
			wxString token = tokenizer.GetNextToken();
			if (token == "alpha") {
				alpha_ind = ind;
			} else if (token == "theta") {
				theta_ind = ind;
			}
			ind++;
		}
		//int numOfCols = ind;
		// c) check if alpha and theta fields exist
		if (alpha_ind == -1) {
			file.Close();
			wxMessageBox("Cannot find alpha column", "CpFit", wxICON_ERROR);
			return;
		}
		if (theta_ind == -1) {
			file.Close();
			wxMessageBox("Cannot find theta column", "CpFit", wxICON_ERROR);
			return;
		}
		// Load parameters values
		std::vector <double> alpha_new, theta_new;
		for (str = file.GetNextLine(); !file.Eof(); str = file.GetNextLine()) {
			// Find and load alpha and theta values from the row
			tokenizer = wxStringTokenizer(str, ";");
			int rowInd = 0;
			bool alphaLoaded = false, thetaLoaded = false;
			while (tokenizer.HasMoreTokens()) {
				// Get new token
				wxString token = tokenizer.GetNextToken();
				// Convert token to double
				double val;
				if (!token.ToDouble(&val)) {
					file.Close();
					wxString msg = wxString::Format("%s is not a number", token);
					wxMessageBox(msg, "CpFit", wxICON_ERROR);
					return;
				}
				// Load values to vectors
				if (rowInd == alpha_ind) {
					alpha_new.push_back(val);
					alphaLoaded = true;
				}
				if (rowInd == theta_ind) {
					theta_new.push_back(val);
					thetaLoaded = true;
				}
				rowInd++;
			}
			// Check if alpha and theta values have been loaded
			if (!alphaLoaded) {
				wxMessageBox("Cannot find alpha value!", "CpFit", wxICON_ERROR);
			}
			if (!thetaLoaded) {
				wxMessageBox("Cannot find theta value!", "CpFit", wxICON_ERROR);
			}
		}
		// Save it into CpFit forms
		WriteAlpha(alpha_new);
		WriteTheta(theta_new);
	}	
}

/**
 * \brief Show "About the program" dialog
 */
void MainWindow::OnAbout(wxCommandEvent &event)
{
	wxString GPLText("This program is free software: you can redistribute it and/or modify\n"
	"it under the terms of the GNU General Public License as published by\n"
	"the Free Software Foundation, either version 2 of the License, or\n"
	"(at your option) any later version.\n\n"
	"This program is distributed in the hope that it will be useful,\n"
	"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
	"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
	"GNU General Public License for more details.\n\n"
	"You should have received a copy of the GNU General Public License\n"
	"along with this program.  If not, see <http://www.gnu.org/licenses/>.");

	wxAboutDialogInfo info;
	info.SetName(("CpFit"));
	info.SetVersion(_("0.9"));
	info.SetDescription(_(
		"This program describes heat capacities experimental data\n"
		"using Einstein-Plank Cp(T) empirical functions\n"
		"and the nonlinear least squares method\n"));
	info.SetCopyright(wxT("(C) 2016-2020 Laboratory of Chemical Thermodynamics\nLomonosow Moscow State University"));
	info.SetWebSite("http://td.chem.msu.ru");
	info.AddDeveloper("Alexey L. Voskov (alvoskov@gmail.com)");
	info.SetLicense(GPLText);
	wxAboutBox(info);
}



/**
 * \brief  Run parameters evaluation (by means of the least squares method)
 */
void MainWindow::OnEvalPar(wxCommandEvent &event)
{
	bool isOk;
	SaveToDataFile();
	double Natoms = box->GetFixedNatoms(&isOk);
	if (!isOk) {
		return;
	}

	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");
		TDBaseLineFuncParams param0 = ReadParameters();
		grwnd->AddData(dataFile);
		grwnd->AddParam0(param0);
		grwnd->AddFixedNatoms(Natoms);
		grwnd->CurveFit();
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Run parameters evaluation using autodefinition of the parameters number.
 * User-defined initial approximation will be ignored! It is based on the least
 * squares method.
 */
void MainWindow::OnAutoEvalPar(wxCommandEvent &event)
{
	bool isOk;
	SaveToDataFile();
	double Natoms = box->GetFixedNatoms(&isOk);
	if (!isOk) {
		return;
	}

	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");
		TDBaseLineFuncParams param0 = ReadParameters(); // Polynomial part is used from here
		grwnd->AddData(dataFile);
		grwnd->AddFixedNatoms(Natoms);
		grwnd->AddParam0(param0);
		grwnd->CurveFit(FIT_VALNTERMS_AUTO);
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Run parameters evaluation using user-defined initial approximation but
 * adds extra terms while they are statistically significant (can be used when
 * MainWindow::OnAutoEvalPar stops at small number of terms, i.e. in local minimum)
 */
void MainWindow::OnAddTermsEvalPar(wxCommandEvent &event)
{
	bool isOk;
	SaveToDataFile();
	double Natoms = box->GetFixedNatoms(&isOk);
	if (!isOk) {
		return;
	} 
	
	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");
		TDBaseLineFuncParams param0 = ReadParameters();
		grwnd->AddData(dataFile);
		grwnd->AddFixedNatoms(Natoms);
		grwnd->AddParam0(param0);
		grwnd->CurveFit(FIT_VALNTERMS_INITAPPROX);
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Run parameters evaluation only for extra terms (baseline parameters are not
 * changing). In this case "initial approximation" parameters are used unchanged.
 */
void MainWindow::OnEvalExTermsOnly(wxCommandEvent &event)
{
	bool isOk;
	SaveToDataFile();
	double Natoms = box->GetFixedNatoms(&isOk);
	if (!isOk) {
		return;
	}

	if (dataFile.GetNumOfPoints() > 0) {
		GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");
		TDBaseLineFuncParams param0 = ReadParameters();
		grwnd->AddData(dataFile);
		grwnd->AddFixedNatoms(Natoms);
		grwnd->AddParam0(param0); // Will be used as baseline WITHOUT CHANGES!
		grwnd->CurveFit(FIT_EXTERMS_ONLY);
		grwnd->Show( true );
	} else {
		wxMessageBox("No data are present", "CpFit", wxOK | wxICON_EXCLAMATION);
	}
}

/**
 * \brief Shows graphs but doesn't optimize the coefficients. Can work either with
 * or without experimental data
 */
void MainWindow::OnNoOptEvalPar(wxCommandEvent &event)
{
	GraphWindow *grwnd = new GraphWindow((MainWindow *) this, "CpFit results");
	TDBaseLineFuncParams param0 = ReadParameters();
	grwnd->AddData(dataFile);
	grwnd->AddParam0(param0);
	grwnd->CurveFit(FIT_NOOPT);
	grwnd->Show( true );
}

/**
 * \brief Sorts all experimental data by reference, by type and by T
 */
void MainWindow::OnDataSort(wxCommandEvent &event)
{
	// Refresh file in memory
	SaveToDataFile();
	// Sort data
	dataFile.SortData();
	// Load sorted data to GUI
	RefreshDataGrid();
	SetIsFileChanged(true);
}

/**
 * \brief Exclude points with w=0 (i.e. zero statistical weights) from
 * all data sets.
 */
void MainWindow::OnExcludeW0Points(wxCommandEvent &event)
{
	// Refresh file in memory
	SaveToDataFile();
	// Exclude points
	dataFile.ExcludeW0Points();
	// Load sorted data to GUI
	RefreshDataGrid();
	SetIsFileChanged(true);
}

/**
 * \brief Some operations before closing the window
 */
void MainWindow::OnExit(wxCommandEvent &event)
{
	Close(true);
}


/**
 * \brief Asks questions about not saved data before closing
 */
void MainWindow::OnClose(wxCloseEvent& event)
{
	if (event.CanVeto() && isFileChanged) {
		int ans = wxMessageBox("Current file is not saved. Exit without changes?", "CpFit",
			wxYES_NO | wxNO_DEFAULT | wxICON_EXCLAMATION);
		if (ans == wxNO) {
			event.Veto();
			return;
		}
	}
	event.Skip();
}

/**
 * \brief Load existing data to the grid
 */
void MainWindow::RefreshDataGrid()
{
	std::vector <wxString> strFileInfo = dataFile.GetFileInfo();
	boxInfo->SetFileInfo(strFileInfo);
	ctlData->LoadDataFile(dataFile);
	// If the file contains any parameters (alpha/theta, polynomial) --
	// then overwrite all parameters in the form (even by empty ones)
	if (dataFile.params_0.alpha.size() != 0 || dataFile.params_0.polyCoeff.size() != 0) {
		WriteAlpha(dataFile.params_0.alpha);
		WriteTheta(dataFile.params_0.theta);
		WritePolyCoeff(dataFile.params_0.polyCoeff);
		WritePolyDegree(dataFile.params_0.polyDegree);
	}
}

/**
 * \brief Saves data entered in the form to the data file.
 * \details Saves experimental data and extra terms, model parameters and
 * the user-defined metainformation.
 */
void MainWindow::SaveToDataFile()
{
	ctlData->SaveToDataFile(dataFile);
	std::vector <wxString> fileInfo = boxInfo->GetFileInfo();
	dataFile.SetFileInfo(fileInfo);
	dataFile.params_0 = ReadParameters();
}

/**
 * \brief Sets isFileChanged flag and then updates the window title
 * \param value New boolean value
 */
void MainWindow::SetIsFileChanged(bool value)
{
	isFileChanged = value;
	RefreshTitle();
}



/**
 * \brief  Refreshes window title using file name
 */
void MainWindow::RefreshTitle()
{
	wxString fileName = dataFile.GetFileName();
	if (fileName == "") {
		fileName = "<untitled>";
	}
	wxString title = wxString::Format("CpFit: %s", fileName);
	if (isFileChanged) {
		title += " (*)";
	}
	SetTitle(title);
}


TDBaseLineFuncParams MainWindow::ReadParameters()
{
	TDBaseLineFuncParams param;
	param.alpha = ReadAlpha();
	param.theta = ReadTheta();
	param.polyCoeff = ReadPolyCoeff();
	param.polyDegree = ReadPolyDegree();
	return param;
}



/**
 * \brief  Collect all alpha values from the grid into a vector and return them
 */
std::vector <double> MainWindow::ReadAlpha()
{
	return box->ReadAlpha();
}

/**
 * \brief  Collect all theta values from the grid into a vector and return them
 */
std::vector <double> MainWindow::ReadTheta()
{
	return box->ReadTheta();
}

std::vector <double> MainWindow::ReadPolyCoeff()
{
	return box->ReadPolyCoeff();
}

std::vector <double> MainWindow::ReadPolyDegree()
{
	return box->ReadPolyDegree();
}


/**
 * \brief  Write alpha values to the table
 *
 * The table will be resized to make sizes of alpha and table grid consistent.
 */
bool MainWindow::WriteAlpha(std::vector <double> alpha)
{
	return box->WriteAlpha(alpha);
}

/**
 * \brief  Write theta values to the table
 *
 * The table will be resized to make sizes of alpha and table grid consistent.
 */
bool MainWindow::WriteTheta(std::vector <double> theta)
{
	return box->WriteTheta(theta);
}

bool MainWindow::WritePolyCoeff(std::vector <double> polyCoeff)
{
	return box->WritePolyCoeff(polyCoeff);
}

bool MainWindow::WritePolyDegree(std::vector <double> polyDegree)
{
	return box->WritePolyDegree(polyDegree);
}




/**
 * \brief  Write extra terms parameters values to the data file (in memory)
 *
 * The function controls parameters types
 */
bool MainWindow::WriteExtraTermsParams(std::vector <ExtraTermContainer> &extraTerms_new)
{
	if (dataFile.extraTerms.size() != extraTerms_new.size()) {
		return false;
	}
	for (size_t i = 0; i < extraTerms_new.size(); i++) {
		if (!extraTerms_new[i].CompareTo(dataFile.extraTerms[i])) {
			return false;
		}
	}
	dataFile.extraTerms = extraTerms_new;
	ctlData->LoadDataFile(dataFile);
	return true;
}


/**
 * \brief  Adds data point to the data file
 */
void MainWindow::AddDataPoint(Point point)
{
	dataFile.data.push_back(point);
	RefreshDataGrid();
	SetIsFileChanged(true);
}

/**
 * \brief  Adds data points to the data file
 */
void MainWindow::AddDataPoints(std::vector <Point> points)
{
	for (size_t i = 0; i < points.size(); i++) {
		dataFile.data.push_back(points[i]);
	}
	RefreshDataGrid();
	SetIsFileChanged(true);
}



/**
 * \brief CpFit main class
 */
class MyApp: public wxApp
{
public:
	virtual bool OnInit();
};

wxIMPLEMENT_APP(MyApp);

/**
 * \brief CpFit entry point
 */
bool MyApp::OnInit()
{
	wxInitAllImageHandlers();
	MainWindow *frame = new MainWindow();
	frame->Show( true );
	return true;
}
