/**
 * \file
 * \brief Command-line version of CpFit (supports file masks and packet mode)
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <functional>
#include <math.h>
#include <wx/wx.h>
#include <wx/grid.h>
#include <wx/dir.h>
#include <wx/textfile.h>
#include <wx/tokenzr.h>
#include <wx/filefn.h>
#include "datafit.h"
#include "xydatafile.h"

void printFileParams(XYDataFile &file)
{
	size_t nterms = file.params_0.alpha.size();
	std::cout << "  ALPHA: ";
	for (size_t i = 0; i < nterms; i++) {
		std::cout << file.params_0.alpha[i] << " ";
	}
	std::cout << std::endl;
	std::cout << "  THETA: ";
	for (size_t i = 0; i < nterms; i++) {
		std::cout << file.params_0.theta[i] << " ";
	}
	std::cout << std::endl;
	for (size_t i = 0; i < file.extraTerms.size(); i++) {
		ExtraTerm *term = file.extraTerms[i].term;
		std::vector <double> lsqparam = term->GetLSQParam();
		std::vector <double> fixparam = term->GetFixParam();		
		std::cout << "EXTRA TERM " << i << " " << term->GetType() << std::endl;
		std::cout << "  LSQ PARAMS: ";
		for (size_t j = 0; j < lsqparam.size(); j++) {
			std::cout << lsqparam[j] << " ";
		}
		std::cout << std::endl;
		std::cout << "  FIX PARAMS: ";
		for (size_t j = 0; j < fixparam.size(); j++) {
			std::cout << fixparam[j] << " ";
		}
		std::cout << std::endl;
	}
	std::cout << std::endl;
}

/*
LSQResult fileDataFit(XYDataFile &file, std::vector <double> param0)
{
	LSQResult result = file.DataFit(param0);
	int nparams = param0.size();
	std::vector <double> param = result.GetParam();
	printf("RSS: %lf\n", result.sumsqr);
	for (int i = 0; i < nparams; i++) {
		printf("%f ", param[i]);
	}
	printf("\n");
	return result;
}*/


LSQResult autoDataFit(XYDataFile &file)
{
	LSQResult result = file.DataFit();
	TDBaseLineFuncParams param = result.GetParam();
	printf("RSS: %lf\n", result.sumsqr);	
	file.params_0 = param;
	file.extraTerms = result.extraTerms;
	printFileParams(file);
	return result;
}

void printHelp()
{
	std::cout << "CpFit command-line version\n";
	std::cout << "(C) 2016-2020 Alexey Voskov\n";
	std::cout << "Usage:\n";
	std::cout << "  cpfit_cmd mode inpfile\n";
	std::cout << "    inpfile -- input file name or mask\n";
	std::cout << "    mode -- optional key for selection of program mode\n";
	std::cout << "    --optimize   Optimize models parameters\n";
	std::cout << "      results will be written in out_inpfile files\n";
	std::cout << "    --tabulate  Calculate tabulated thermodynamic functions values\n";
	std::cout << "      results will be written in inpfile.csv files\n";
	std::cout << "    --tabulatetex  Calculate tabulated thermodynamic functions values\n";
	std::cout << "      results will be written in inpfile.tex files\n";
	std::cout << "Examples:\n";
	std::cout << "  cpfit_cmd natrolite.dat\n";
	std::cout << "  cpfit_cmd \"*.dat\"\n";
	std::cout << std::endl;
}

wxArrayString *GetDataFiles(wxString &filespec)
{
	wxArrayString *files = new wxArrayString();
	wxDir::GetAllFiles(".", files, filespec, wxDIR_FILES);
	return files;
}

/**
 * @brief  Processes one input file and writes results to one output file
 * @param  file  Opened and loaded data file
 * @param  inpFileName Input file name (required to generate output file name)
 * @param  mode Command line mode key (--optimize, --tabulate, --tabulatetex)
 */
void ProcessFile(XYDataFile &file, wxString inpFileName, wxString mode)
{
	wxString outFileName;
	inpFileName.Replace("\\", "/");
	if (mode == "--optimize") {
		// Optimized models parameters (including anomalies)
		autoDataFit(file);
		outFileName = inpFileName.BeforeLast('/') + "/out_" + inpFileName.AfterLast('/');
		file.SaveFile(outFileName);
	} else if (mode == "--tabulate" || mode == "--tabulatetex") {
		// Print tabulated values
		// a) calculate
		double Tmax = file.GetXiMax();
		std::vector <double> T = {5, 10, 20, 30, 40, 50};
		for (double Ti = 75; Ti <= Tmax + 25; Ti += 25) {
			T.push_back(Ti);
			if (Ti >= 275 && Ti < 300)
				T.push_back(298.15);
		}
		std::vector <double> Cp, H, S;
		LSQResult lsqRes = file.NoDataFit();
		Cp = lsqRes.CpFunc(T);
		H = lsqRes.HFunc(T);
		S = lsqRes.SFunc(T);
		// b) write to the file
		outFileName = inpFileName.BeforeLast('.') + ( (mode == "--tabulate") ? ".csv" : ".tex");
		wxTextFile outfile;
		if (wxFileExists(outFileName)) {
			std::cout << "File " << outFileName << " already exists" << std::endl;
			return;
		}
		if (!outfile.Create(outFileName)) {
			std::cout << "Cannot create file " << outFileName << std::endl;
			return;
		}
		if (mode == "--tabulate") { // Ordinary .csv output
			outfile.AddLine("T;Cp;S;H");
			for (size_t i = 0; i < T.size(); i++) {
				outfile.AddLine(wxString::Format("%.2f;%g;%g;%g", T[i], Cp[i], S[i], H[i]));
			}
		} else { // LaTeX output (useful for documents)
			outfile.AddLine("\\documentclass{article}");
			outfile.AddLine("\\usepackage[utf8]{inputenc}");
			outfile.AddLine("\\usepackage[english]{babel}");
			outfile.AddLine("\\begin{document}");
			outfile.AddLine("\\begin{tabular}{cccc}");
			outfile.AddLine("\\hline");
			outfile.AddLine(
				"$T$ / K & "
				"$\\frac{C_p^\\circ}{\\mathrm{J\\cdot (mol\\cdot K)^{-1} }}$ & "
				"$\\frac{S^\\circ}{\\mathrm{J\\cdot (mol\\cdot K)^{-1} }}$ & "
				"$\\frac{H_T^\\circ-H_0^\\circ}{\\mathrm{J\\cdot mol^{-1} }}$ \\\\");
			outfile.AddLine("\\hline");
			for (size_t i = 0; i < T.size(); i++) {
				outfile.AddLine(wxString::Format("%g & %g & %g & %g \\\\",
					T[i], Cp[i], S[i], H[i]));
			}
			outfile.AddLine("\\hline");
			outfile.AddLine("\\end{tabular}");
			outfile.AddLine("\\end{document}");
		}
		outfile.Write();
		outfile.Close();
	} else {
		std::cout << "Unknown mode " << mode << std::endl;
	}
	
}

int main(int argc, char *argv[])
{
	wxString inpFileMask, inpFileName, outFileName, mode;
	XYDataFile file;
	int errcode = 0;

	if (argc < 3) {
		printHelp();
		return 0;
	}
	// Process commant line arguments
	for (int i = 0; i < argc; i++)
		printf("%s\n", argv[i]);
	if (argc >= 3) {
		mode = wxString(argv[1]);
		inpFileMask = wxString(argv[2]);
	}
	// Load all extra terms	
	ExtraTermContainer::InitTermsTypes();
	// Process all files that match file mask	
	wxArrayString *files = GetDataFiles(inpFileMask);
	if (files->size() == 0) {
		std::cout << "No files matching " << inpFileMask << " mask found" << std::endl;
		return 0;
	}
	for (size_t i = 0; i < files->size(); i++) {
		inpFileName = (*files)[i];
		std::cout << "TASK " << (i + 1) << " FILE " << inpFileName << std::endl;
		if (!file.LoadFile(inpFileName)) {
			std::cout << "Cannot load file " << inpFileName << std::endl;
			errcode = 1;
		} else {
//			autoDataFit(file);
//			inpFileName.Replace("\\", "/");
//			outFileName = inpFileName.BeforeLast('/') + "/out_" + inpFileName.AfterLast('/');
//			file.SaveFile(outFileName);
			ProcessFile(file, inpFileName, mode);
		}
	}
	delete files;
	return errcode;
}
