/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */


#include <stdio.h>
#include <stdlib.h>
#include <vector>
#include <functional>
#include <wx/wx.h>
#include <wx/grid.h>
#include <wx/bookctrl.h>
#include <wx/dcbuffer.h>
#include "datafit.h"
#include "misc.h"
#include "customgrids.h"

// Icons for GNU/Linux and some other platforms
#ifndef __WXMSW__
#include "term_default.xpm"
#include "term_empty.xpm"
#include "term_leftexp2.xpm"
#include "term_leftexp.xpm"
#include "term_rightexp2.xpm"
#include "term_rightexp.xpm"
#include "term_skewed.xpm"
#include "term_twopeaks.xpm"
#endif

enum {
	DATACOL_TYPE = 0,
	DATACOL_T,
	DATACOL_T2,
	DATACOL_Y,
	DATACOL_ISRELDEV,
	DATACOL_WEIGHT,
	DATACOL_SERID
};


wxBEGIN_EVENT_TABLE(DataSeriesGrid, wxGrid)
	EVT_GRID_CELL_CHANGING(DataSeriesGrid::OnGridCellChanging)
	EVT_SIZE(DataSeriesGrid::OnSize)
wxEND_EVENT_TABLE()

const wxString DataSeriesGrid::dataTypeChoices[] = {"Cp", "H", "S"};


DataSeriesGrid::DataSeriesGrid(wxWindow *parent, std::function <void()> cbfFileChanged, wxWindowID id, wxPoint pos, wxSize size) : wxGrid(parent, id, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;
	CreateGrid(0, 7);
	SetColLabelValue(DATACOL_T, wxT("T1 / K"));
	SetColLabelValue(DATACOL_Y, wxT("Y"));
	SetColLabelValue(DATACOL_ISRELDEV, wxT("Rel.Dev."));
	SetColLabelValue(DATACOL_WEIGHT, wxT("\u03C9")); // omega
	SetColLabelValue(DATACOL_T2, wxT("T2 / K"));
	SetColLabelValue(DATACOL_TYPE, wxT("Type"));
	SetColLabelValue(DATACOL_SERID, wxT("Series"));
	SetColFormatBool(DATACOL_ISRELDEV);
 	SetRowLabelSize(40); 
	for (int i = 0; i < 7; i++) {
		SetColSize(i, 80);
	}
}

/**
 * \brief Resizes columns inside the table
 */
void DataSeriesGrid::OnSize(wxSizeEvent &event)
{
	ResizeCols();
}


/**
 * \brief Load existing data to the grid
 */
bool DataSeriesGrid::LoadDataFile(LSQContext &dataFile)
{
	int numOfPoints = dataFile.GetNumOfPoints();
	if (GetNumberRows() > 0) {
		DeleteRows(0, GetNumberRows());
	}

	InsertRows(0, numOfPoints);
	for (int i = 0; i < numOfPoints; i++) {
		Point point = dataFile.GetPoint(i);
		SetCellValue(i, DATACOL_T, wxString::Format("%.14g", point.x));
		SetCellValue(i, DATACOL_Y, wxString::Format("%.14g", point.y));
		SetCellValue(i, DATACOL_ISRELDEV, wxString::Format("%d", (int) point.isRelDev));
		SetCellValue(i, DATACOL_WEIGHT, wxString::Format("%.14g", point.w));
		SetCellValue(i, DATACOL_SERID, point.getSeriesID());
		
		switch (point.dataType) {		
		case DATATYPE_CP:
			SetCellValue(i, DATACOL_TYPE, "Cp");
			SetReadOnly(i, DATACOL_T2, true);
			break;
		case DATATYPE_S:
			SetCellValue(i, DATACOL_TYPE, "S");
			SetReadOnly(i, DATACOL_T2, true);
			break;
		case DATATYPE_H:
			SetCellValue(i, DATACOL_T, wxString::Format("%.14g", point.x1));
			SetCellValue(i, DATACOL_T2, wxString::Format("%.14g", point.x2));
			SetCellValue(i, DATACOL_TYPE, "H");
			SetReadOnly(i, DATACOL_T2, false);
			break;
		}
		SetCellEditor(i, DATACOL_TYPE,
			new wxGridCellChoiceEditor(WXSIZEOF(DataSeriesGrid::dataTypeChoices), DataSeriesGrid::dataTypeChoices, false));
	}
	return true;
}

/**
 * \brief  Reads data from a specially prepared wxGrid class example
 * \param  dataFile pointer to the data file in memory
 */
bool DataSeriesGrid::SaveToDataFile(LSQContext &dataFile)
{
	// Check grid size
	if (GetNumberRows() == 0) {
		dataFile.data.clear();
		return true;
	}
	// Convert grid to the data
	int npoints = GetNumberRows();
	dataFile.data.clear();
	for (int i = 0; i < npoints; i++) {
		Point point;
		long int isRelDev;
		GetCellValue(i, DATACOL_T).ToDouble(&point.x);
		GetCellValue(i, DATACOL_Y).ToDouble(&point.y);
		GetCellValue(i, DATACOL_ISRELDEV).ToLong(&isRelDev); point.isRelDev = isRelDev;
		GetCellValue(i, DATACOL_WEIGHT).ToDouble(&point.w);
		point.setSeriesID(GetCellValue(i, DATACOL_SERID));
		wxString dataType = GetCellValue(i, DATACOL_TYPE);
		if (dataType == "Cp") {
			GetCellValue(i, DATACOL_T).ToDouble(&point.x);
			point.dataType = DATATYPE_CP;
		} else if (dataType == "H") {
			GetCellValue(i, DATACOL_T).ToDouble(&point.x1);
			GetCellValue(i, DATACOL_T2).ToDouble(&point.x2);
			point.dataType = DATATYPE_H;
		} else if (dataType == "S") {
			GetCellValue(i, DATACOL_T).ToDouble(&point.x);
			point.dataType = DATATYPE_S;
		} else {
			// TODO: ERROR HANDLING
		}

		dataFile.data.push_back(point);
	}
	return true;
}


void DataSeriesGrid::Reset()
{
	if (GetNumberRows() > 0) {
		DeleteRows(0, GetNumberRows());
	}
}

/**
 * \brief Add new experimental point with default values
 */
void DataSeriesGrid::AddPoint()
{
	int ind = GetNumberRows();
	InsertRows(ind, 1);
	SetCellValue(ind, DATACOL_TYPE, "Cp");
	SetCellValue(ind, DATACOL_T, "1");
	SetCellValue(ind, DATACOL_Y, "10");
	SetCellValue(ind, DATACOL_ISRELDEV, "0");
	SetCellValue(ind, DATACOL_WEIGHT, "1");
	SetCellEditor(ind, DATACOL_TYPE,
		new wxGridCellChoiceEditor(WXSIZEOF(DataSeriesGrid::dataTypeChoices), DataSeriesGrid::dataTypeChoices, false));
	SetReadOnly(ind, DATACOL_T2, true);
}

/**
 * \brief Delete existing point(s) (selected rows)
 */
void DataSeriesGrid::DeleteSelectedPoints()
{
	wxArrayInt rows = GetSelectedRows();
	for (int i = rows.GetCount() - 1; i >= 0; i--) {
		DeleteRows(rows[i], 1);
	}
}

/**
 * \brief  Set all deviations in the table to relative
 */
void DataSeriesGrid::SetRelDev()
{
	for (int i = 0; i < GetNumberRows(); i++) {
		SetCellValue(i, DATACOL_ISRELDEV, "1");
	}
}

/**
 * \brief  Set all deviations in the table to absolute
 */
void DataSeriesGrid::SetAbsDev()
{
	for (int i = 0; i < GetNumberRows(); i++) {
		SetCellValue(i, DATACOL_ISRELDEV, "0");
	}
}

wxPoint DataSeriesGrid::GetPosition()
{
	return wxGrid::GetPosition();
}

void DataSeriesGrid::SetSize(int width, int height)
{
	wxGrid::SetSize(width, height);
}

void DataSeriesGrid::ResizeCols()
{
	ResizeWxGridCols(this);
}

/**
 * \brief Controls user input and prevents changes if it is incorrect
 */
void DataSeriesGrid::OnGridCellChanging(wxGridEvent& event)
{
	wxString strNotNumeric = wxString::FromUTF8("Not numeric parameter");
	int col = event.GetCol();
	double t;
	if (col == DATACOL_T || col == DATACOL_T2 || col == DATACOL_Y) {
		if (!event.GetString().ToDouble(&t) || t <= 0.0) {
			wxMessageBox(strNotNumeric, "CpFit", wxICON_ERROR);
			event.Veto();
		}
	} else if (col == DATACOL_WEIGHT) {
		if (!event.GetString().ToDouble(&t) || t < 0.0) {
			wxMessageBox(strNotNumeric, "CpFit", wxICON_ERROR);
			event.Veto();
		}
	} else if (col == DATACOL_TYPE) { // data type row
		wxString dataType = event.GetString();
		int row = event.GetRow();
		if (dataType == "Cp") {
			SetReadOnly(row, DATACOL_T2); // Disable T2 column
		} else if (dataType == "H") {
			SetReadOnly(row, DATACOL_T2, false); // Enable T2 column
		} else if (dataType == "S") {
			SetReadOnly(row, DATACOL_T2); // Disable T2 column
		} else {
			event.Veto();
		}
	}
	// We don't check DATACOL_ISRELDEV because if is boolean
	// We don't check DATACOL_SERID because it contains series ID in free form
	cbfFileChanged();
}


///////////////////////////////////////////////
/**
 * \brief Controls IDs
 */
enum {
	ID_BTN_CALC = 100,
	ID_CMB_PARNUM,
	ID_BTN_RESETPAR,
	ID_BTN_ADDPOINT,
	ID_BTN_DELPOINT,
	ID_BTN_SETREL,
	ID_BTN_SETABS,
	ID_GR_DATA,
};


wxBEGIN_EVENT_TABLE(DataSeriesEditor, wxPanel)
	EVT_BUTTON(ID_BTN_ADDPOINT, DataSeriesEditor::OnAddPoint)
	EVT_BUTTON(ID_BTN_DELPOINT, DataSeriesEditor::OnDeleteSelectedPoints)
	EVT_BUTTON(ID_BTN_SETREL, DataSeriesEditor::OnSetRelDev)
	EVT_BUTTON(ID_BTN_SETABS, DataSeriesEditor::OnSetAbsDev)
wxEND_EVENT_TABLE()

DataSeriesEditor::DataSeriesEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size) : wxPanel(parent, wxID_ANY, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;
	grData = new DataSeriesGrid(this, this->cbfFileChanged, ID_GR_DATA, wxDefaultPosition, wxDefaultSize);
	btnAddPoint = new wxButton(this, ID_BTN_ADDPOINT, "Add point");
	btnDelPoint = new wxButton(this, ID_BTN_DELPOINT, "Delete point");
	btnSetAbs   = new wxButton(this, ID_BTN_SETABS, "Set abs.dev.");
	btnSetRel   = new wxButton(this, ID_BTN_SETREL, "Set rel.dev.");

	// Sizers
	wxBoxSizer *buttonsSizer = new wxBoxSizer(wxHORIZONTAL);
	buttonsSizer->Add(btnAddPoint, 1, wxALL | wxEXPAND);
	buttonsSizer->Add(btnDelPoint, 1, wxALL | wxEXPAND);
	buttonsSizer->Add(btnSetAbs, 1, wxALL | wxEXPAND);
	buttonsSizer->Add(btnSetRel, 1, wxALL | wxEXPAND);
	// Main sizer
	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(grData, 1, wxALL | wxEXPAND);
	mainSizer->Add(buttonsSizer, 0, wxALL | wxEXPAND);
	this->SetSizer(mainSizer);
	this->Layout();
}

bool DataSeriesEditor::LoadDataFile(LSQContext &dataFile)
{
	return grData->LoadDataFile(dataFile);
}

bool DataSeriesEditor::SaveToDataFile(LSQContext &dataFile)
{
	return grData->SaveToDataFile(dataFile);
}

void DataSeriesEditor::Reset()
{
	grData->Reset();
}

/**
 * \brief Add new experimental point
 */
void DataSeriesEditor::OnAddPoint(wxCommandEvent &event)
{
	grData->AddPoint();
	cbfFileChanged();
}

/**
 * \brief Delete existing point(s) (selected rows)
 */
void DataSeriesEditor::OnDeleteSelectedPoints(wxCommandEvent &event)
{
	grData->DeleteSelectedPoints();
	cbfFileChanged();
}

/**
 * \brief  Set all deviations in the table to relative
 */
void DataSeriesEditor::OnSetAbsDev(wxCommandEvent &event)
{
	grData->SetAbsDev();
	cbfFileChanged();
}

/**
 * \brief  Set all deviations in the table to relative
 */
void DataSeriesEditor::OnSetRelDev(wxCommandEvent &event)
{
	grData->SetRelDev();
	cbfFileChanged();
}

void DataSeriesEditor::ResizeCols()
{
	grData->ResizeCols();
}


//////////////////////////////////////////////
enum {
	ID_BTN_ADDTERM = 100,
	ID_BTN_DELTERM,
	ID_CMB_TERMSLIST,
	ID_CMB_TERMTYPE,
	ID_GR_FIXPARAM,
	ID_GR_LSQPARAM,
	ID_TXT_TMIN,
	ID_TXT_TMAX
};

wxBEGIN_EVENT_TABLE(ExtraTermsEditor, wxPanel)
	EVT_BUTTON(ID_BTN_ADDTERM, ExtraTermsEditor::OnAddTerm)
	EVT_BUTTON(ID_BTN_DELTERM, ExtraTermsEditor::OnDeleteTerm)
	EVT_COMBOBOX(ID_CMB_TERMSLIST, ExtraTermsEditor::OnCmbTermsList)
	EVT_COMBOBOX(ID_CMB_TERMTYPE, ExtraTermsEditor::OnCmbTermType)
	EVT_GRID_CELL_CHANGING(ExtraTermsEditor::OnGridCellChanging)
	EVT_TEXT(ID_TXT_TMIN, ExtraTermsEditor::OnTminChange)
	EVT_TEXT(ID_TXT_TMAX, ExtraTermsEditor::OnTmaxChange)
wxEND_EVENT_TABLE()

/**
 * \brief Creates panel for editing extra terms in Cp models
 * \details The dialog has the next layout:
 *
 *     +------------{mainSizer}-------------------+
 *     | +------------+-----+{gridSizer}--------+ |
 *     | |            |     | +--{btnSizer}---+ | | 
 *     | | Terms:     | vvv | |  Add  |Delete | | |
 *     | |            |     | +-------+-------+ | | 
 *     | +------------+-----+----------+--------+ |
 *     | | Term type: | vvv | bitmap            | |
 *     | +------------+-----+----------+--------+ |
 *     | |            |     | +--{TmaxSizer}--+ | |
 *     | | Tmin, K:   | OOO | |Tmax, K:| OOOO | | |
 *     | |            |     | +--------+------+ | |
 *     | +-----------------------------+--------+ |
 *     +------------------------------------------+ 
 *     | +------------------+-{paramSizer}------+ |
 *     | | Fixed parameters | LSQ parameters    | |
 *     | +------------------+-------------------+ |
 *     +------------------------------------------+
 */
ExtraTermsEditor::ExtraTermsEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size)
	: wxPanel(parent, wxID_ANY, pos, size)
{
	TermsListMap = {
		{"DEFAULT", {"", wxBITMAP(term_default)}},
		{"TWOPEAKS", {"", wxBITMAP(term_twopeaks)}},
		{"LEFTEXP", {"", wxBITMAP(term_leftexp)}},
		{"RIGHTEXP", {"", wxBITMAP(term_rightexp)}},
		{"LEFTEXP2", {"", wxBITMAP(term_leftexp2)}},
		{"RIGHTEXP2", {"", wxBITMAP(term_rightexp2)}},
		{"SKEWED", {"", wxBITMAP(term_skewed)}},
		{"EMPTY",  {"", wxBITMAP(term_empty)}},
	};
	this->cbfFileChanged = cbfFileChanged;
	// Initialize terms list for combobox
	termsTypesList = ExtraTermContainer::GetTermTypesList();
	std::vector <std::string> termsBriefDescrList = ExtraTermContainer::GetTermBriefDescrList();
	int numOfTermsTypes = termsTypesList.size();
	wxString strTermTypes[numOfTermsTypes];
	for (int i = 0; i < numOfTermsTypes; i++) {
		strTermTypes[i] = termsBriefDescrList[i];
	}
	// Terms selection menu
	wxStaticText *lblTermsList = new wxStaticText(this, wxID_ANY, "Terms:");
	cbTermsList = new wxComboBox(this, ID_CMB_TERMSLIST, "");
	btnAddTerm = new wxButton(this, ID_BTN_ADDTERM, "Add");
	btnDelTerm = new wxButton(this, ID_BTN_DELTERM, "Delete");
	// Term editor
	// a) type field
	wxStaticText *lblTermType = new wxStaticText(this, wxID_ANY, "Term type:");
	cbTermType = new wxComboBox(this, ID_CMB_TERMTYPE, strTermTypes[0], wxDefaultPosition, wxDefaultSize, numOfTermsTypes, strTermTypes, wxCB_READONLY);
	// b) formula field
 	bmpFormula = new wxStaticBitmap(this, wxID_ANY, wxBITMAP(term_default));
	// c) Tmin field
	wxStaticText *lblTmin = new wxStaticText(this, wxID_ANY, "Tmin, K:");
	txtTmin = new wxTextCtrl(this, ID_TXT_TMIN, "");
	// c) Tmax field
	wxStaticText *lblTmax = new wxStaticText(this, wxID_ANY, "Tmax, K:");
	txtTmax = new wxTextCtrl(this, ID_TXT_TMAX, "");
	// d) Constants grid
	boxLSQParam = new wxStaticBox(this, wxID_ANY, "LSQ parameters");
	boxFixParam = new wxStaticBox(this, wxID_ANY, "Fixed parameters");
	wxStaticBoxSizer *boxLSQParamSizer = new wxStaticBoxSizer(boxLSQParam, wxVERTICAL);
	wxStaticBoxSizer *boxFixParamSizer = new wxStaticBoxSizer(boxFixParam, wxVERTICAL);
	grFixParam = new wxGridAutoColSize(this, ID_GR_FIXPARAM);
	boxFixParamSizer->Add(grFixParam, 0, wxALL | wxEXPAND);
	grFixParam->CreateGrid(0, 2);
	grFixParam->SetRowLabelSize(40);
	grFixParam->SetColLabelValue(0, "Name");
	grFixParam->SetColLabelValue(1, "Value");
	// e) LSQ parameters grid
	grLSQParam = new wxGridAutoColSize(this, ID_GR_LSQPARAM);
	boxLSQParamSizer->Add(grLSQParam, 0, wxALL | wxEXPAND);
	grLSQParam->CreateGrid(0, 4);
	grLSQParam->SetRowLabelSize(40);
	grLSQParam->SetColLabelValue(0, "Name");
	grLSQParam->SetColLabelValue(1, "Value");
	grLSQParam->SetColLabelValue(2, "Low.bound.");
	grLSQParam->SetColLabelValue(3, "Up.bound.");
	// Sizers
	int borderWidth = 2;

	wxFlexGridSizer *gridSizer = new wxFlexGridSizer(3, 3, 5, 5);
	gridSizer->Add(lblTermsList, 1, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(cbTermsList, 1, wxALL | wxEXPAND, borderWidth);
	wxBoxSizer *btnSizer = new wxBoxSizer(wxHORIZONTAL);
	btnSizer->Add(btnAddTerm, 0, wxALL | wxEXPAND, borderWidth);
	btnSizer->Add(btnDelTerm, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(btnSizer, 2, wxALL | wxEXPAND, borderWidth);

	gridSizer->Add(lblTermType, 1, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(cbTermType, 1, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(bmpFormula, 2, wxALL | wxEXPAND, borderWidth);

	gridSizer->Add(lblTmin, 1, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(txtTmin, 1, wxALL | wxEXPAND, borderWidth);
	wxBoxSizer *TmaxSizer = new wxBoxSizer(wxHORIZONTAL);
	TmaxSizer->Add(lblTmax, 0, wxALL | wxEXPAND, borderWidth);
	TmaxSizer->Add(txtTmax, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(TmaxSizer, 2, wxALL | wxEXPAND, borderWidth);

	wxBoxSizer *paramSizer = new wxBoxSizer(wxHORIZONTAL);
	paramSizer->Add(boxLSQParamSizer, 1, wxALL | wxEXPAND, borderWidth);
	paramSizer->Add(boxFixParamSizer, 1, wxALL | wxEXPAND, borderWidth);

	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(gridSizer, 0, wxALL | wxEXPAND, borderWidth);
	mainSizer->Add(paramSizer, 1, wxALL | wxEXPAND, borderWidth);	
	this->SetSizer(mainSizer);
	this->Layout();

	// Some default settings
	currentTermIndex = -1; // Nothing selected
	ClearEditor();
	DisableEditor();
}

/**
 * \brief Removes all extra terms from the internal buffer
 * and clears the form
 */
void ExtraTermsEditor::Reset()
{
	termsBuf.clear();
	LoadTerms(termsBuf);
}

/**
 * \brief Clear the form (Tmin and Tmax values, table, formula bitmap etc.)
 */
void ExtraTermsEditor::ClearEditor()
{
	std::vector <double> emptyVector;
	VecToTblRow(grFixParam, emptyVector);
	VecToTblRow(grLSQParam, emptyVector);

	wxBitmap emptyBitmap;
	emptyBitmap.Create(1, 1);
	bmpFormula->SetBitmap(emptyBitmap);

	txtTmin->ChangeValue("");
	txtTmax->ChangeValue("");
}

/**
 * \brief Enable/disable form for editing the extra term
 */
void ExtraTermsEditor::EnableEditor(bool enable)
{	
	txtTmin->Enable(enable);
	txtTmax->Enable(enable);
	cbTermType->Enable(enable);
	grFixParam->Enable(enable);
	grLSQParam->Enable(enable);
	btnDelTerm->Enable(enable);
}

void ExtraTermsEditor::DisableEditor()
{
	EnableEditor(false);
}

/**
 * \brief Controls grLSQParam and grFixParam wxGrid objects
 *
 * Controls data type (only doubles are accepted) and values ranges for grLSQParam
 */
void ExtraTermsEditor::OnGridCellChanging(wxGridEvent& event)
{
	// Data type control
	double t;
	bool changed = true;
	if (!event.GetString().ToDouble(&t)) {
		wxMessageBox("Not numeric parameter", "CpFit", wxICON_ERROR);
		changed = false;
	}
	// Grid-specific things
	if (event.GetId() == ID_GR_LSQPARAM) {
		size_t row = event.GetRow();
		size_t col = event.GetCol();
		if (col == 1 && row < lsqparam_lb.size() && (t < lsqparam_lb[row] || t > lsqparam_ub[row]) ) { // param
			wxMessageBox(
				wxString::Format("Parameter boundaries are [%g;%g]", lsqparam_lb[row], lsqparam_ub[row]),
				"CpFit",
				wxICON_ERROR);
			changed = false;
		} else if (col == 2 && row < lsqparam_lb.size() && t > lsqparam_ub[row]) { // param_lb
			wxMessageBox(
				wxString::Format("Lower boundary cannot be higher than upper boundary"),
				"CpFit",
				wxICON_ERROR);
			changed = false;
		} else if (col == 3 && row < lsqparam_ub.size() && t < lsqparam_lb[row]) { // param_ub
			wxMessageBox(
				wxString::Format("Upper boundary cannot be lower than lower boundary"),
				"CpFit",
				wxICON_ERROR);
			changed = false;
		}

	}
	// Change control
	if (changed) { // New value is correct: set "file is changed" flag
		SaveTerm(termsBuf[currentTermIndex]);
		LoadTerm(termsBuf[currentTermIndex]);
		cbfFileChanged();
	} else { // New value is incorrect: cancel editing
		event.Veto();
	}
}

/**
 * \brief Used by ExtraTermsEditor::OnTminChange and ExtraTermsEditor::OnTmaxChange
 * for control of T value (must be a real number between 0 and 20000 K)
 */
double ExtraTermsEditor::OnTChange(wxString str)
{
	double t = -1;
	if (!str.ToDouble(&t)) {
		wxMessageBox("Not numeric parameter", "CpFit", wxICON_ERROR);
		t = -1;
	} else if (t < 0 || t > 20000) {
		wxMessageBox("Invalid T value", "CpFit", wxICON_ERROR);
		t = -1;
	};
	return t;
}

/**
 * \brief Controls Tmin value (in GUI) change
 */
void ExtraTermsEditor::OnTminChange(wxCommandEvent &event)
{	
	double t = OnTChange(event.GetString());
	if (t == -1) {
		txtTmin->ChangeValue(wxString::Format("%g", termsBuf[currentTermIndex].term->Tmin));
	} else {
		cbfFileChanged();
	}
}

/**
 * \brief Controls Tmax value (in GUI) change
 */
void ExtraTermsEditor::OnTmaxChange(wxCommandEvent &event)
{
	double t = OnTChange(event.GetString());
	if (t == -1) {
		txtTmax->ChangeValue(wxString::Format("%g", termsBuf[currentTermIndex].term->Tmax));
	} else {
		cbfFileChanged();
	}
}


/**
 * Removes all rows from the table and adds empty rows of required size
 */
void ExtraTermsEditor::ClearTblRows(wxGrid *gr, std::vector <double> &p)
{
	int nrows = gr->GetNumberRows();
	if (nrows > 0) {
		gr->DeleteRows(0, nrows);
	}
	gr->InsertRows(0, p.size());
}

/**
 * \brief Converts double vector to wxGrid row
 */
void ExtraTermsEditor::VecToTblRow(wxGrid *gr, std::vector <double> &p, int row)
{
	for (size_t i = 0; i < p.size(); i++) {
		gr->SetCellValue(i, row, wxString::Format("%g", p[i]));
	}
}

/**
 * \brief Converts wxGrid row to double vector
 */
std::vector <double> ExtraTermsEditor::TblRowToVec(wxGrid *gr, int row)
{
	std::vector <double> vec;
	int nrows = gr->GetNumberRows();
	for (int i = 0; i < nrows; i++) {
		double val;
		gr->GetCellValue(i, row).ToDouble(&val);
		vec.push_back(val);
	}
	return vec;
}


/**
 * \brief Loads data from ExtraTerm class example into GUI
 */
bool ExtraTermsEditor::LoadTerm(ExtraTermContainer &term)
{
	// Load bitmap with formula for selected term
	std::string type = term.term->GetType();
	size_t ind = 0;
	for (size_t i = 0; i < termsTypesList.size(); i++) {
		if (termsTypesList[i] == type) {
			ind = i;
			break;
		}
	}
	cbTermType->SetSelection(ind);
	bmpFormula->SetBitmap(TermsListMap[type].second);
	Refresh();
	// Load term parameters
	txtTmin->SetValue(wxString::Format("%g", term.term->Tmin));
	txtTmax->SetValue(wxString::Format("%g", term.term->Tmax));
	std::vector <double> fixparam = term.term->GetFixParam();
	std::vector <double> lsqparam = term.term->GetLSQParam();
	ClearTblRows(grFixParam, fixparam); VecToTblRow(grFixParam, fixparam);
	ClearTblRows(grLSQParam, lsqparam); VecToTblRow(grLSQParam, lsqparam);
	lsqparam_lb = term.term->GetLSQParamLB(); VecToTblRow(grLSQParam, lsqparam_lb, 2);
	lsqparam_ub = term.term->GetLSQParamUB(); VecToTblRow(grLSQParam, lsqparam_ub, 3);
	// Load names of parameters
	std::vector <std::string> fixparam_names = term.term->GetFixParamNames();
	std::vector <std::string> lsqparam_names = term.term->GetLSQParamNames();
	for (size_t i = 0; i < fixparam_names.size(); i++) {
		grFixParam->SetCellValue(i, 0, fixparam_names[i]);
		grFixParam->SetReadOnly(i, 0, true);
	}
	for (size_t i = 0; i < lsqparam_names.size(); i++) {
		grLSQParam->SetCellValue(i, 0, lsqparam_names[i]);
		grLSQParam->SetReadOnly(i, 0, true);
	}
	return true;
}

/**
 * \brief Save data from GUI to ExtraTerm class examle
 */
bool ExtraTermsEditor::SaveTerm(ExtraTermContainer &term)
{
	int ind = cbTermType->GetSelection();	
	if (ind < 0) {
		ind = 0;
	}

	ExtraTermContainer ans(GetGUITermType());
	if (!txtTmin->GetValue().ToDouble(&ans.term->Tmin)) {
		return false;
	}
	if (!txtTmax->GetValue().ToDouble(&ans.term->Tmax)) {
		return false;
	}
	std::vector <double> lsqparam = TblRowToVec(grLSQParam);
	std::vector <double> fixparam = TblRowToVec(grFixParam);
	std::vector <double> lsqparam_lb = TblRowToVec(grLSQParam, 2);
	std::vector <double> lsqparam_ub = TblRowToVec(grLSQParam, 3);
	ans.term->SetLSQParam(lsqparam);
	ans.term->SetLSQParamLB(lsqparam_lb);
	ans.term->SetLSQParamUB(lsqparam_ub);
	ans.term->SetFixParam(fixparam);
	
	term = ans;	
	return true;
}

std::string ExtraTermsEditor::GetGUITermType()
{
	int ind = cbTermType->GetSelection();	
	if (ind < 0) {
		ind = 0;
	}
	return termsTypesList[ind];
}

/**
 * \brief Loads ExtraTerm class vector into ExtraTermsEditor internal variables
 * and shows the first term at the page
 */
bool ExtraTermsEditor::LoadTerms(std::vector <ExtraTermContainer> &terms)
{
	termsBuf = terms;
	// a) Initialize combo box
	cbTermsList->Clear();
	for (size_t i = 0; i < termsBuf.size(); i++) {
		cbTermsList->Append(wxString::Format("%d", (int) (i + 1)));
	}
	if (termsBuf.size() > 0) {
		cbTermsList->SetSelection(0);
		LoadTerm(termsBuf[0]);
		EnableEditor();
		currentTermIndex = 0;
	} else {
		ClearEditor();
		DisableEditor();
	}
	return false;
}

bool ExtraTermsEditor::SaveTerms(std::vector <ExtraTermContainer> &terms)
{
	terms = termsBuf;
	return false;
}

bool ExtraTermsEditor::LoadDataFile(LSQContext &dataFile)
{
	return LoadTerms(dataFile.extraTerms);
}

bool ExtraTermsEditor::SaveToDataFile(LSQContext &dataFile)
{
	if (currentTermIndex != -1) {
		// Save current GUI content
		SaveTerm(termsBuf[currentTermIndex]);
	}
	return SaveTerms(dataFile.extraTerms);
}


void ExtraTermsEditor::OnAddTerm(wxCommandEvent &event)
{
	ExtraTermContainer c;
	termsBuf.push_back(c);
	wxString termName = wxString::Format("%d", (int) termsBuf.size());
	cbTermsList->Append(termName);
	cbTermsList->SetValue(termName);
	if (currentTermIndex != -1) {
		// Save current GUI content
		SaveTerm(termsBuf[currentTermIndex]);
	}
	currentTermIndex = termsBuf.size() - 1;
	LoadTerm(termsBuf[currentTermIndex]);
	cbTermsList->SetSelection(termsBuf.size() - 1);
	EnableEditor();
	cbfFileChanged();
}

void ExtraTermsEditor::OnDeleteTerm(wxCommandEvent &event)
{
	int ind = cbTermsList->GetSelection();
	if (termsBuf.size() > 0 && ind >= 0) {
		wxString strMsg = wxString::Format("Do you want to delete extra term %d?", ind + 1);
		int ans = wxMessageBox(strMsg, "CpFit", wxYES_NO | wxICON_EXCLAMATION);
		if (ans == wxYES) {
			termsBuf.erase(termsBuf.begin() + ind);
			LoadTerms(termsBuf);
			if (cbTermsList->GetCount() > 0) {
				cbTermsList->SetSelection(0);
			}
			cbfFileChanged();
		}
	}
}

/**
 * \params Shows selected term parameters in the GUI
 */
void ExtraTermsEditor::OnCmbTermsList(wxCommandEvent &event)
{
	int ind = cbTermsList->GetSelection();
	if (currentTermIndex != -1) {
		// Save GUI content
		SaveTerm(termsBuf[currentTermIndex]);
	}
	LoadTerm(termsBuf[ind]);
	currentTermIndex = ind;
}

/**
 * \params Processes selection of another term type in the GUI
 * (e.g. parameters sizes etc.)
 */
void ExtraTermsEditor::OnCmbTermType(wxCommandEvent &event)
{
	// Update term type
	if (currentTermIndex != -1) {
		double Tmin = -1, Tmax = -1;
		// Save Tmax and Tmin from the old term
		if (ExtraTermsEditor::SaveTerm(termsBuf[currentTermIndex])) {
			Tmin = termsBuf[currentTermIndex].term->Tmin;
			Tmax = termsBuf[currentTermIndex].term->Tmax;
		}
		// Initialize new term, set Tmin/Tmax, save to buffer and load to form
		termsBuf[currentTermIndex] = ExtraTermContainer(GetGUITermType());
		if (Tmin != -1) {
			termsBuf[currentTermIndex].term->Tmin = Tmin;
			termsBuf[currentTermIndex].term->Tmax = Tmax;
		}
		LoadTerm(termsBuf[currentTermIndex]);
		// File changed
		cbfFileChanged();
	}
}

/** 
 * @brief Refreshes list of terms in the memory.
 * Note! It considers only new terms, i.e. assumes that terms cannot be removed!
 */
void ExtraTermsEditor::RefreshExtraTermsTypes()
{
	// Initialize terms list for combobox
	std::vector <std::string> termsTypesList_new = ExtraTermContainer::GetTermTypesList();
	std::vector <std::string> termsBriefDescrList_new = ExtraTermContainer::GetTermBriefDescrList();
	std::vector <std::string> termsFormulaList_new = ExtraTermContainer::GetTermFormulaList();
	int numOfTermsTypes = termsTypesList.size();
	int numOfTermsTypes_new = termsTypesList_new.size();
	if (numOfTermsTypes_new > numOfTermsTypes) {
		for (int i = numOfTermsTypes; i < numOfTermsTypes_new; i++) {
			std::string type = termsTypesList_new[i];
			std::string descr = termsBriefDescrList_new[i];
			std::string formula = termsFormulaList_new[i];
			termsTypesList.push_back(type);
			cbTermType->Append(descr);
			
			// Generate bitmap with formula using user-defined string
			// a) clean white canvas
			wxBitmap bmp(wxSize(300, 30));
			wxMemoryDC dc;
			dc.SelectObject(bmp);
			dc.SetPen( *wxTRANSPARENT_PEN );
			dc.SetBrush(wxBrush(wxColor(255, 255, 255)));
			dc.DrawRectangle(0,0,300, 30);
			// b) render the string
			wxFont f(12,
				wxFONTFAMILY_ROMAN,
				wxFONTSTYLE_NORMAL,
				wxFONTWEIGHT_NORMAL,
				false, "Times New Roman", wxFONTENCODING_CP1251);
			wxFont f_small(10,
				wxFONTFAMILY_ROMAN,
				wxFONTSTYLE_NORMAL,
				wxFONTWEIGHT_NORMAL,
				false, "Times New Roman", wxFONTENCODING_CP1251);
			dc.SetPen( *wxBLACK_PEN );
			dc.SetFont(f);
			wxString text = wxString::FromUTF8(formula.c_str());
			int xpos = 2;
			bool isUpper = false, isLower = false;
			for (size_t i = 0; i < text.length(); i++) {
				int w, h;
				wxString sym = text[i];
				// Special symbols processing
				if (sym == "_") {
					isLower = true; isUpper = false;
					dc.SetFont(f_small);
					continue;
				} else if (sym == "^") {
					isLower = false; isUpper = true;
					dc.SetFont(f_small);
					continue;
				}
				// Render the letter
				dc.GetTextExtent(sym, &w, &h);

				int ypos = 30 / 2 - h / 2;
				if (isLower) {
					ypos += h/3;
				} else if (isUpper) {
					ypos -= h/3;
				}
				dc.DrawLabel(sym, wxRect(xpos, ypos, w, h));
				xpos += w + 1;

				if (isLower || isUpper) {
					dc.SetFont(f);
					isLower = false; isUpper = false;
				}
			}			
			dc.SelectObject(wxNullBitmap);
			// Add description and bitmap
			TermsListMap[type] = std::pair <std::string, wxBitmap> (descr, bmp);
		}
	}
}


/////////////////////////////////////////////////////////
DataFileEditor::DataFileEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size)
	: wxPanel(parent, wxID_ANY, pos, size)
{
	this->cbfFileChanged = cbfFileChanged;
	// Create notebook with sizer	
	wxStaticBox *box = new wxStaticBox(this, wxID_ANY, "Input data");
	wxStaticBoxSizer *boxSizer = new wxStaticBoxSizer(box, wxVERTICAL);

	wxNotebook *nbData = new wxNotebook(box, wxID_ANY);
	boxSizer->Add(nbData, 1, wxALL | wxEXPAND, 5);
	// Create and add tabs
	ctlData = new DataSeriesEditor(nbData, this->cbfFileChanged, wxPoint(20, 30), wxSize(380, 230));
	ctlExtraTerms = new ExtraTermsEditor(nbData, this->cbfFileChanged, wxPoint(20, 30), wxSize(380, 230));
	nbData->AddPage(ctlData, "Data points");
	nbData->AddPage(ctlExtraTerms, "Extra terms");
	// Main sizer
	wxBoxSizer *mainSizer = new wxBoxSizer(wxHORIZONTAL);
	mainSizer->Add(boxSizer, 1, wxALL | wxEXPAND);
	this->SetSizer(mainSizer);
	this->Layout();
}

void DataFileEditor::RefreshExtraTermsTypes()
{
	ctlExtraTerms->RefreshExtraTermsTypes();
}


bool DataFileEditor::LoadDataFile(LSQContext &dataFile)
{
	if (!ctlData->LoadDataFile(dataFile)) {
		return false;
	}
	if (!ctlExtraTerms->LoadDataFile(dataFile)) {
		return false;
	}
	return true;
}

bool DataFileEditor::SaveToDataFile(LSQContext &dataFile)
{
	if (!ctlData->SaveToDataFile(dataFile)) {
		return false;
	}
	if (!ctlExtraTerms->SaveToDataFile(dataFile)) {
		return false;
	}
	return true;
}

void DataFileEditor::Reset()
{
	ctlData->Reset();
	ctlExtraTerms->Reset();
}
