/**
 * \file 
 * \brief Implementation of custom grids and other controls
 *
 * Custom controls (including grids) grids allow user
 * to view and edit data series
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */


#ifndef __CUSTOMGRIDS_H
#define __CUSTOMGRIDS_H
#include "xydatafile.h"
class DataSeriesGrid : public wxGrid
{
public:
	DataSeriesGrid(wxWindow *parent, std::function <void()> cbfFileChanged, wxWindowID id, wxPoint pos, wxSize size);
	bool LoadDataFile(LSQContext &dataFile);
	bool SaveToDataFile(LSQContext &dataFile);
	void Reset();
	void AddPoint();
	void DeleteSelectedPoints();
	void SetAbsDev();
	void SetRelDev();
	void ResizeCols();
	
	wxPoint GetPosition();
	void SetSize(int width, int height);

	void OnGridCellChanging(wxGridEvent& event);
	void OnSize(wxSizeEvent &event);

	wxDECLARE_EVENT_TABLE();
	
private:
	static const wxString dataTypeChoices[];
	std::function <void()> cbfFileChanged;
};


/**
 * \brief  Designed for extra terms editing
 */
class ExtraTermsEditor : public wxPanel
{
public:
	ExtraTermsEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size);
	void Reset();
	void ClearEditor();
	void EnableEditor(bool enable = true);
	void DisableEditor();
	bool LoadDataFile(LSQContext &dataFile);
	bool SaveToDataFile(LSQContext &dataFile);
	void RefreshExtraTermsTypes();
	std::string GetGUITermType();
	void OnTminChange(wxCommandEvent &event);
	void OnTmaxChange(wxCommandEvent &event);
	void OnAddTerm(wxCommandEvent &event);
	void OnDeleteTerm(wxCommandEvent &event);
	void OnCmbTermsList(wxCommandEvent &event);
	void OnCmbTermType(wxCommandEvent &event);
	void OnGridCellChanging(wxGridEvent& event);
	wxDECLARE_EVENT_TABLE();
	
private:
	double OnTChange(wxString str);

	void ClearTblRows(wxGrid *gr, std::vector <double> &p);
	void VecToTblRow(wxGrid *gr, std::vector <double> &p, int row = 1);
	std::vector <double> TblRowToVec(wxGrid *gr, int row = 1);
	bool LoadTerm(ExtraTermContainer &term);
	bool SaveTerm(ExtraTermContainer &term);
	bool LoadTerms(std::vector <ExtraTermContainer> &terms);
	bool SaveTerms(std::vector <ExtraTermContainer> &terms);
	
	std::vector <ExtraTermContainer> termsBuf;
	
	wxStaticBitmap *bmpFormula;
	wxComboBox *cbTermsList;
	wxComboBox *cbTermType;
	wxButton *btnAddTerm;
	wxButton *btnDelTerm;
	wxTextCtrl *txtTmin;
	wxTextCtrl *txtTmax;
	
	wxGrid *grFixParam;
	wxGrid *grLSQParam;
	wxStaticBox *boxFixParam;
	wxStaticBox *boxLSQParam;
	std::vector <double> lsqparam_lb;
	std::vector <double> lsqparam_ub;
	
	std::function <void()> cbfFileChanged;
	int currentTermIndex;

	std::map <std::string, std::pair <std::string, wxBitmap>> TermsListMap;
	std::vector <std::string> termsTypesList;
};


/**
 * \brief  Class for working with experimental data points (without extra terms)
 */
class DataSeriesEditor : public wxPanel
{
public:
	DataSeriesEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size);

	bool LoadDataFile(LSQContext &dataFile);
	bool SaveToDataFile(LSQContext &dataFile);
	void Reset();
	void OnAddPoint(wxCommandEvent &event);
	void OnDeleteSelectedPoints(wxCommandEvent &event);
	void OnSetAbsDev(wxCommandEvent &event);
	void OnSetRelDev(wxCommandEvent &event);
	void ResizeCols();
	// Controls	
	DataSeriesGrid *grData;
	// Buttons
	wxButton *btnAddPoint, *btnDelPoint, *btnSetAbs, *btnSetRel;
	
	wxDECLARE_EVENT_TABLE();
private:
	std::function <void()> cbfFileChanged;
};

/**
 * \brief Class for editing entire data file (experimental points + extra terms).
 * GUI has the format of notebook with 2 tabs for poitns and extra terms respectively
 */
class DataFileEditor : public wxPanel
{
public:
	DataFileEditor(wxWindow *parent, std::function <void()> cbfFileChanged, wxPoint pos, wxSize size);
	bool LoadDataFile(LSQContext &dataFile);
	bool SaveToDataFile(LSQContext &dataFile);
	void Reset();
	void RefreshExtraTermsTypes();
private:	
	DataSeriesEditor *ctlData;
	ExtraTermsEditor *ctlExtraTerms;
	std::function <void()> cbfFileChanged;
};


#endif
