/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <float.h>
#include <time.h>
#include <algorithm>
#include "datafit.h"
#include "dfunc3.h"
#include "levmar.h"


// Import internal function from LEVMAR library that calculates
// covariance matrix
extern "C" int dlevmar_covar(double *JtJ, double *C, double sumsq, int m, int n);

/**
 * \brief Constructor creates a point with invalid x and y values,
 * absolute deviation, Cp data type and w = 1
 */
Point::Point()
{
	x = -1.0;
	x1 = -1.0;
	x2 = -1.0;
	y = -1.0;
	isRelDev = 0;
	w = 1.0;
	dataType = DATATYPE_CP;
	seriesID[0] = 0; // ASCIIZ-string
}

void Point::setSeriesID(const char *id)
{
	int i;
	for (i = 0; i < 15 && id[i] != 0; i++) {
		seriesID[i] = id[i];
	}
	seriesID[i] = 0; // End of line
}

std::string Point::getSeriesID()
{
	return std::string(seriesID);
}


// ===== TDPEFunc constants =====
const double TDPEFunc::T0 = 298.15;


// ===== TDPEFunc constructors =====

/**
 * \brief Class example with no model parameters
 */
TDPEFunc::TDPEFunc()
{
}

/**
 * \brief Class example with user-defined model parameters
 */
TDPEFunc::TDPEFunc(const TDBaseLineFuncParams &param)
{
	this->param = param;
}

/**
 * \brief Returns current parameters set
 */
TDBaseLineFuncParams TDPEFunc::GetParam()
{
	return param;
}

int TDPEFunc::GetNumOfParams()
{
	return param.NumParams();
}

void TDPEFunc::AddTerm(double alpha, double theta)
{
	param.AddTerm(alpha, theta);
}


// ===== TDPEFunc methods =====
/**
 * \brief Cp(T) function implementation (internal parameters used)
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::CpFunc(std::vector <double> T)
{
	int npoints = T.size();
	std::vector <double> Cp(npoints);
	for (int i = 0; i < npoints; i++) {
		Cp[i] = CpFunc(T[i]);
	}
	return Cp;
}

/**
 * \brief S(T)-S(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::SFunc(std::vector <double> T)
{
	int npoints = T.size();
	std::vector <double> S(npoints);
	for (int i = 0; i < npoints; i++) {
		S[i] = SFunc(T[i]);
	}
	return S;
}

/**
 * \brief H(T)-H(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::HFunc(std::vector <double> T)
{
	int npoints = T.size();
	std::vector <double> H(npoints);
	for (int i = 0; i < npoints; i++) {
		H[i] = HFunc(T[i]);
	}
	return H;
}

/**
 * \brief H(T2)-H(T1) function implementation (internal parameteres used)
 * \param T1 Lower temperature, K
 * \param T2 Higher temperatre, K
 */
std::vector <double> TDPEFunc::DHFunc(std::vector <double> T1, std::vector <double> T2)
{
	int npoints = T1.size();
	std::vector <double> DH(npoints);
	if (T1.size() != T2.size()) {
		return DH;
	}
	for (int i = 0; i < npoints; i++) {
		DH[i] = DHFunc(T1[i], T2[i]);
	}
	return DH;
}

/**
 * \brief Cp(T) function implementation (internal parameters used)
 * \param T Temperature, K
 */
double TDPEFunc::CpFunc(double T)
{
	double Cp = 0.0;
	// a) E.-P. functions part
	size_t nterms = param.NumAlphaThetaTerms();
	for (size_t j = 0; j < nterms; j++) {
		double x = param.theta[j] / T;
		double ex = exp(x);

		if (x < 500) {
			Cp += 3*param.alpha[j]*ex*x*x / pow(ex - 1, 2);
		} else {
			Cp += 3*param.alpha[j]*x*x / ex;
		}
	}
	// b) Polynomial part
	size_t npolycoeffs = param.polyCoeff.size();
	for (size_t j = 0; j < npolycoeffs; j++) {
		Cp += param.polyCoeff[j] * pow(T / T0, param.polyDegree[j]);
	}
	Cp = Cp * CONST_R;
	// c) Extra terms part
	int nextraterms = extraTerms.size();
	for (int j = 0; j < nextraterms; j++) {
		Cp += extraTerms[j].term->CpFuncEval(T);
	}
	return Cp;
}

/**
 * \brief (&part;Cp/&part;&beta;) derivatives calculation
 * (required for Jacobian calculation)
 * \param T temperature, K
 * \param J pointer to the Jacobian buffer (must have at least nparams
 * free elements)
 *
 * This function is actively used by LSQContext::GetJVector to calculate
 * Jacobian. &beta; is generalized notation for model parameters, i.e.
 *
 * &beta;[0] = &alpha;[0], &beta;[1] = &theta;[0], &beta;[2] = &alpha;[1], &beta;[3] = &theta;[1] etc.
 *
 * Order of derivatives in memory correspond this rule
 */
void TDPEFunc::dCpdBFunc(double T, double *J)
{
	// a) E.-P. functions part
	size_t nterms = param.NumAlphaThetaTerms();
	for (size_t j = 0; j < nterms; j++) {
		double x = param.theta[j] / T;
		double ex = exp(x);

		if (x < 500) {
			*J++ = 3*ex*x*x / pow(ex - 1, 2) * CONST_R;
			*J++ = 3*param.alpha[j] / (T * pow(ex - 1, 2)) * x * ex * ( 
				(2 + x) - 2*x*ex / (ex - 1)
			) * CONST_R;
		} else {
			*J++ = 3*x*x / (ex) * CONST_R;
			*J++ = 3*param.alpha[j]*x / (T * ex) * (2 - x) * CONST_R;
		}
	}
	// b) Polynomial part
	size_t npolycoeffs = param.polyCoeff.size();
	for (size_t j = 0; j < npolycoeffs; j++) {
		*J++ = pow(T / T0, param.polyDegree[j]) * CONST_R;
	}
}


/**
 * \brief (&part;H/&part;&beta;) derivatives calculation
 * (required for Jacobian calculation)
 * \param T temperature, K
 * \param J pointer to the Jacobian buffer (must have at least nparams
 * free elements)
 *
 * This function is actively used by LSQContext::GetJVector to calculate
 * Jacobian. See TDPEFunc::param for model parameters order. 
 * Order of derivatives in memory corresponds this rule
 */
void TDPEFunc::dHdBFunc(double T, double *J)
{
	// a) E.-P. part
	size_t nterms = param.NumAlphaThetaTerms();
	for (size_t j = 0; j < nterms; j++) {
		double x = param.theta[j] / T;
		double ex = exp(x);

		if (x < 500) {
			*J++ = 3 * x / (exp(x) - 1) * T * CONST_R; 
			*J++ = 3 * param.alpha[j] * (ex - 1 - x*ex) / pow(ex - 1, 2) * CONST_R;
		} else {
			*J++ = 3 * x / (exp(x) - 1) * T * CONST_R; 
			*J++ = 3 * param.alpha[j] * (1 - x) / ex * CONST_R;
		}
	}
	// b) Polynomial part
	size_t npolycoeffs = param.polyCoeff.size();
	for (size_t j = 0; j < npolycoeffs; j++) {
		double n = param.polyDegree[j];
		if (n == -1) { // Integrated (T/T0)^(-1): 
			*J++ = T0 * log(T / T0) * CONST_R;
		} else { // Integrated (T/T0)^n
			*J++ = pow(T / T0, n) * T / (n + 1) * CONST_R;
		}
	}
}

/**
 * \brief (&part;S/&part;&beta;) derivatives calculation
 * (required for Jacobian calculation)
 * \param T temperature, K
 * \param J pointer to the Jacobian buffer (must have at least nparams
 * free elements)
 *
 * This function is actively used by LSQContext::GetJVector to calculate
 * Jacobian. See TDPEFunc::param for model parameters order. 
 * Order of derivatives in memory corresponds this rule.
 */
void TDPEFunc::dSdBFunc(double T, double *J)
{
	// a) E.-P. part
	size_t nterms = param.NumAlphaThetaTerms();
	for (size_t j = 0; j < nterms; j++) {
		double x = param.theta[j] / T;
		double ex = exp(x);

		if (x < 500) {		
			*J++ = 3*(x / (ex - 1) - log(1 - exp(-x))) * CONST_R;
			*J++ = -3*param.alpha[j]*param.theta[j] * ex / pow(T * (ex - 1), 2) * CONST_R;
		} else {
			*J++ = 3*(x / (ex - 1) - log(1 - exp(-x))) * CONST_R;
			*J++ = -3*param.alpha[j]*param.theta[j] / ex * CONST_R;
		}
	}
	// b) Polynomial part
	size_t npolycoeffs = param.polyCoeff.size();
	for (size_t j = 0; j < npolycoeffs; j++) {
		double n = param.polyDegree[j];
		*J++ = pow(T / T0, n) / n * CONST_R;
	}
}

/**
 * \brief S(T)-S(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
double TDPEFunc::SFunc(double T)
{
	double S = 0.0;
	// a) E.-P. part
	size_t nterms = param.NumAlphaThetaTerms();
	for (size_t j = 0; j < nterms; j++) {
		double x = param.theta[j] / T;
		double ex = exp(x);		
		// TODO: FOR X > 500
		S += 3*param.alpha[j]*(x / (ex - 1) - log(1 - exp(-x)));
	}
	S *= CONST_R;
	// b) Polynomial part
	size_t npolycoeffs = param.polyCoeff.size();
	for (size_t j = 0; j < npolycoeffs; j++) {
		double n = param.polyDegree[j];
		S += param.polyCoeff[j] * pow(T / T0, n) / n * CONST_R;
	}
	// c) Extra terms part
	int nextraterms = extraTerms.size();
	for (int j = 0; j < nextraterms; j++) {
		S += extraTerms[j].term->SFuncEval(T);
	}
	return S;
}

/**
 * \brief H(T)-H(0) function implementation (internal parameters used)
 * \param T Temperature, K
 */
double TDPEFunc::HFunc(double T)
{
	double H = 0.0;
	// a) E.-P. part
	size_t nterms = param.NumAlphaThetaTerms();
	for (size_t j = 0; j < nterms; j++) {
		double x = param.theta[j] / T;
		H += 3 * param.alpha[j] * x / (exp(x) - 1) * T;
	}
	// b) Polynomial part
	size_t npolycoeffs = param.polyCoeff.size();
	for (size_t j = 0; j < npolycoeffs; j++) {
		double n = param.polyDegree[j];
		if (n == -1) { // Integrated (T/T0)^(-1): 
			H += param.polyCoeff[j] * T0 * log(T / T0);
		} else { // Integrated (T/T0)^n
			H += param.polyCoeff[j] * pow(T / T0, n) * T / (n + 1);
		}
	}
	H *= CONST_R;
	// c) Extra terms part
	int nextraterms = extraTerms.size();
	for (int j = 0; j < nextraterms; j++) {
		H += extraTerms[j].term->HFuncEval(T);
	}
	return H;
}

//std::vector <double> DHFunc(std::vector <double> T1, std::vector <double> T2);
/**
 * \brief H(T2)-H(T1) function implementation (internal parameteres used)
 * \param T1 Lower temperature, K
 * \param T2 Higher temperatre, K
 */
double TDPEFunc::DHFunc(double T1, double T2)
{
	return HFunc(T2) - HFunc(T1);
}

/**
 * \brief Generalized implementation of (Cp(T), S(T)-S(0) and H(T)-H(0) functions) that use
 * internal parameters. 
 * \param dataType  Used data type
 * \param T Temperature, K
 */
std::vector <double> TDPEFunc::YFunc(TDDataType dataType, std::vector <double> T)
{
	std::vector <double> y;
	switch (dataType) {
	case DATATYPE_CP:
		y = CpFunc(T);
		break;
	case DATATYPE_S:
		y = SFunc(T);
		break;
	case DATATYPE_H:
		y = HFunc(T);
		break;
	case DATATYPE_UNKNOWN: // TODO: ERROR-LIKE BEHAVIOUR
		y = T;
		break;
	}	
	return y;
}

// Used only by TDPEFunc::EvalDebyeTInPoint
void func_debyep(double *TD, double *res, int m, int n, void *adata)
{
	double *vals = (double *) adata;
	double T0 = vals[0], natoms = vals[1], Cp = vals[2];
	
	double Cp_D = natoms * 3.0*CONST_R*c_debye_func(TD[0] / T0);
	res[0] = Cp_D - Cp;
}

/**
 * \brief Evaluates Debye temperature for one point with T0 using Cp curve
 *
 * This function uses the next estimation of Debye temperature TD:
 * 1) Calculate Cp(T0)
 * 2) Find TD for Debye model by means of the least squares method using Debye function
 */
double TDPEFunc::EvalDebyeTInPoint(double T0, int natoms)
{
	double inpdata[] = {T0, (double) natoms, this->CpFunc(T0)};
	double TD = 100.0, TD_lb = 0.0, TD_ub  = 1e6, covar;
	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];

	dlevmar_bc_dif(
		func_debyep, &TD, NULL, 1, 1,
		&TD_lb, &TD_ub, NULL, // Constraints (box)
		500, // Itmax
		opts, info, NULL, &covar,
		(void *) inpdata);
	// TODO: ERROR CHECK!!!
	return TD;
}

// Used only by TDPEFunc::EvalDebyeTNatInPoint
void func_debyepder(double *par, double *res, int m, int n, void *adata)
{
	double *vals = (double *) adata;
	double T0 = vals[0], Cp = vals[1], dCp_dT = vals[2];
	double TD = par[0], natoms = par[1];
	double dT = 1e-2;

	// Calculate function value	
	double Cp_D = natoms * 3.0*CONST_R*c_debye_func(TD / T0);
	res[0] = (Cp_D - Cp) / Cp_D;
	// And derivatives
	double Cp_D2 = natoms * 3.0*CONST_R*c_debye_func(TD / (T0 + dT));
	double Cp_D1 = natoms * 3.0*CONST_R*c_debye_func(TD / (T0 - dT));
	res[1] = ((Cp_D2 - Cp_D1) / (2 * dT) - dCp_dT) / dCp_dT;
}


/**
 * \brief Evaluates Debye temperature and number of atoms for one point with T0 using
 * Cp furve
 *
 * This function uses the next estimation of Debye temperature TD:
 * 1) Calculate Cp(T0) and dCp_dT(T0)
 * 2) Find TD and Natoms for Debye model by means of the least squares method using Debye function
 *
 * Returns a vector with [TD, Natoms, res] values where
 * TD is Debye temperature, Natoms is number of atoms, res is residual for equation solving
 */
std::vector <double> TDPEFunc::EvalDebyeTNatInPoint(double T0)
{
	double inpdata[] = {T0, this->CpFunc(T0), 0.0};
	double par[] = {100.0, 10}, par_lb[] = {0.0, 0.0}, par_ub[] = {1e6, 1e6};
	double covar[] = {0.0, 0.0, 0.0, 0.0};
	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];
	double dT = 1e-2;

	// Find dCp_dT numerically
	inpdata[2] = (this->CpFunc(T0 + dT) - this->CpFunc(T0 - dT)) / (2*dT);		
	// Optimize parameters
	dlevmar_bc_dif(
		func_debyepder, par, NULL, 2, 2,
		par_lb, par_ub, NULL, // Constraints (box)
		500, // Itmax
		opts, info, NULL, covar,
		(void *) inpdata);
	// Calculate uncertainties
	double rr[2];
	func_debyepder(par, rr, 2, 2, (void *) inpdata);
	// Return TD and Natoms
	std::vector <double> res;
	res.push_back(par[0]); // TD
	res.push_back(par[1]); // Natoms
	res.push_back(sqrt((rr[0]*rr[0] + rr[1]*rr[1]) / 2)); // std for eq.solver
	return res;
}


/**
 * \brief  Generates TDPEFunc without extra terms
 */
TDPEFunc TDPEFunc::GetBaseLineFuncs()
{
	return TDPEFunc(param);
}




// =============== LSQContext methods ========================/

LSQContext::LSQContext()
{
	useConstraints = false;
	Natoms = -1; // Varied number of Natoms
}


/**
 * \brief Return a point with a required index
 */
Point LSQContext::GetPoint(int index)
{
	return data[index];
}

/**
 * \brief Return the number of points
 */
int LSQContext::GetNumOfPoints()
{
	return data.size();
}

/**
 * \brief Return the number of points that are included into baseline
 * (i.e. part of data that doesn't contain 2nd order phase transitions)
 */
int LSQContext::GetNumOfBaseLinePoints()
{
	return baseLineData.size();
}

/**
 * \brief Add new point to the class
 */
void LSQContext::PushPoint(Point point)
{
	data.push_back(point);
}


bool LSQContext::IsSeriesEmpty(std::string serID)
{
	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].getSeriesID() == serID) {
			return false;
		}
	}
	return true;
}

bool LSQContext::IsSeriesEmpty(std::string serID, TDDataType ftype)
{
	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].getSeriesID() == serID && data[i].dataType == ftype) {
			return false;
		}
	}
	return true;
}

double LSQContext::GetXiMin(std::string serID)
{
	double xmin = 1e100;
	bool allSeries = (serID == "-");

	for (unsigned int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			if (data[i].dataType == DATATYPE_H) {
				if (xmin > data[i].x1) xmin = data[i].x1;
				if (xmin < data[i].x2) xmin = data[i].x2;
			} else {
				if (xmin > data[i].x) xmin = data[i].x;
			}
		}
	}
	return xmin;
}

double LSQContext::GetXiMax(std::string serID)
{
	double xmax = -1e100;
	bool allSeries = (serID == "-");

	for (unsigned int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			if (data[i].dataType == DATATYPE_H) {
				if (xmax < data[i].x1) xmax = data[i].x1;
				if (xmax < data[i].x2) xmax = data[i].x2;
			} else {
				if (xmax < data[i].x) xmax = data[i].x;
			}
		}
	}
	return xmax;
}



/**
 * \brief  Returns a vector that contains all T values for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetXVector(std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (unsigned int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			x.push_back(data[i].x);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains T values for the
 * selected experimental data subset (Cp, H or S).
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetXVector(TDDataType dataType, std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");

	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			x.push_back(data[i].x);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains all T1 values for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetX1Vector(std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (unsigned int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			x.push_back(data[i].x1);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains T1 values for the
 * selected experimental data subset (Cp, H or S).
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetX1Vector(TDDataType dataType, std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			x.push_back(data[i].x1);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains all T2 values for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetX2Vector(std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");
	
	for (unsigned int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			x.push_back(data[i].x2);
		}
	}
	return x;
}

/**
 * \brief  Returns a vector that contains T2 values for the
 * selected experimental data subset (Cp, H or S).
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetX2Vector(TDDataType dataType, std::string serID)
{
	std::vector <double> x;
	bool allSeries = (serID == "-");

	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			x.push_back(data[i].x2);
		}
	}
	return x;
}


/**
 * \brief  Returns a vector that contains all values (Cp, H, S) for 
 * all points that are kept inside the class
 */
std::vector <double> LSQContext::GetYVector(std::string serID)
{
	std::vector <double> y;
	bool allSeries = (serID == "-");
	
	for (unsigned int i = 0; i < data.size(); i++) {
		if (allSeries || data[i].getSeriesID() == serID) {
			y.push_back(data[i].y);
		}
	}
	return y;
}

/**
 * \brief  Returns a vector that contains values (Cp, H or S) for the
 * selected experimental data subset.
 *
 * \param dataType  Type of data used for selection of experimental data subset
 */
std::vector <double> LSQContext::GetYVector(TDDataType dataType, std::string serID)
{
	std::vector <double> y;
	bool allSeries = (serID == "-");

	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].dataType == dataType && (allSeries || data[i].getSeriesID() == serID)) {
			y.push_back(data[i].y);
		}
	}
	return y;
}


/**
 * \brief Calculates Jacobian for the parameters
 * \details It uses the next formula for derivatives:
 *
 * \f[
 * \frac{\partial f_i}{\partial\xi_j} =
 *   \sum_{k=1}^m \frac{\partial f_i}{\partial \alpha_k} \frac{\partial \alpha_k}{\partial \xi_j} =
 *   \sum_{k=j}^m \frac{\partial f_i}{\partial \alpha_k} \frac{\partial \alpha_k}{\partial \xi_j}
 * \f]
 *
 * \f[
 * \frac{\partial\alpha_k}{\partial\xi_j} = \begin{cases}
 *   0 & \textrm{where } j > k \\
 *   \alpha_k / \xi_j = N_{atoms}\prod_{p=1,p\neq j}^{k-1} \xi_p & \textrm{where } j < k \textrm{ and } k = m \\
 *   \alpha_k / \xi_j = N_{atoms}(1 - \xi_k)\prod_{p=1,p\neq j}^{k-1} \xi_p & \textrm{where } j < k \textrm{ and } k \neq m \\
 *   \alpha_k / (\xi_j - 1) = -N_{atoms}\prod_{p=1}^{k-1} \xi_p & \textrm{where } j = k \textrm{ and } k \neq m
 * \end{cases}
 * \f]
 */
std::vector <double> LSQContext::GetJVectorKsi(TDPEFunc &tdf)
{
	// Prepare dalphai_dksij derivatives table
	auto fparam = tdf.GetParam();
	auto ksi = fparam.GetKsi();
	auto alpha = fparam.alpha;
	int nterms = fparam.NumAlphaThetaTerms();
	int nksiparam = fparam.NumParams() - 1;
	// Calculate dalpha_dksi derivatives dtable
	std::vector <std::vector<double>> dalpha_dksi(nterms, std::vector<double>(nterms - 1, 0.0));
	std::vector <double> ksiprod(nterms - 1, 1.0);
	double dalphak_dksik = -Natoms;
	for (int k = 0; k < nterms; k++) {
		// a1) prepare ksi product
		if (k > 0) {
			for (int j = 0; j < nterms - 1; j++) {
				if (k - 1 != j) {
					ksiprod[j] *= ksi[k - 1];
				}
			}
		}
		// a2) outside the main diagonal
		for (int j = 0; j < k; j++) {
			if (k != nterms - 1) {
				dalpha_dksi[k][j] = Natoms * (1 - ksi[k]) * ksiprod[j];
			} else {
				dalpha_dksi[k][j] = Natoms * ksiprod[j];
			}
		}
		// a3) at the main diagonal
		if (k < nterms - 1) {
			dalpha_dksi[k][k] = dalphak_dksik;
			dalphak_dksik *= ksi[k];
		}
	}
	// Get Jacobian for alpha-theta
	std::vector <double> Jv = GetJVector(tdf);
	int npoints = Jv.size() / (nksiparam + 1);
	std::vector <double> Jvksi(nksiparam*npoints, 0.0);
	// and convert to Jacobian for ksi-theta
	double *Jptr = Jv.data();
	double *Jksiptr = Jvksi.data();
	for (int i = 0; i < npoints; i++) {
		// Calculate _dalpha
		for (int j = 0; j < nterms - 1; j++) {
			double dfi_dksij = 0.0;
			for (int k = j; k < nterms; k++) {
				double dfi_dalphak = Jptr[2*k];
				dfi_dksij += dfi_dalphak * dalpha_dksi[k][j];
			}
			Jksiptr[2*j + 1] = dfi_dksij;
		}
		// Copy _dtheta
		for (int j = 0; j < nterms; j++) {
			Jksiptr[2*j] = Jptr[2*j + 1];
		}
		// Copy polynomial part
		for (size_t j = 0; j < fparam.polyCoeff.size(); j++) {
			Jksiptr[2*nterms - 1 + j] = Jptr[2*nterms + j];
		}
		// Update pointers
		Jptr += nksiparam + 1;
		Jksiptr += nksiparam;
	}
	// Return Jacobian
	return Jvksi;

}


/**
 * \brief Returns residuals vectors using experimental data inside the class
 * and user-supplied functions.
 *
 * \param tdf  User-supplied functions
 *
 * This function automatically takes into account data types (Cp, H, S) and
 * the type of error (absolute or residual).
 */
std::vector <double> LSQContext::GetResVector(TDPEFunc &tdf)
{
	std::vector <Point> &dat = baseLineData;
	std::vector <double> res(dat.size());

	for (unsigned int i = 0; i < dat.size(); i++) {
		// a) Calculated value
		double ycalc = 0.0;
		switch (dat[i].dataType) {
		case DATATYPE_CP:
			ycalc = tdf.CpFunc(dat[i].x);
			break;
		case DATATYPE_S:
			ycalc = tdf.SFunc(dat[i].x);
			break;
		case DATATYPE_H:
			ycalc = tdf.HFunc(dat[i].x2) - tdf.HFunc(dat[i].x1);
			break;
		case DATATYPE_UNKNOWN:
			ycalc = NAN;
			break;
		}
		// b) weight
		double w = dat[i].w;
		if (dat[i].isRelDev) {
			w /= dat[i].y;
		}
		// c) calculate residual
		res[i] = (ycalc - dat[i].y) * w;
	}
	return res;
}

/**
 * \brief Returns Jacobian (matrix) in the form of vector
 *
 * It uses the next order of elements:
 *
 * [&part;Y<sub>0</sub>/&part;&beta;, &part;Y<sub>1</sub>/&part;&beta;, ...]
 */
std::vector <double> LSQContext::GetJVector(TDPEFunc &tdf)
{
	std::vector <Point> &dat = baseLineData;
	int npoints = dat.size();
	int nparams = tdf.GetNumOfParams();
	std::vector <double> J(npoints * nparams);
	double *Jptr = &J[0];
	double der2[nparams]; 
	double der1[nparams];

	
	for (int i = 0; i < npoints; i++) {
		// a) Calculated value
		switch (dat[i].dataType) {
		case DATATYPE_CP:
			tdf.dCpdBFunc(dat[i].x, Jptr);
			break;
		case DATATYPE_S:
			tdf.dSdBFunc(dat[i].x, Jptr);
			break;
		case DATATYPE_H:
			tdf.dHdBFunc(dat[i].x2, der2);
			tdf.dHdBFunc(dat[i].x1, der1);
			for (int j = 0; j < nparams; j++) {
				Jptr[j] = der2[j] - der1[j];
			}
			break;
		case DATATYPE_UNKNOWN: // TODO: ERROR-LIKE BEHAVIOUR
			break;
		}
		// b) weight
		double w = dat[i].w;
		if (dat[i].isRelDev) {
			w /= dat[i].y;
		}
		for (int j = 0; j < nparams; j++) {
			Jptr[j] *= w;
		}
		// Go to the next line
		Jptr += nparams;
	}
	return J;
}


/**
 * \brief  Runs parameters optimization procedure using
 * the least squares method. It doesn't require an initial approximation
 */
LSQResult LSQContext::DataFit(bool optExTerms)
{
	TDBaseLineFuncParams param0;
	if (Natoms < 0) {
		if (param0.polyCoeff.size() > 0) {
			param0.AddTerm(0.25, 100);
		} else {
			param0.AddTerm(0.1, 10);
		}
	} else {
		param0.AddTerm(Natoms, 10);
	}
	return DataFitAddTerms(param0, optExTerms);
}

/**
 * \brief  Runs parameters optimization procedure using
 * the least squares method. It doesn't require an initial approximation
 * for alpha-theta but uses user-supplied polynomial coefficients
 */
LSQResult LSQContext::DataFitWithPoly(const std::vector <double> &polyCoeff,
	const std::vector <double> &polyDegree, bool optExTerms)

{
	TDBaseLineFuncParams param0;
	param0.polyCoeff = polyCoeff;
	param0.polyDegree = polyDegree;
	if (Natoms < 0) {
		if (param0.polyCoeff.size() > 0) {
			param0.AddTerm(0.25, 100);
		} else {
			param0.AddTerm(0.1, 10);
		}
	} else {
		param0.AddTerm(Natoms, 10);
	}
	return DataFitAddTerms(param0, optExTerms);
}


/**
 * \brief  Runs parameters optimization procedure with addition of extra
 * statistically significant terms
 */
LSQResult LSQContext::DataFitAddTerms(const TDBaseLineFuncParams &param, bool optExTerms)
{
	LSQResult res, resPrev;
	bool isSignificant, isFirst = true;
	// Initial approximation
	res.SetParam(param);
	// Parameters search cycle
	do {
		res = DataFit(res.GetParam(), optExTerms);
		isSignificant = res.AreAllParamsSignificant();
		// If parameters are significant --- update initial approximation
		if (isSignificant/* && res.IsExitCodeGood()*/) {
			resPrev = res;
			if (Natoms < 0) {
				if (param.polyCoeff.size() > 0) {
					res.AddTerm(0.25, 100);
				} else {
					res.AddTerm(1e-3, 1);
				}
			} else {
				res.AddTerm(1, 100);
			}
		} else if (isFirst) {
			resPrev = res;
		}
		// All other iterations will not be first
		isFirst = false;
	} while (isSignificant);
	// Sort (baseline) parameters
	auto resSortedParams = resPrev.GetParam();
	resSortedParams.SortParameters();
	resPrev = DataFit(resSortedParams, optExTerms);
	// Return result with optimized and sorted parameters
	return resPrev;
}

/**
 * \brief Runs &alpha; and &theta; parameters optimization procedure
 * for the fixed number of atoms, i.e. \f$ \sum_i \alpha_i = N_\mathrm{atoms}\f$
 */
void LSQContext::DataFitBaseLineFixNat(LSQResult &ans, const TDBaseLineFuncParams &param)
{
	paramsBuf = param;
	size_t nparams = paramsBuf.NumParams();
	size_t npoints = this->data.size();
	size_t nBLpoints = baseLineData.size();
	ans.numOfPoints = npoints;
	paramsBuf.Natoms = Natoms;
	std::vector <double> ksiparam = paramsBuf.SerializeFixNatoms();
	std::vector <double> ksiparam_out = ksiparam;

	
	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];
	
	// Function returns residuals vector to levmar library	
	auto func = [] (double *ksiparam, double *res, int m, int n, void *adata) -> void {
		LSQContext *lsq = (LSQContext *) adata;
		std::vector <double> ksiparam_v(ksiparam, ksiparam + m);
		lsq->paramsBuf.DeserializeFixNatoms(ksiparam_v);
		TDPEFunc tdf(lsq->paramsBuf);
	
		std::vector <double> resv = lsq->GetResVector(tdf);
		std::copy(resv.begin(), resv.end(), res);
	};
	
	// Function returns residuals vector Jacobian to levmar library
	auto jacf = [] (double *ksiparam, double *J, int m, int n, void *adata) -> void {
		LSQContext *lsq = (LSQContext *) adata;
		std::vector <double> ksiparam_v(ksiparam, ksiparam + m);
		lsq->paramsBuf.DeserializeFixNatoms(ksiparam_v);
		TDPEFunc tdf(lsq->paramsBuf);
		
		std::vector <double> Jv = lsq->GetJVectorKsi(tdf);
		std::copy(Jv.begin(), Jv.end(), J);
	};	
	// Prepare constraints
	// a) ksi/theta
	size_t nterms = paramsBuf.NumAlphaThetaTerms();
	size_t npolycoeff = paramsBuf.polyCoeff.size();
	std::vector <double> ksiparam_lb(nparams - 1), ksiparam_ub(nparams - 1);
	for (size_t i = 0, j = 1; i < nterms - 1; i++, j += 2) { // ksi
		ksiparam_lb[j] = 0; ksiparam_ub[j] = 1;
	}
	for (size_t i = 0, j = 0; i < nterms; i++, j += 2) { // theta
		ksiparam_lb[j] = 0; ksiparam_ub[j] = 100000;
	}
	// b) Polynomial
	for (size_t i = 0; i < npolycoeff; i++) {
		ksiparam_lb[2*nterms - 1 + i] = 0;
		ksiparam_ub[2*nterms - 1 + i] = 1e10;
	}
	// Run optimization
	dlevmar_bc_der( // dif,der
		func, jacf, &ksiparam_out[0], NULL, nparams - 1, nBLpoints,
		&ksiparam_lb[0], &ksiparam_ub[0], NULL, // Constraints (box)
		500, // Itmax
		opts, info, NULL, NULL,
		(void *) this);
	ans.SetExitCode(info[6]);
	paramsBuf.DeserializeFixNatoms(ksiparam_out);
	// Save parameters and covariance matrix
	ans.SetParam(paramsBuf, *this); // To set up covariance matrix
}

/**
 * \brief Runs optimization of &alpha; and &theta; parameters and ignores
 * extra excess terms.
 */
void LSQContext::DataFitBaseLine(LSQResult &ans, const TDBaseLineFuncParams &param)
{
	paramsBuf = param;
	int nparams = paramsBuf.NumParams();
	int npoints = this->data.size();
	int nBLpoints = baseLineData.size();
	ans.numOfPoints = npoints;
	std::vector <double> param_out_vec = paramsBuf.Serialize();
	std::vector <double> covar(nparams * nparams);

	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-20, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];

	// Function returns residuals vector to levmar library	
	auto func = [] (double *param, double *res, int m, int n, void *adata) -> void {
		LSQContext *lsq = (LSQContext *) adata;
		std::vector <double> param_vec(param, param + m);
		lsq->paramsBuf.Deserialize(param_vec);
		TDPEFunc tdf(lsq->paramsBuf);
	
		std::vector <double> resv = lsq->GetResVector(tdf);
		std::copy(resv.begin(), resv.end(), res);
	};

	// Function returns residuals vector Jacobian to levmar library
	auto jacf = [] (double *param, double *J, int m, int n, void *adata) -> void {
		LSQContext *lsq = (LSQContext *) adata;
		std::vector <double> param_vec(param, param + m);
		lsq->paramsBuf.Deserialize(param_vec);
		TDPEFunc tdf(lsq->paramsBuf);
		
		std::vector <double> Jv = lsq->GetJVector(tdf);
		std::copy(Jv.begin(), Jv.end(), J);	
	};

	if (useConstraints) { // Constrained optimization
		std::vector <double> param_lb = paramsBuf.GetLowerBoundaries().Serialize();
		std::vector <double> param_ub = paramsBuf.GetUpperBoundaries().Serialize();
		dlevmar_bc_der(
			func, jacf, &param_out_vec[0], NULL, nparams, nBLpoints,
			&param_lb[0], &param_ub[0], NULL, // Constraints (box)
			500, // Itmax
			opts, info, NULL, &covar[0],
		(void *) this);
	} else { // Unconstrained optimization (can give negative values of parameters)	
		dlevmar_der(
			func, jacf, &param_out_vec[0], NULL, nparams, nBLpoints,
			500, // Itmax
			opts, info, NULL, &covar[0],
			(void *) this);
	}
	ans.SetExitCode(info[6]);
	// Save parameters and covariance matrix
	paramsBuf.Deserialize(param_out_vec);
	ans.SetParam(paramsBuf);
	ans.SetCovar(covar);	
}


/**
 * \brief  Runs parameters optimization procedure using
 * the least squares method and user-supplied initial approximation.
 * \param param Initial approximation
 * \param optExTerms Optimize/not optimize extra terms parameters
 */
LSQResult LSQContext::DataFit(const TDBaseLineFuncParams &param, bool optExTerms, bool optBaseLine)
{
	LSQResult ans;

	InitBaseLineData();
	
	// levmar library call
	if (optBaseLine) {
		if (Natoms < 0) {
			DataFitBaseLine(ans, param);
		} else {
			DataFitBaseLineFixNat(ans, param);			
		}
	} else {
		// Use non-optimized parameters
		ans.SetParam(param, *this);
	}
	// Evaluate extra terms if required
	if (optExTerms) {
		// Real numeric evaluation
		std::vector <ExtraTermContainer> termsBuf;
		// a) Create copies and optimize parameters
		for (size_t i = 0; i < extraTerms.size(); i++) {
			termsBuf.push_back(extraTerms[i]);
			termsBuf[i].term->DataFit(this->data, ans);
		}
		// b) Write optimized parameters to the answer
		for (size_t i = 0; i < termsBuf.size(); i++) {
			ans.extraTerms.push_back(termsBuf[i]);
		}
	} else {
		// Just copy extra terms without optimization
		std::vector <ExtraTermContainer> termsBuf;
		for (size_t i = 0; i < extraTerms.size(); i++) {
			termsBuf.push_back(extraTerms[i]);
			// TODO: CONFIDENCE INTERVALS ESTIMATOR!!!
		}
		for (size_t i = 0; i < extraTerms.size(); i++) {
			ans.extraTerms.push_back(termsBuf[i]);
		}
	}
	// Estimate absolute, relative and maximal deviations for all variables and data series
	ans.LoadDeviations(*this);
	// Return results
	return ans;
}

/**
 * \brief Just loads model parameters into LSQResult class example (together with
 * initializations of residuals and covariance matrices)
 */
LSQResult LSQContext::NoDataFit(const TDBaseLineFuncParams &param)
{
	LSQResult lsqRes;
	TDPEFunc baseLineFuncs;
	lsqRes.SetParam(param, *this);
	// b) Extra terms parameters
	baseLineFuncs = lsqRes.GetBaseLineFuncs();
	lsqRes.extraTerms = this->extraTerms;
	for (size_t i = 0; i < this->extraTerms.size(); i++) {
		std::vector <double> p = lsqRes.extraTerms[i].term->GetLSQParam();
		lsqRes.extraTerms[i].term->SetLSQParam(p, this->data, baseLineFuncs);
	}
	lsqRes.SetParam(param, *this); // To update deviations (HACK!!!)
	return lsqRes;
}

/**
 * \brief Load model parameters from alpha_0 and theta_0 variables (used by XYDataFile
 * class) into LSQResult class example (together with initializations of residuals
 * and covariance matrices)
 */
LSQResult LSQContext::NoDataFit()
{
	return NoDataFit(params_0);
}



struct seriesIDCustomLess {
	inline bool operator() (const Point &a, const Point &b) const {
		return ( (strcmp(a.seriesID, b.seriesID) < 0) ? true : false );
	}
};

struct seriesIDCustomIsEq {
	inline bool operator() (const Point &a, const Point &b) const {
		return ( (strcmp(a.seriesID, b.seriesID) == 0) ? true : false );
	}
};



std::vector <std::string> LSQContext::GetSeriesIDList(bool returnEmptySeries, TDDataType ftype)
{
	std::vector <std::string> idList;
	std::vector <Point> d = data;

	
	std::sort(d.begin(), d.end(), seriesIDCustomLess());
	d.erase(std::unique(d.begin(), d.end(), seriesIDCustomIsEq()), d.end());

	if (ftype == DATATYPE_UNKNOWN) {
		for (unsigned int i = 0; i < d.size(); i++) {
			if (returnEmptySeries || !IsSeriesEmpty(d[i].getSeriesID())) {
				idList.push_back(d[i].getSeriesID());
			}
		}
	} else {
		for (unsigned int i = 0; i < d.size(); i++) {
			if (returnEmptySeries || !IsSeriesEmpty(d[i].getSeriesID(), ftype)) {
				idList.push_back(d[i].getSeriesID());
			}
		}		
	}

	return idList;
}

bool LSQContext::IsPointOnBaseLine(Point &p)
{
	bool isBaseLine = true;
	int nterms = extraTerms.size();
	for (int j = 0; j < nterms; j++) {
		ExtraTerm *term = extraTerms[j].term;
		// TODO: ENTHALPY ISSUES!
		if ( (p.dataType == DATATYPE_CP && (term->Tmin <= p.x && p.x <= term->Tmax)) ||
		     (false && p.dataType == DATATYPE_H && (
		     (p.x1 <= term->Tmin && term->Tmin <= p.x2) ||
		     (p.x1 <= term->Tmax && term->Tmax <= p.x2) ))
		) {
			isBaseLine = false;
		}
	}
	return isBaseLine;
}


void LSQContext::InitBaseLineData()
{
	int npoints = data.size();
	baseLineData.clear();	
	for (int i = 0; i < npoints; i++) {
		Point &p = data[i];
		if (IsPointOnBaseLine(p)) {
			baseLineData.push_back(p);
		}
	}
}

/**
 * \brief Excludes all points with w=0, i.e. zero statistical weights.
 */
void LSQContext::ExcludeW0Points()
{
	std::vector <Point> filteredData;
	for (size_t i = 0; i < data.size(); i++) {
		if (data[i].w != 0)
			filteredData.push_back(data[i]);
	}
	data = filteredData;
}


/**
 * \brief Sorts all data points by seriesID, data type and T
 * Data type is more important than T. In the case of heat content
 * T1 is more important than T2.
 */
void LSQContext::SortData()
{
	std::function <bool (const Point &, const Point &)> sortfunc = [] (const Point &a, const Point &b) -> bool
		{
			// Compare seriesID
			Point a_cpy = a, b_cpy = b;
			std::string ref1 = a_cpy.getSeriesID(), ref2 = b_cpy.getSeriesID();
			if (ref1 < ref2) {
				return true;
			} else if (ref1 > ref2) {
				return false;
			}			
			// Compare data type (sort inside references ID)			
			if (a.dataType > b.dataType) {
				return false;
			} else if (a.dataType < b.dataType) {
				return true;
			}
			// Sorting inside one datatype			
			if (a.dataType == DATATYPE_H) { // a) heat content
				if (a.x1 > b.x1) { // a) sort by T1
					return false;
				} else if (a.x1 < b.x1) {
					return true;
				} else { // b) sort by T2 (if T1 are equal)
					return (a.x2 < b.x2) ? true : false;
				}
			} else { // b) everything else
				return (a.x < b.x) ? true : false;
			}
		};

	
	std::sort(data.begin(), data.end(), sortfunc);
}



LSQResult::LSQResult()
{
	sumsqr = 0.0;
	sigma2 = 0.0;	
	lsqExitCode = 0;
}

/**
 * \brief  Loads covariance matrix and sets coefficients uncertainties
 * (standard deviations and confidence intervals)
 *
 * Requires data array, i.e experimental points, to be loaded
 * (to estimate t distribution quantile)
 */
bool LSQResult::SetCovar(const std::vector <double> &covar)
{
	// Check matrix size and copy it
	size_t nparams = param.NumParams();
	if (covar.size() != nparams * nparams) {
		return false;
	}
	this->covar = covar;
	// Estimate coefficients standard deviations
	double tinv = tdistr_inv(0.975, numOfPoints - nparams); // For 95% conf int (t-distr)
	std::vector <double> sparam_vec(nparams), dparam_vec(nparams);
	for (size_t i = 0; i < nparams; i++) {
		size_t ind = i * nparams + i; // Main diagonal
		double sp = sqrt(this->covar[ind]);
		sparam_vec[i] = sp;
		dparam_vec[i] = sp * tinv;
	}
	sparam = param; sparam.Deserialize(sparam_vec);
	dparam = param; dparam.Deserialize(dparam_vec);
	
	return true;
}


bool LSQResult::isParamPositive()
{
	return param.IsAllAlphaThetaPositive();
}


/**
 * \brief  Return parameters standard deviations
 */
TDBaseLineFuncParams LSQResult::GetParamStd()
{
	return sparam;
}

/**
 * \brief  Returns parameters confidence intervals (95%, t-distr)
 */
TDBaseLineFuncParams LSQResult::GetParamConfInt()
{
	return dparam;
}



/**
 * \brief  Loads model parameter and resets all deviations
 */
void LSQResult::SetParam(const TDBaseLineFuncParams &param)
{
	sumsqr = 0.0;
	sigma2 = 0.0;
	this->param = param;
	this->sparam = param; 
	this->dparam = param;
	// Fill coefficients approximations with zeros
	this->sparam.SetZeros();
	this->dparam.SetZeros();
	// Fill standard deviations with zeros
	size_t nparams = sparam.NumParams();
	this->covar = std::vector <double>(nparams*nparams);
	for (size_t i = 0; i < nparams * nparams; i++) {
		this->covar[i] = 0.0;
	}
}

void LSQResult::AddTerm(double alpha, double theta)
{
	param.AddTerm(alpha, theta);
	SetParam(param);
}

/**
 * \brief  Calculate and load deviations from LSQContext to LSQResult
 */
void LSQResult::LoadDeviations(LSQContext &lsqCont)
{
	
	// Estimate absolute, relative and maximal deviations for all variables
	devCp = GetYDeviations(lsqCont, DATATYPE_CP);
	devH  = GetYDeviations(lsqCont, DATATYPE_H);
	devS  = GetYDeviations(lsqCont, DATATYPE_S);
	// Clear possible old data series
	devs.clear();
	devsBaseLine.clear();
	devsAnomaly.clear();
	// Estimate absolute, relative and maximal deviations for all data series
	std::vector <std::string> serIDList = lsqCont.GetSeriesIDList();
	for (size_t i = 0; i < serIDList.size(); i++) {
		std::string serID = serIDList[i];
		// a) all points
		LSQDevInfo devInfoCp = GetYDeviations(lsqCont, DATATYPE_CP, serID);
		LSQDevInfo devInfoH = GetYDeviations(lsqCont, DATATYPE_H, serID);
		LSQDevInfo devInfoS = GetYDeviations(lsqCont, DATATYPE_S, serID);
		// b) points inside anomaly
		LSQDevInfo devInfoCpBL = GetYDeviations(lsqCont, DATATYPE_CP, serID, POINTSFILTER_BASELINE);
		LSQDevInfo devInfoHBL = GetYDeviations(lsqCont, DATATYPE_H, serID, POINTSFILTER_BASELINE);
		LSQDevInfo devInfoSBL = GetYDeviations(lsqCont, DATATYPE_S, serID, POINTSFILTER_BASELINE);
		// c) points outside anomaly
		LSQDevInfo devInfoCpAn = GetYDeviations(lsqCont, DATATYPE_CP, serID, POINTSFILTER_ANOMALY);
		LSQDevInfo devInfoHAn = GetYDeviations(lsqCont, DATATYPE_H, serID, POINTSFILTER_ANOMALY);
		LSQDevInfo devInfoSAn = GetYDeviations(lsqCont, DATATYPE_S, serID, POINTSFILTER_ANOMALY);
		// d) Add all deviations
		if (devInfoCp.npoints > 0) {
			devs.push_back(devInfoCp);
			devsBaseLine.push_back(devInfoCpBL);
			devsAnomaly.push_back(devInfoCpAn);
		}
		if (devInfoH.npoints > 0) {
			devs.push_back(devInfoH);
			devsBaseLine.push_back(devInfoHBL);
			devsAnomaly.push_back(devInfoHAn);
		}
		if (devInfoS.npoints > 0) {
			devs.push_back(devInfoS);
			devsBaseLine.push_back(devInfoSBL);
			devsAnomaly.push_back(devInfoSAn);
		}
	}
}

/**
 * \brief  Loads model parameters and tries to restore confidence intervals
 *
 * If lsqCont contains some points deviations, covariance matrix and
 * confidence intervals will be calculated
 */
void LSQResult::SetParam(const TDBaseLineFuncParams &param, LSQContext &lsqCont)
{
	// Set parameter (with initialization of some structures)
	SetParam(param);
	size_t nparams = this->param.NumParams();
	lsqCont.InitBaseLineData();
	TDPEFunc baseLineFunc(param);
	// Estimate absolute, relative and maximal deviations for all variables
	LoadDeviations(lsqCont);
	// Estimate Jacobian
	size_t npoints = lsqCont.GetNumOfBaseLinePoints();
	numOfPoints = npoints;
	if (npoints > 0) {
		// a) Sum of squares
		std::vector <double> res = lsqCont.GetResVector(baseLineFunc);
		for (size_t i = 0; i < npoints; i++) {
			sumsqr += res[i] * res[i];
		}		
		sigma2 = sumsqr / (npoints - nparams);
		// b) Jacobian and covariance matrix
		std::vector <double> Jvec = lsqCont.GetJVector(baseLineFunc);
		std::vector <double> JTJ(nparams * nparams);
		for (size_t i = 0; i < nparams; i++) {
			for (size_t j = 0; j < nparams; j++) {
				double *Jelem = &JTJ[i*nparams + j];
				*Jelem = 0;
				for (size_t k = 0; k < npoints; k++) {
					*Jelem += Jvec[k*nparams + i] * Jvec[k*nparams + j];
				}
			}
		}
		dlevmar_covar(&JTJ[0], &covar[0], sumsqr, nparams, npoints);
		SetCovar(covar);
	}
}


/**
 * \brief  Calculates standard deviations for chosen function type (Cp(T), S(T), H(T))
 * and for the given T value. This standard deviation can be used to calculate
 * a confidence interval of the function.
 *
 * The next formula is used to calculate standard deviation of the function:
 *
 * <i>s</i>(<i>Y</i>) = &sigma;                                                                                                                                                                                                                                                                                                                                                                                      (<i>J</i><sub>x</sub>'(<i>J</i>'<i>J</i>)<sup>-1</sup> <i>J</i><sub>x</sub> )<sup>0.5</sup>
 *                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                
 * where <i>J</i><sub>x</sub> is column vector and &sigma;(<i>J</i>'<i>J</i>)<sup>-1</sup> is covariance matrix
 */
double LSQResult::FuncConfIntStd(double T, TDDataType ftype, bool useConfIntervals)
{
	// Formula without HTML formatting
	// (s(Y) = sqrt(Jx' * sigma^2*inv(J'*J) * Jx)

	int m = param.NumParams();
	int n = GetNumOfPoints();
	double *C = &covar[0];
	std::vector <double> Jx(m);
	
	switch (ftype) {
	case DATATYPE_CP:
		dCpdBFunc(T, &Jx[0]);
		break;
	case DATATYPE_S:
		dSdBFunc(T, &Jx[0]);
		break;
	case DATATYPE_H:
		dHdBFunc(T, &Jx[0]);
		break;
	case DATATYPE_UNKNOWN: // TODO: SOME ERROR-LIKE BEHAVIOUR
		break;
	}

	// A = Jx' * sigma^2*inv(J'*J) product	
	std::vector <double> A(m);
	for (int i = 0; i < m; i++) {
		A[i] = 0.0;
		for (int j = 0; j < m; j++) {
			A[i] += Jx[j] * C[i*m + j]; // covar is symmetric
		}
	}
	// T * Jx product
	double ci_std = 0.0;
	for (int i = 0; i < m; i++) {
		ci_std += A[i] * Jx[i];
	}
	ci_std = sqrt(ci_std);
	if (useConfIntervals) {
		ci_std *= tdistr_inv(0.975, n - m);
	}
	// Part 2: extra terms conf.int.
	double ci_std_extra = 0.0;
	switch (ftype) {
	case DATATYPE_CP:
		for (unsigned int j = 0; j < extraTerms.size(); j++) {
			ExtraTerm *et = extraTerms[j].term;
			if (et->Tmin <= T && T <= et->Tmax) {
				double s = et->FuncConfIntStd(T, ftype, useConfIntervals);
				ci_std_extra += s*s;
			}
		}
		break;
	case DATATYPE_H:
	case DATATYPE_S:
		for (unsigned int j = 0; j < extraTerms.size(); j++) {
			ExtraTerm *et = extraTerms[j].term;
			if (et->Tmin <= T) {
				double s = et->FuncConfIntStd((T > et->Tmax) ? et->Tmax : T, ftype, useConfIntervals);
				ci_std_extra += s*s;
			}
		}
		break;
	case DATATYPE_UNKNOWN: // TODO: SOME ERROR-LIKE BEHAVIOUR
		break;
	}
	ci_std_extra = sqrt(ci_std_extra);

	return sqrt(ci_std*ci_std + ci_std_extra*ci_std_extra);
}

/**
 * \brief Returns precalculated deviations for a requested subset of
 * experimental data (Cp(T), S(T)-S(0), H(T)-H(0)
 * 
 * \param ftype  Type of experimental data
 */
LSQDevInfo LSQResult::GetYDeviations(TDDataType ftype)
{
	LSQDevInfo devEmpty;
	switch (ftype) {
	case DATATYPE_CP:
		return devCp;
		break;
	case DATATYPE_H:
		return devH;
		break;
	case DATATYPE_S:
		return devS;
		break;
	case DATATYPE_UNKNOWN:
		return devEmpty;
		break;
	}
	return LSQDevInfo();
}

/**
 * \brief Returns precalculated deviations for all data series
 */
std::vector <LSQDevInfo> LSQResult::GetYDeviations(PointsFilterID mode)
{
	switch (mode) {
	case POINTSFILTER_ALL: return devs;
	case POINTSFILTER_BASELINE: return devsBaseLine;
	case POINTSFILTER_ANOMALY: return devsAnomaly;
	default: return devs;
	}
}

/**
 * \brief Calculate deviations for Y value given by user in LSQContext class example
 * described by the current parameters set (inside current LSQResult class example).
 * For internal use only.
 */
LSQDevInfo LSQResult::GetYDeviations(LSQContext &lsqCont, TDDataType ftype)
{
	// Get all required data
	std::vector <double> T = lsqCont.GetXVector(ftype);
	std::vector <double> Y = lsqCont.GetYVector(ftype);
	std::vector <double> T1 = lsqCont.GetX1Vector(ftype);
	std::vector <double> T2 = lsqCont.GetX2Vector(ftype);
	
	if (T.size() == 0 || Y.size() == 0) {
		return LSQDevInfo();
	}
	std::vector <double> Ycalc = (ftype != DATATYPE_H) ? YFunc(ftype, T) : DHFunc(T1, T2);
	// Calculate deviations
	LSQDevInfo devs(Ycalc, Y, (ftype != DATATYPE_H) ? T : T2, 0); // We don't use parameters size (n^(-1) in std formula!)
	return devs;
}

/**
 * \brief Calculate deviations for Y value given by user in LSQContext class example
 * described by the current parameters set (inside current LSQResult class example).
 * For internal use only.
 */
LSQDevInfo LSQResult::GetYDeviations(LSQContext &lsqCont, TDDataType ftype, std::string seriesID, PointsFilterID mode)
{
	// Get all data points
	std::vector <Point> data;
	for (size_t i = 0; i < lsqCont.data.size(); i++) {
		Point &p = lsqCont.data[i];
		if (p.getSeriesID() == seriesID && p.dataType == ftype)
		{
			if (mode == POINTSFILTER_ALL) {
				data.push_back(p);
			} else {
				bool isOnBaseLine = lsqCont.IsPointOnBaseLine(p);
				if ( (mode == POINTSFILTER_BASELINE && isOnBaseLine) ||
				     (mode == POINTSFILTER_ANOMALY && !isOnBaseLine)
				   ) {
					data.push_back(p);
				}
			}
		}
	}
	if (data.size() == 0) {
		LSQDevInfo devs;
		devs.dataType = ftype;
		devs.seriesID = seriesID;
		return devs;
	}
	// Vectors
	std::vector <double> T;
	std::vector <double> Y;
	std::vector <double> T1;
	std::vector <double> T2;
	for (size_t i = 0; i < data.size(); i++) {
		T.push_back(data[i].x);
		T1.push_back(data[i].x1);
		T2.push_back(data[i].x2);
		Y.push_back(data[i].y);	
	}
	// Calculate deviations
	std::vector <double> Ycalc = (ftype != DATATYPE_H) ? YFunc(ftype, T) : DHFunc(T1, T2);
	LSQDevInfo devs(Ycalc, Y, (ftype != DATATYPE_H) ? T : T2, 0); // We don't use parameters size (n^(-1) in std formula!)
	devs.dataType = ftype;
	devs.seriesID = seriesID;
	return devs;
}


int LSQResult::GetExitCode()
{
	return lsqExitCode;
}

bool LSQResult::SetExitCode(int exitCode)
{
	if (0 <= exitCode && exitCode <= 7) {
		lsqExitCode = exitCode;
		return true;
	} else {
		return false;
	}
}


std::string LSQResult::GetExitMessage()
{
	return std::string(lsqExitMessages[lsqExitCode]);
}

bool LSQResult::IsExitCodeGood()
{
	if (lsqExitCode == 0 || lsqExitCode == 1 ||
	    lsqExitCode == 2 || lsqExitCode == 6) {
		return true;
	} else {
		return false;
	}
}

/**
 * \brief  Check if all parameters are statistically significant
 * (t-distribution based criteria)
 */
bool LSQResult::AreAllParamsSignificant()
{
	return param.IsAllParamsSignificant(dparam);
}

int LSQResult::GetNumOfPoints()
{
	return numOfPoints;
}



const char *LSQResult::lsqExitMessages[8] = {
	"", // Code 0
	"Stopped by small gradient J^T e.", // Code 1
	"Stopped by small Dp.", // Code 2
	"Maximum of number of iterations exceeded. Try another initial approximation or decrease the number of parameters.", // Code 3
	"Singular matrix. Restart from current p with increased mu.", // Code 4
	"No further error reduction is possible. Restart with increased mu.", // Code 5
	"Stopped by small ||e||_2.", // Code 6
	"Stopped by invalid (i.e. NaN or Inf) \"func\" values. This is a user error. Try another initial approximation or decrease the number of parameters." // Code 7
};


/**
 * \brief Initializes the class example by invalid values (to show to a user
 * that it requires filling with valid data)
 */
LSQDevInfo::LSQDevInfo()
{
	maxAbsDev = -1.0;
	maxRelDev = -1.0;
	maxAbsDevX = -1.0;
	maxRelDevX = -1.0;
	stdAbsDev = -1.0;
	stdRelDev = -1.0;
	medAbsDev = -1.0;
	medRelDev = -1.0;
	npoints = 0;
}

/**
 * \brief Initializes the class example using data series
 * \param Ycalc  Calculated values  of dependent variable
 * \param Yobs   Observed values of dependend variable
 * \param X      Value of independent variable
 * \param nparams Number of model parameters (required to calculate degrees of freedom
 * as npoints - nparam where npoints is the number of points
 */
LSQDevInfo::LSQDevInfo(std::vector <double> &Ycalc, std::vector <double> Yobs, std::vector <double> &X, int nparams)
{
	LSQDevInfo::Initialize(Ycalc, Yobs, X, nparams);
}

bool LSQDevInfo::Initialize(std::vector <double> &Ycalc, std::vector <double> Yobs, std::vector <double> &X, int nparams)
{
	// Calculate deviations
	unsigned int npoints = Ycalc.size();
	if (npoints != Yobs.size() || npoints != X.size()) {
		return false;
	}
	std::vector <double> absDevVec(npoints), relDevVec(npoints);
	double sumSqrAbs = 0.0, sumSqrRel = 0.0;
	maxAbsDev = 0.0;
	maxRelDev = 0.0;
	for (unsigned int i = 0; i < npoints; i++) {
		double absDev = fabs(Ycalc[i] - Yobs[i]); absDevVec[i] = absDev;
		double relDev = fabs(absDev / Yobs[i]);   relDevVec[i] = relDev;
		sumSqrAbs += absDev * absDev;
		sumSqrRel += relDev * relDev;
		if (absDev > maxAbsDev) {
			maxAbsDev = absDev;
			maxAbsDevX = X[i];
		}
		if (relDev > maxRelDev) {
			maxRelDev = relDev;
			maxRelDevX = X[i];
		}
	}
	stdAbsDev = sqrt(sumSqrAbs / (npoints - nparams));
	stdRelDev = sqrt(sumSqrRel / (npoints - nparams));
	this->npoints = npoints;
	this->nparams = nparams;

	// Calculate median
	std::sort(absDevVec.begin(), absDevVec.end());
	std::sort(relDevVec.begin(), relDevVec.end());
	
	if (npoints % 2 == 0) { // O O ..MED.. O O 
		medAbsDev = (absDevVec[npoints / 2 - 1] + absDevVec[npoints / 2]) / 2;
		medRelDev = (relDevVec[npoints / 2 - 1] + relDevVec[npoints / 2]) / 2;
	} else { // O O O[MED] O O
		medAbsDev = absDevVec[npoints / 2] * MAD_TO_STD;
		medRelDev = relDevVec[npoints / 2] * MAD_TO_STD;
	}
	// Success	
	return true;
}

bool LSQDevInfo::IsInitialized()
{
	if (maxAbsDev == -1.0 && stdAbsDev == -1.0) {
		return false;
	} else {
		return true;
	}
}


////////////////
ExtraTerm::ExtraTerm()
{
	sigma2 = 0.0;
	sumsqr = 0.0;
}

ExtraTerm::~ExtraTerm()
{
}

double ExtraTerm::HFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	auto ff = [this, &p, &c, &Tmin, &Tmax] (double Ti) -> double { return CpFunc(Ti, p, c, Tmin, Tmax);};
	if (T < Tmin) {
		return 0.0;		
	} else {
		double T1 = (Tmin - 50 > 1e-3) ? (Tmin - 50) : 1e-3; 
		double T2 = (T > Tmax + 50) ? (Tmax + 50) : T;
		return int_adaptive_simpson(ff, T1, T2);
	}
}

double ExtraTerm::SFunc(double T, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	auto ff = [this, &p, &c, &Tmin, &Tmax] (double Ti) -> double { return CpFunc(Ti, p, c, Tmin, Tmax) / Ti;};
	if (T < Tmin) {
		return 0.0;
	} else {
		double T1 = (Tmin - 50 > 1e-3) ? (Tmin - 50) : 1e-3; 
		double T2 = (T > Tmax + 50) ? (Tmax + 50) : T;
		return int_adaptive_simpson(ff, T1, T2);
	}
}


double ExtraTerm::dCpdBFunc(double T, double *J, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	for (unsigned int i = 0; i < p.size(); i++) {
		double dB = 1e-3;
		std::vector <double> pB = p; pB[i] = pB[i] + dB;
		std::vector <double> pA = p; pA[i] = pA[i] - dB;
		J[i] = (CpFunc(T, pB, c, Tmin, Tmax) - CpFunc(T, pA, c, Tmin, Tmax)) / (2*dB);
	}
	return 0.0;
}

double ExtraTerm::dHdBFunc(double T, double *J, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	for (unsigned int i = 0; i < p.size(); i++) {
		double dB = 1e-3;
		std::vector <double> pB = p; pB[i] = pB[i] + dB;
		std::vector <double> pA = p; pA[i] = pA[i] - dB;
		J[i] = (HFunc(T, pB, c, Tmin, Tmax) - HFunc(T, pA, c, Tmin, Tmax)) / (2*dB);
	}
	return 0.0;
}

double ExtraTerm::dSdBFunc(double T, double *J, std::vector <double> &p, std::vector <double> &c, double Tmin, double Tmax)
{
	for (unsigned int i = 0; i < p.size(); i++) {
		double dB = 1e-3;
		std::vector <double> pB = p; pB[i] = pB[i] + dB;
		std::vector <double> pA = p; pA[i] = pA[i] - dB;
		J[i] = (SFunc(T, pB, c, Tmin, Tmax) - SFunc(T, pA, c, Tmin, Tmax)) / (2*dB);
	}
	return 0.0;
}


/** 
 * \brief Initialize vectors required for storage of LSQ parameters,
 * their standard deviations and covariance matrix (required for confidence
 * intervals).
 */
void ExtraTerm::InitParamStd()
{
	size_t nparams = lsqparam.size();
	if (slsqparam.size() != nparams ||
	    dlsqparam.size() != nparams)
	{
		slsqparam.clear();
		dlsqparam.clear();
		covar.clear();
		for (size_t i = 0; i < nparams; i++) {
			slsqparam.push_back(0.0);
			dlsqparam.push_back(0.0);
		}
		covar = std::vector <double> (nparams * nparams);
	}
}


bool ExtraTerm::SetLSQParam(std::vector <double> &p)
{
	if (p.size() == lsqparam.size()) {
		lsqparam = p;
		return true;
	} else {
		return false;
	}
}

bool ExtraTerm::SetLSQParamLB(std::vector <double> &p_lb)
{
	if (p_lb.size() == lsqparam_lb.size()) {
		lsqparam_lb = p_lb;
		return true;
	} else {
		return false;
	}
}


bool ExtraTerm::SetLSQParamUB(std::vector <double> &p_ub)
{
	if (p_ub.size() == lsqparam_ub.size()) {
		lsqparam_ub = p_ub;
		return true;
	} else {
		return false;
	}
}


bool ExtraTerm::SetLSQParam(std::vector <double> &p, std::vector <Point> &data, TDPEFunc &baseLine)
{
	// Set LSQ parameter
	if (!SetLSQParam(p)) {
		return false;
	}
	// And load experimental data
	if (!DataLoad(data, baseLine)) {
		return false;
	} else {
		int npoints = etdata.size();
		int nparams = lsqparam.size();
		residuals.resize(npoints, 0.0);
		InitResVector(&residuals[0]); // Emulate DataFit run

		sigma2 = 0.0;
		sumsqr = 0.0;		
		if (npoints > 0) {
			// a) Sum of squares
			for (int i = 0; i < npoints; i++) {
				sumsqr += residuals[i] * residuals[i];
			}		
			sigma2 = sumsqr / (npoints - nparams);
			// b) Jacobian and covariance matrix
			std::vector <double> Jvec = GetJVector();
			std::vector <double> JTJ(nparams * nparams);
			for (int i = 0; i < nparams; i++) {
				for (int j = 0; j < nparams; j++) {
					double *Jelem = &JTJ[i*nparams + j];
					*Jelem = 0;
					for (int k = 0; k < npoints; k++) {
						*Jelem += Jvec[k*nparams + i] * Jvec[k*nparams + j];
					}
				}
			}
			covar = std::vector <double> (nparams * nparams);
			dlevmar_covar(&JTJ[0], &covar[0], sumsqr, nparams, npoints);
			SetCovar(covar);
		}
		return true;
	}
}

std::vector <double> ExtraTerm::GetLSQParam()
{
	return lsqparam;
}

std::vector <std::string> ExtraTerm::GetLSQParamNames()
{
	return lsqparam_names;
}

std::vector <double> ExtraTerm::GetFixParam()
{
	return fixparam;
}

std::vector <std::string> ExtraTerm::GetFixParamNames()
{
	return fixparam_names;
}

std::vector <double> ExtraTerm::GetLSQParamStd()
{
	InitParamStd();
	return slsqparam;
}

std::vector <double> ExtraTerm::GetLSQParamCi()
{
	InitParamStd();
	return dlsqparam;
}

std::vector <double> ExtraTerm::GetLSQParamLB()
{
	return lsqparam_lb;
}

std::vector <double> ExtraTerm::GetLSQParamUB()
{
	return lsqparam_ub;
}


bool ExtraTerm::SetFixParam(std::vector <double> &c)
{
	if (c.size() == fixparam.size()) {
		fixparam = c;
		return true;
	} else {
		return false;
	}
}


/**
 * \brief  Loads experimental data with its selection and subtraction
 * of the baseline
 */
bool ExtraTerm::DataLoad(std::vector <Point> &data, TDPEFunc &baseLine)
{
	etdata.clear();
	// Preselect experimental data (with baseline subtraction)
	for (unsigned int i = 0; i < data.size(); i++) {
		Point p = data[i];
		if (p.dataType == DATATYPE_CP && Tmin <= p.x && p.x <= Tmax) {
			p.y = p.y - baseLine.CpFunc(p.x);
			etdata.push_back(p);
		}
	}
	// Check data size
	if (etdata.size() <= lsqparam.size()) {
		return false; // Not enough experimental data (or too many model parameters)
	} else {
		return true; // Successful
	}
}

void ExtraTerm::InitResVector(double *res)
{
	// Calculate residuals
	int n = etdata.size();
	for (int i = 0; i < n; i++) {
		// a) calculated value (for Cp only!)
		double T = etdata[i].x;
		double ycalc = CpFunc(T, lsqparam, fixparam, Tmin, Tmax);
		double yexp = etdata[i].y;
		// b) weight
		double w = etdata[i].w;
		if (etdata[i].isRelDev) {
			w /= yexp;
			}
		// c) calculate residual
		res[i] = (ycalc - yexp) * w;			
	}
}


/**
 * \brief Loads experimental data and fits the model parameters
 */
bool ExtraTerm::DataFit(std::vector <Point> &data, TDPEFunc &baseLine)
{
	if (!DataLoad(data, baseLine)) {
		return false;
	}
	// lambda for optimization procedure
	auto minfunc = [] (double *param, double *res, int m, int n, void *adata) -> void
	{
//		static int nfcall = 0;
//		FILE *fp = fopen("lualog.log", "a");
//		fprintf(fp, "FCALL %d\n", nfcall++);
//		fclose(fp);
		ExtraTerm *term = (ExtraTerm *) adata;
		// Create lsqparam
		std::vector <double> lsqparam(m);
		for (int i = 0; i < m; i++) {
			lsqparam[i] = param[i];
		}
		// Calculate residuals
		term->lsqparam = lsqparam;
		term->InitResVector(res);
	};
	// Optimization procedure
	int nparams = lsqparam.size();
	int npoints = etdata.size();
	std::vector <double> lsqparam_out = lsqparam;
	covar = std::vector <double> (nparams * nparams);

	double opts[LM_OPTS_SZ] = {LM_INIT_MU, 1e-15, 1e-15, 1e-15, LM_DIFF_DELTA};
	double info[LM_INFO_SZ];

	dlevmar_bc_dif(
		minfunc, &lsqparam_out[0], NULL, nparams, npoints,
		&lsqparam_lb[0], &lsqparam_ub[0], NULL, // Constraints (box)
		250, // Itmax
		opts, info, NULL, &covar[0],
		(void *) this);

	lsqparam = lsqparam_out;
	// Calculate residuals
	residuals = std::vector<double>(npoints);
	minfunc(&lsqparam[0], &residuals[0], nparams, npoints, (void *) this);
	sigma2 = 0.0; sumsqr = 0.0;		
	if (npoints > 0) {
		// a) Sum of squares
		for (int i = 0; i < npoints; i++) {
			sumsqr += residuals[i] * residuals[i];
		}		
		sigma2 = sumsqr / (npoints - nparams);
	}
	// Calculate coefficients confidence intervals
	SetCovar(covar);
	return true;
}

std::string ExtraTerm::GetType()
{
	return type;
}

std::string ExtraTerm::GetBriefDescr()
{
	return briefDescr;
}

std::string ExtraTerm::GetFullDescr()
{
	return fullDescr;
}

std::string ExtraTerm::GetFormula()
{
	return formula;
}



int ExtraTerm::GetNumOfPoints()
{
	return etdata.size();
}

std::vector <double> ExtraTerm::GetResVector()
{
	return residuals;
}

std::vector <double> ExtraTerm::GetTVector()
{
	std::vector <double> Tvec(etdata.size());
	for (unsigned int i = 0; i < etdata.size(); i++) {
		Tvec[i] = etdata[i].x;
	}
	return Tvec;
}

std::vector <double> ExtraTerm::GetYVector()
{
	std::vector <double> Yvec(etdata.size());
	for (unsigned int i = 0; i < etdata.size(); i++) {
		Yvec[i] = etdata[i].y;
	}
	return Yvec;
}

std::vector <bool> ExtraTerm::GetIsRelDevVector()
{
	std::vector <bool> isRelDev(etdata.size());
	for (unsigned int i = 0; i < etdata.size(); i++) {
		isRelDev[i] = etdata[i].isRelDev;
	}
	return isRelDev;
}


double ExtraTerm::CpFuncEval(double T)
{
	return CpFunc(T, lsqparam, fixparam, Tmin, Tmax);
}

double ExtraTerm::HFuncEval(double T)
{
	return HFunc(T, lsqparam, fixparam, Tmin, Tmax);
}

double ExtraTerm::SFuncEval(double T)
{
	return SFunc(T, lsqparam, fixparam, Tmin, Tmax);
}

void ExtraTerm::dCpdBFuncEval(double T, double *J)
{
	dCpdBFunc(T, J, lsqparam, fixparam, Tmin, Tmax);
}

void ExtraTerm::dHdBFuncEval(double T, double *J)
{
	dHdBFunc(T, J, lsqparam, fixparam, Tmin, Tmax);
}

void ExtraTerm::dSdBFuncEval(double T, double *J)
{
	dSdBFunc(T, J, lsqparam, fixparam, Tmin, Tmax);
}

/**
 * \brief Returns Jacobian (matrix) in the form of vector using
 * the current data set and values of lsqparam and fixparam.
 *
 * It uses the next order of elements:
 *
 * [&part;Y<sub>0</sub>/&part;&beta;, &part;Y<sub>1</sub>/&part;&beta;, ...]
 */
std::vector <double> ExtraTerm::GetJVector()
{
	std::vector <Point> &dat = etdata;
	int npoints = dat.size();
	int nparams = lsqparam.size();
	std::vector <double> J(npoints * nparams);
	double *Jptr = &J[0];
	double der2[nparams]; 
	double der1[nparams];
		
	for (int i = 0; i < npoints; i++) {
		// a) Calculated value
		switch (dat[i].dataType) {
		case DATATYPE_CP:
			dCpdBFuncEval(dat[i].x, Jptr);
			break;
		case DATATYPE_S:
			dSdBFuncEval(dat[i].x, Jptr);
			break;
		case DATATYPE_H:
			dHdBFuncEval(dat[i].x2, der2);
			dHdBFuncEval(dat[i].x1, der1);
			for (int j = 0; j < nparams; j++) {
				Jptr[j] = der2[j] - der1[j];
			}
			break;
		case DATATYPE_UNKNOWN:
			for (int j = 0; j < nparams; j++) {
				Jptr[j] = NAN;
			}
			break;
		}
		// b) weight
		double w = dat[i].w;
		if (dat[i].isRelDev) {
			w /= dat[i].y;
		}
		for (int j = 0; j < nparams; j++) {
			Jptr[j] *= w;
		}
		// Go to the next line
		Jptr += nparams;
	}
	return J;
}


/**
 * \brief Calculates coefficients confidence intervals using existing covariance
 * matrix. 
 * \details Formula without HTML formatting:
 *     (s(Y) = sqrt(Jx' * sigma^2*inv(J'*J) * Jx)
 */
double ExtraTerm::FuncConfIntStd(double T, TDDataType ftype, bool useConfIntervals)
{
	int m = lsqparam.size();
	int n = etdata.size();
	double *C = &covar[0];
	std::vector <double> Jx(m);
	
	switch (ftype) {
	case DATATYPE_CP:
		dCpdBFuncEval(T, &Jx[0]);
		break;
	case DATATYPE_S:
		dSdBFuncEval(T, &Jx[0]);
		break;
	case DATATYPE_H:
		dHdBFuncEval(T, &Jx[0]);
		break;
	case DATATYPE_UNKNOWN:
		for (int i = 0; i < m; i++) {
			Jx[i] = NAN;
		}
		break;
	}

	// A = Jx' * sigma^2*inv(J'*J) product	
	std::vector <double> A(m);
	for (int i = 0; i < m; i++) {
		A[i] = 0.0;
		for (int j = 0; j < m; j++) {
			A[i] += Jx[j] * C[i*m + j]; // covar is symmetric
		}
	}
	// T * Jx product
	double ci_std = 0.0;
	for (int i = 0; i < m; i++) {
		ci_std += A[i] * Jx[i];
	}
	ci_std = sqrt(ci_std);
	if (useConfIntervals) {
		ci_std *= tdistr_inv(0.975, n - m);
	}
	return ci_std;
}


/**
 * \brief  Loads covariance matrix and sets coefficients uncertainties
 * (standard deviations and confidence intervals)
 *
 * Requires data array, i.e experimental points, to be loaded
 * (to estimate t distribution quantile)
 */
bool ExtraTerm::SetCovar(const std::vector <double> &covar)
{
	// Check matrix size and copy it
	unsigned int nparams = lsqparam.size();
	unsigned int npoints = etdata.size();
	if (covar.size() != nparams * nparams) {
		return false;
	}
	this->covar = covar;
	// Estimate coefficients standard deviations
	double tinv = tdistr_inv(0.975, npoints - nparams); // For 95% conf int (t-distr)
	slsqparam.clear();
	dlsqparam.clear();
	for (unsigned int i = 0; i < nparams; i++) {
		unsigned int ind = i * nparams + i; // Main diagonal
		double sp = sqrt(this->covar[ind]);
		slsqparam.push_back(sp);
		dlsqparam.push_back(sp * tinv);
	}
	return true;
}

void ExtraTerm::CopyData(ExtraTerm *term)
{
	Tmin = term->Tmin;
	Tmax = term->Tmax;
	etdata = term->etdata;
	sumsqr = term->sumsqr;
	sigma2 = term->sigma2;

	lsqparam = term->lsqparam;
	slsqparam = term->slsqparam;
	dlsqparam = term->dlsqparam;
	lsqparam_lb = term->lsqparam_lb;
	lsqparam_ub = term->lsqparam_ub;
	fixparam = term->fixparam;
	residuals = term->residuals;
	covar = term->covar;

	fixparam_names = term->fixparam_names;
	lsqparam_names = term->lsqparam_names;

	briefDescr = term->briefDescr;
	fullDescr = term->fullDescr;
	formula = term->formula;
}

std::string ExtraTerm::GetErrorMessage()
{
	return errorMessage;
}


////////////////////////////////////////////////////////////////
std::map <std::string, ExtraTerm *> ExtraTermContainer::termsTypes;
std::vector <std::string> ExtraTermContainer::termsTypesNames;

bool ExtraTermContainer::AddExtraTermType(ExtraTerm * term)
{
	if (term == NULL) {
		return false;
	}
	// Check if such type is already present
	std::string type = term->GetType();
	if (termsTypes.count(type) > 0) {
		return false;
	}
	// Type not present -- add it
	termsTypesNames.push_back(type);
	termsTypes[type] = term->Clone();
	delete term;
	return true;
}


void ExtraTermContainer::InitTermsTypes()
{
	if (termsTypesNames.size() > 0) {
		return;
	}
	
	AddExtraTermType(new ExtraTermDefault());
	AddExtraTermType(new ExtraTermTwoPeaks());
	AddExtraTermType(new ExtraTermLeftExp());
	AddExtraTermType(new ExtraTermRightExp());
	AddExtraTermType(new ExtraTermLeftExp2());
	AddExtraTermType(new ExtraTermRightExp2());
	AddExtraTermType(new ExtraTermSkewed());
	AddExtraTermType(new ExtraTermEmpty());
}


std::vector <std::string> ExtraTermContainer::GetTermTypesList()
{
	ExtraTermContainer::InitTermsTypes();
	return termsTypesNames;
}

std::vector <std::string> ExtraTermContainer::GetTermBriefDescrList()
{
	std::vector <std::string> descrList;
	ExtraTermContainer::InitTermsTypes();
	for (size_t i = 0; i < termsTypesNames.size(); i++) {
		ExtraTerm *term = termsTypes[termsTypesNames[i]];
		descrList.push_back(term->GetBriefDescr());
	}
	return descrList;
}

std::vector <std::string> ExtraTermContainer::GetTermFormulaList()
{
	std::vector <std::string> formulaList;
	ExtraTermContainer::InitTermsTypes();
	for (size_t i = 0; i < termsTypesNames.size(); i++) {
		ExtraTerm *term = termsTypes[termsTypesNames[i]];
		formulaList.push_back(term->GetFormula());
	}
	return formulaList;
}


void ExtraTermContainer::ExtraTermInitPtr(std::string termType)
{
	if (term != NULL) {
		delete term;
	}

	if (termsTypes.count(termType) > 0) {
		term = termsTypes[termType]->Clone();
	} else {
		term = NULL;
	}
}


ExtraTermContainer::ExtraTermContainer()
{
	ExtraTermContainer::InitTermsTypes();
	term = NULL;
	ExtraTermInitPtr("DEFAULT");
}


ExtraTermContainer::ExtraTermContainer(std::string termType)
{
	term = NULL;
	ExtraTermInitPtr(termType);
}

ExtraTermContainer::~ExtraTermContainer()
{
	if (term != NULL)
		delete term;
	term = NULL;
}


ExtraTermContainer::ExtraTermContainer(const ExtraTermContainer &c)
{
	if (c.term == NULL) {
		term = NULL;
	} else {
		term = NULL;
		ExtraTermInitPtr(c.term->GetType());
		if (term != NULL) {
			term->CopyData(c.term);
		}
	}
}

ExtraTermContainer & ExtraTermContainer::operator=(const ExtraTermContainer &c)
{
	if (this == &c) {
		return *this;
	}

	if (c.term == NULL) {
		term = NULL;
	} else {
		term = NULL;
		ExtraTermInitPtr(c.term->GetType());
		if (term != NULL) {
			term->CopyData(c.term);
		}
	}
	return *this;
}


bool ExtraTermContainer::IsValid()
{
	return (term == NULL) ? false : true;
}

bool ExtraTermContainer::CompareTo(const ExtraTermContainer &c)
{
	if (term->GetType() != c.term->GetType()) {
		return false;
	} else {
		return true;
	}
}

