/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <wx/wx.h>
#include <wx/grid.h>
#include <wx/bookctrl.h>
#include <wx/textfile.h>
#include <wx/tokenzr.h>
#include "datafit.h"
#include "plotcanvas.h"
#include "cp.h"
#include "graph.h"
#include "tdtable.h"
#include "xyplots.h"
#include "misc.h"
#include "dfunc3.h"



enum {
	ID_MNU_VIRTPOINTS = 2000,
};


/**
 * \brief Window with low-temperature Debye correction term
 */
class DebyeCorrResultWindow : public wxFrame
{
public:
	DebyeCorrResultWindow(wxFrame *parent, TDPEFunc *tdpefunc, MainWindow *mainWnd, double natoms, double T0);
	void OnClose(wxCloseEvent &event);
	void OnAddVirtualPoints(wxCommandEvent &event);
private:
	void EvaluateCorrections();

	double natoms;
	double T0;
	double TD;
	TDPEFunc tdpefunc;
	MainWindow *mainWnd;
	
	wxTextCtrl *txtNAtoms;
	wxTextCtrl *txtT0;
	wxTextCtrl *txtTD;
	wxTextCtrl *txtSE;
	wxTextCtrl *txtSD;
	wxTextCtrl *txtDS;

	PlotCanvas *pl;
	
	
	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(DebyeCorrResultWindow, wxFrame)
	EVT_CLOSE(DebyeCorrResultWindow::OnClose)
	EVT_MENU(ID_MNU_VIRTPOINTS, DebyeCorrResultWindow::OnAddVirtualPoints)
wxEND_EVENT_TABLE()

/**
 * \brief Calculates Cp corrections using Debye model and visualizes it.
 * It also creates all GUI elements and its layout.
 * \details The window has the next layout defined by sizers:
 * 
 *     +-------{mainSizer}----+------+
 *     | +--{infoSizer}-----+ |      |
 *     | | +--{gridSizer}-+ | |      |
 *     | | | name | value | | |      |
 *     | | +------+-------+ | |      |
 *     | | | name | value | | | PLOT |
 *     | | +------+-------+ | |      |
 *     | +------------------+ |      |
 *     | | comment 1        | |      |
 *     | | comment 2        | |      |
 *     | +------------------+ |      |
 *     +----------------------+------+
 */
DebyeCorrResultWindow::DebyeCorrResultWindow(wxFrame *parent, TDPEFunc *tdpefunc, MainWindow *mainWnd, double natoms, double T0) :
	wxFrame(parent, wxID_ANY, "Low-T entropy correction result", wxPoint(50, 50), wxSize(700, 500))
{
	this->tdpefunc = *tdpefunc;
	this->mainWnd = mainWnd;
	this->natoms = natoms;
	this->T0 = T0;
	this->TD = TD;
	
	wxPanel *panel = new wxPanel(this, -1);
	// Labels
	// a) For textboxes
	wxStaticText *lblNAtoms = new wxStaticText(panel, wxID_ANY, "Number of atoms");
	wxStaticText *lblT0 = new wxStaticText(panel, wxID_ANY, "T0 (lowest T), K");
	wxStaticText *lblTD = new wxStaticText(panel, wxID_ANY, "TD (debye T), K");
	wxStaticText *lblSE = new wxStaticText(panel, wxID_ANY, "S (E.-P.), J/(mol*K)");
	wxStaticText *lblSD = new wxStaticText(panel, wxID_ANY, "S (Debye), J/(mol*K)");
	wxStaticText *lblDS = new wxStaticText(panel, wxID_ANY, "SD - SE, J/(mol*K)");
	// b) Graph legend
	wxStaticText *lblInfo1 = new wxStaticText(panel, wxID_ANY, "Black line - E.-P. functions sum");
	wxStaticText *lblInfo2 = new wxStaticText(panel, wxID_ANY, "Blue line - Debye function");
	// TextBoxes for values
	txtNAtoms = new wxTextCtrl(panel, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	txtT0 = new wxTextCtrl(panel, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	txtTD = new wxTextCtrl(panel, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	txtSE = new wxTextCtrl(panel, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	txtSD = new wxTextCtrl(panel, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	txtDS = new wxTextCtrl(panel, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	// Plot canvas with graphs
	pl = new PlotCanvas(panel, -1, wxDefaultPosition, wxDefaultSize);
	pl->SetXLabel(wxT("T / K"));
	pl->SetYLabel(wxT("Cp / J/(mol\u00B7K)")); // \cdot	
	// Window menu
	wxMenu *menuData = new wxMenu();
	menuData->Append(ID_MNU_VIRTPOINTS, "Add virtual poins", "Add virtual points using Debye model");
	wxMenuBar *menuBar = new wxMenuBar();
	menuBar->Append(menuData, "Data");
	SetMenuBar(menuBar);
	// Set window icon
	SetIcon(wxICON(sample));
	// Sizers

	int borderWidth = 5;
	wxGridSizer *gridSizer = new wxGridSizer(2, 5, 5);
	gridSizer->Add(lblNAtoms, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtNAtoms, 0, wxALL | wxEXPAND);
	gridSizer->Add(lblT0, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtT0, 0, wxALL | wxEXPAND);
	gridSizer->Add(lblTD, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtTD, 0, wxALL | wxEXPAND);
	gridSizer->Add(lblSE, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtSE, 0, wxALL | wxEXPAND);
	gridSizer->Add(lblSD, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtSD, 0, wxALL | wxEXPAND);
	gridSizer->Add(lblDS, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtDS, 0, wxALL | wxEXPAND);

	wxBoxSizer *infoSizer = new wxBoxSizer(wxVERTICAL);
	infoSizer->Add(gridSizer, 0, wxALL | wxEXPAND, borderWidth);
	infoSizer->Add(lblInfo1, 0, wxALL | wxEXPAND, borderWidth);
	infoSizer->Add(lblInfo2, 0, wxALL | wxEXPAND, borderWidth);
	infoSizer->AddStretchSpacer();

	wxBoxSizer *mainSizer = new wxBoxSizer(wxHORIZONTAL);
	mainSizer->Add(infoSizer, 0, wxALL | wxEXPAND, borderWidth);
	mainSizer->Add(pl, 1, wxALL | wxEXPAND, borderWidth);

	panel->SetSizer(mainSizer);
	panel->Layout();
	// Set window coordinates and minimal size
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetMinSize(wxSize(700, 500));
	// Load data
	EvaluateCorrections();
}

/**
 * \brief Evaluate Debye function based corrections for heat capacity
 * extrapolation, show them in textual form and in the form of plot.
 */
void DebyeCorrResultWindow::EvaluateCorrections()
{
	double SD, SE, resnorm = 0.0;
	if (natoms > 0) {
		TD = this->tdpefunc.EvalDebyeTInPoint(T0, natoms);
		SD = natoms * 3*CONST_R*S_debye_func(T0 / TD);
		SE = this->tdpefunc.SFunc(T0);
	} else {
		std::vector <double> par = this->tdpefunc.EvalDebyeTNatInPoint(T0);
		TD = par[0];
		natoms = par[1]; this->natoms = natoms;
		resnorm = par[2];
		SD = natoms * 3*CONST_R*S_debye_func(T0 / TD);
		SE = this->tdpefunc.SFunc(T0);
	}
	// Check warning if resnorm is larger than 1e-6
	if (resnorm > 1e-6) {
		wxString msg = wxString::Format("Debye functions are found inaccurate; resnorm = %g\n", resnorm);
		wxMessageBox(msg, wxT("CpFit"), wxICON_WARNING);
	}
	// Load values
	txtNAtoms->ChangeValue(wxString::Format("%.4g", natoms));
	txtT0->ChangeValue(wxString::Format("%.4g", T0));
	txtTD->ChangeValue(wxString::Format("%.4g", TD));
	txtSE->ChangeValue(wxString::Format("%.4g", SE));
	txtSD->ChangeValue(wxString::Format("%.4g", SD));
	txtDS->ChangeValue(wxString::Format("%.4g", SD - SE));
	// Visualize at plot
	std::vector <double> T, CpDv, CpEv;
	for (double Ti = 1e-5; Ti <= T0 + 1e-10; Ti += T0 * 0.001) {
		double CpDi = natoms * 3*CONST_R*c_debye_func(TD / Ti);
		T.push_back(Ti);
		CpDv.push_back(CpDi);

	}
	CpEv = this->tdpefunc.CpFunc(T);
	pl->Plot(T, CpEv, PlotMarker(wxColor(0,0,0), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	pl->Plot(T, CpDv, PlotMarker(wxColor(0,0,255), wxColor(0,0,0), LINE_SOLID, MARKER_NONE));
	pl->FitAxes();
}

void DebyeCorrResultWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}


void DebyeCorrResultWindow::OnAddVirtualPoints(wxCommandEvent &event)
{
	bool isValueOk = false;
	std::vector <Point> virtPoints;
	// Ask if the user wants to enter points or the number of points
	int ans = wxMessageBox("Do you want to enter temperatures manually? If no then number of virtual points will be asked.", "Virtual points addition",
		wxYES_NO | wxNO_DEFAULT | wxICON_EXCLAMATION);
	// Different variants of virtual points addition dialog
	if (ans == wxNO) {
		// Get number of virtual points
		wxTextEntryDialog *dlg = new wxTextEntryDialog(
			this,
			"Enter the number of virtual points",
			"Virtual points addition",
			"5");
		long numOfPoints = 0;
		do {
			dlg->SetValue("5");
			if (dlg->ShowModal() == wxID_CANCEL) {
				return;
			}
			isValueOk = dlg->GetValue().ToLong(&numOfPoints);
			if (numOfPoints < 1 || numOfPoints > 100) {
				isValueOk = false;
			}
			if (!isValueOk) {
				wxMessageBox("Invalid number of virtual points", "CpFit", wxICON_ERROR);
			}
		} while (!isValueOk);
		// Add virtual points
		double dT = T0 / (numOfPoints + 1);
		for (int i = 0; i < numOfPoints; i++) {
			Point p;
			p.x = dT * (i + 1);
			p.y = natoms * 3.0*CONST_R*c_debye_func(TD / p.x);
			virtPoints.push_back(p);
		}
	} else {
		// Get user-given virtual points
		wxTextEntryDialog *dlg = new wxTextEntryDialog(
			this,
			"Enter space-delimited temperatures of virtual points",
			"Virtual points addition",
			"5 10 20 30");
		do {
			if (dlg->ShowModal() == wxID_CANCEL) {
				return;
			}
			wxString strPoints = dlg->GetValue().Trim();
			wxStringTokenizer tokenizer(strPoints, " \t");
			while ( tokenizer.HasMoreTokens() ) {
				wxString token = tokenizer.GetNextToken();
				double Ti;
				if (!token.ToDouble(&Ti) || Ti <= 0.0 || Ti >= 10000.0) {
					wxMessageBox(wxString::Format("Invalid T value %g", Ti), "Virtual points addition", wxICON_ERROR);
					isValueOk = false;
					break;
				} else {
					Point p;
					p.x = Ti;
					p.y = natoms * 3.0*CONST_R*c_debye_func(TD / p.x);
					virtPoints.push_back(p);
				}
			}
			isValueOk = true;
		} while (!isValueOk);
	}
	// Add generated virtual points
	mainWnd->AddDataPoints(virtPoints);
}



//////////////////////////////////////////////////////////
enum {
	ID_BTN_EVALDEBYECORR = 1000,
	ID_CHK_DERIVATIVES
};


class DebyeCorrWindow : public wxFrame
{
public:
	DebyeCorrWindow(wxFrame *parent, TDPEFunc *tdpefunc, MainWindow *mainWnd);
	void OnSize(wxSizeEvent &event);
	void OnClose(wxCloseEvent &event);
	void OnEvalDebyeCorr(wxCommandEvent &event);
	void OnChkDerivatives(wxCommandEvent &event);
private:
	wxTextCtrl *txtNAtoms;
	wxTextCtrl *txtT0;
	wxCheckBox *chkDerivatives;
	MainWindow *mainWnd;
	TDPEFunc tdpefunc;

	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(DebyeCorrWindow, wxFrame)
	EVT_SIZE(DebyeCorrWindow::OnSize)
	EVT_CLOSE(DebyeCorrWindow::OnClose)
	EVT_BUTTON(ID_BTN_EVALDEBYECORR, DebyeCorrWindow::OnEvalDebyeCorr)
	EVT_CHECKBOX(ID_CHK_DERIVATIVES, DebyeCorrWindow::OnChkDerivatives)
wxEND_EVENT_TABLE()


/**
 * @brief  Opens window to evaluate influence of differences between Einstein-Planck and Debye functions
 * on entropy values at low (T < 30 K) temperatures
 */
DebyeCorrWindow::DebyeCorrWindow(wxFrame *parent, TDPEFunc *tdpefunc, MainWindow *mainWnd)
	: wxFrame(parent, wxID_ANY, "Low-T entropy correction", wxPoint(50, 50), wxSize(300, 220))
{
	this->tdpefunc = *tdpefunc;
	this->mainWnd = mainWnd;

	wxPanel *panel = new wxPanel(this, -1);
	wxStaticText *lblNAtoms = new wxStaticText(panel, wxID_ANY, "Number of atoms");
	wxStaticText *lblT0 = new wxStaticText(panel, wxID_ANY, "T0 (lowest T), K");
	
	txtNAtoms = new wxTextCtrl(panel, wxID_ANY, "");
	txtT0 = new wxTextCtrl(panel, wxID_ANY, "");
	chkDerivatives = new wxCheckBox(panel, ID_CHK_DERIVATIVES, "Use derivatives (vary Natoms)", wxPoint(10, 95), wxSize(200, 25));
	chkDerivatives->SetValue(false);
	
	wxButton *btnEvalDebyeCorr = new wxButton(panel, ID_BTN_EVALDEBYECORR, "Evaluate", wxPoint(170,135), wxSize(100,30));
	// Sizers
	wxGridSizer *gridSizer = new wxGridSizer(2, 5, 5);
	gridSizer->Add(lblNAtoms, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtNAtoms, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblT0, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtT0, 1, wxALL | wxEXPAND);

	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(gridSizer, 0, wxALL | wxEXPAND, 10);
	mainSizer->Add(chkDerivatives, 0, wxALL | wxEXPAND, 10);
	mainSizer->AddStretchSpacer();
	mainSizer->Add(btnEvalDebyeCorr, 0, wxALL | wxEXPAND, 10);
	panel->SetSizer(mainSizer);
	panel->Layout();
	// Set window icon
	SetIcon(wxICON(sample));
}


void DebyeCorrWindow::OnSize(wxSizeEvent &event)
{	
	event.Skip();
}


void DebyeCorrWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

void DebyeCorrWindow::OnChkDerivatives(wxCommandEvent &event)
{
	
	txtNAtoms->SetEditable(!chkDerivatives->GetValue());
}

/**
 * @brief Runs the evaluation process
 */
void DebyeCorrWindow::OnEvalDebyeCorr(wxCommandEvent &event)
{
	double natoms, T0;
	// Get values	
	wxString strNAtoms = txtNAtoms->GetValue();
	wxString strT0 = txtT0->GetValue();
	if (!strNAtoms.ToDouble(&natoms) || natoms < 1.0) {
		wxMessageBox("Invalid natoms value", "CpFit", wxICON_ERROR);
		return;
	}
	if (!strT0.ToDouble(&T0) || T0 <= 0.0 || T0 > 500.0) {
		wxMessageBox("Invalid T0 value", "CpFit", wxICON_ERROR);
		return;
	}
	// Find Debye temperature and entropy values
	if (chkDerivatives->GetValue()) {
		natoms = -1;
	}
	DebyeCorrResultWindow *wnd = new DebyeCorrResultWindow(this, &tdpefunc, mainWnd, natoms, T0);
	wnd->Show(true);	
	event.Skip();
}


///////////////////////////

enum {
	ID_BTN_BUILDDHPLOT = 1000,
	ID_TXT_DHPLOT_T0,
	ID_TXT_DHPLOT_DT0
};

/**
 * \brief  Dialog for H(T)-H(T0) and (H(T)-H(T0))/(T - T0) plots
 * parameters customization
 */
class DHPlotDialog : public wxFrame
{
public:
	DHPlotDialog(wxFrame *parent, LSQContext &lsqCont, LSQResult &lsqRes);
	void OnClose(wxCloseEvent &event);
	void OnBuildDHPlot(wxCommandEvent &event);
	void OnTxtT0(wxCommandEvent &event);
	void OnTxtDT0(wxCommandEvent &event);
private:
	wxTextCtrl *txtT0;
	wxTextCtrl *txtDT0;
	wxCheckBox *chkNormed;
	LSQContext lsqCont;
	LSQResult lsqRes;
	double T0, DT0;
	
	wxDECLARE_EVENT_TABLE();
};

wxBEGIN_EVENT_TABLE(DHPlotDialog, wxFrame)
	EVT_CLOSE(DHPlotDialog::OnClose)
	EVT_BUTTON(ID_BTN_BUILDDHPLOT, DHPlotDialog::OnBuildDHPlot)
	EVT_TEXT(ID_TXT_DHPLOT_T0, DHPlotDialog::OnTxtT0)
	EVT_TEXT(ID_TXT_DHPLOT_DT0, DHPlotDialog::OnTxtDT0)
wxEND_EVENT_TABLE()


DHPlotDialog::DHPlotDialog(wxFrame *parent, LSQContext &lsqCont, LSQResult &lsqRes)
	: wxFrame(parent, wxID_ANY, "H(T)-H(T0) plot construction", wxPoint(50, 50), wxSize(300, 220))
{
	this->lsqCont = lsqCont;
	this->lsqRes = lsqRes;
	T0 = 298.15; DT0 = 0.5;

	wxPanel *panel = new wxPanel(this, -1);
	wxStaticText *lblT0 = new wxStaticText(panel, wxID_ANY, "T0, K");
	wxStaticText *lblDT0 = new wxStaticText(panel, wxID_ANY, wxT("\u0394T0, K"));

	txtT0 = new wxTextCtrl(panel, ID_TXT_DHPLOT_T0, wxString::Format("%g", T0));
	txtDT0 = new wxTextCtrl(panel, ID_TXT_DHPLOT_DT0, wxString::Format("%g", DT0));
	chkNormed = new wxCheckBox(panel, wxID_ANY, "Divide by (T - T0)");
	chkNormed->SetValue(true);

	wxButton *btnBuildPlot = new wxButton(panel, ID_BTN_BUILDDHPLOT, "Build plot"/*, wxPoint(170,135), wxSize(100,30)*/);

	// Sizers
	int borderWidth = 5;
	wxGridSizer *gridSizer = new wxGridSizer(2, 2, 5, 5);
	gridSizer->Add(lblT0, 2, wxALL | wxEXPAND, borderWidth); // Line 1
	gridSizer->Add(txtT0, 1, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(lblDT0, 2, wxALL | wxEXPAND, borderWidth); // Line 2
	gridSizer->Add(txtDT0, 1, wxALL | wxEXPAND, borderWidth);

	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(gridSizer, 0, wxALL | wxEXPAND, borderWidth);
	mainSizer->AddStretchSpacer(1);
	mainSizer->Add(chkNormed, 0, wxALL | wxEXPAND, borderWidth);
	mainSizer->Add(btnBuildPlot, 0, wxALL | wxEXPAND, borderWidth);
	panel->SetSizer(mainSizer);
	panel->Layout();

	// Set window icon
	SetIcon(wxICON(sample));
}

void DHPlotDialog::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

void DHPlotDialog::OnBuildDHPlot(wxCommandEvent &event)
{
	wxString strTitle = wxString::Format("CpFit H(T) - H(%g) data", T0);
	DHDataWindow *dHwnd = new DHDataWindow(this, strTitle, lsqCont, lsqRes, T0, DT0, chkNormed->GetValue());
	dHwnd->Show();
}

/**
 * \brief Controls T0 input and saves successful attempts inside the object (as double value)
 */
void DHPlotDialog::OnTxtT0(wxCommandEvent &event)
{
	double T;
	if (!event.GetString().ToDouble(&T) || T < 0.0 || T > 20000.0) {
		txtT0->ChangeValue(wxString::Format("%g", T0)); // Bad input: restore old value
	} else {
		T0 = T; // Good input: remember new value
	}
}

/**
 * \brief Controls DT0 input and saves successful attempts inside the object (as double value)
 */
void DHPlotDialog::OnTxtDT0(wxCommandEvent &event)
{
	double T;
	if (!event.GetString().ToDouble(&T) || T < 0 || T > 100) {
		txtDT0->ChangeValue(wxString::Format("%g", DT0)); // Bad input: restore old value
	} else {
		DT0 = T; // Good input: remember new value
	}
}


////////////////////////////////////////////////////////////


/**
 * \details The form layout:
 *
 *     +----------------{mainSizer}--------------+
 *     + +--------------{gridSizer}--------+---+ |
 *     | | Std.abs.dev: |   | MAD |        |   | |
 *     | +--------------+---+-----+--------+---+ |
 *     | | Std.rel.dev. |   | MAD |        |   | |
 *     | +--------------+---+-----+--------+---+ |
 *     | | Max.abs.dev. |   | at  |        | K | |
 *     | +--------------+---+-----+--------+---+ |
 *     | | Max.rel.dev. |   | at  |        | K | |
 *     | +--------------+---+-----+--------+---+ |
 *     +-----------------------------------------+
 *     | All abs.errors ad in J/(mol*K)          |
 *     +-----------------------------------------+
 * 
 */
DevInfoPanel::DevInfoPanel(wxWindow *parent, wxPoint pos, wxSize size, wxString unit)
	: wxPanel(parent, wxID_ANY, pos, size)
{
	// a) Absolute deviations
	wxStaticText *lblStdAbsDev = new wxStaticText(this, wxID_ANY, "Std.absolute dev.:"); // Names
	txtStdAbsDev = new wxTextCtrl(this, wxID_ANY, ""); // Devs
	txtStdAbsDev->SetEditable(false);
	// a1) MAD
	wxStaticText *lblMADAbsDev = new wxStaticText(this, wxID_ANY, "MAD"); // Names
	txtMADAbsDev = new wxTextCtrl(this, wxID_ANY, ""); // Devs
	txtMADAbsDev->SetEditable(false);
	// b) Relative deviations
	wxStaticText *lblStdRelDev = new wxStaticText(this, wxID_ANY, "Std.relative dev.(%):"); // At
	txtStdRelDev = new wxTextCtrl(this, wxID_ANY, ""); // K text
	txtStdRelDev->SetEditable(false);
	// b1) MAD
	wxStaticText *lblMADRelDev = new wxStaticText(this, wxID_ANY, "MAD"); // Names
	txtMADRelDev = new wxTextCtrl(this, wxID_ANY, ""); // Devs
	txtMADRelDev->SetEditable(false);
	// c) Max. Abs. Deviations (with T)
	wxStaticText *lblMaxAbsDev = new wxStaticText(this, wxID_ANY, "Max.absolute dev.:"); // Names
	txtMaxAbsDev = new wxTextCtrl(this, wxID_ANY, ""); // Devs
	txtMaxAbsDev->SetEditable(false);
	wxStaticText *lblMaxAbsDevAt = new wxStaticText(this, wxID_ANY, "at"); // At
	txtMaxAbsDevX = new wxTextCtrl(this, wxID_ANY, ""); // K text
	txtMaxAbsDevX->SetEditable(false);
	wxStaticText *lblMaxAbsDevK  = new wxStaticText(this, wxID_ANY, "K"); // K label
	// d) Max Rel. Dev (with T)
	wxStaticText *lblMaxRelDev = new wxStaticText(this, wxID_ANY, "Max.relative dev.(%):"); // Names
	txtMaxRelDev = new wxTextCtrl(this, wxID_ANY, ""); // Devs
	txtMaxRelDev->SetEditable(false);
	wxStaticText *lblMaxRelDevAt = new wxStaticText(this, wxID_ANY, "at"); // At
	txtMaxRelDevX = new wxTextCtrl(this, wxID_ANY, ""); // K text
	txtMaxRelDevX->SetEditable(false);
	wxStaticText *lblMaxRelDevK  = new wxStaticText(this, wxID_ANY, "K"); // K label
	// e) Info about dimensions
	wxStaticText *lblDimInfo = new wxStaticText(this, wxID_ANY,
		wxString::Format(L"All absolute errors are in %s", unit));

	// Sizers
	int borderWidth = 5;
	wxFlexGridSizer *gridSizer = new wxFlexGridSizer(4, 5, 5, 5);
	gridSizer->Add(lblStdAbsDev, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtStdAbsDev, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblMADAbsDev, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtMADAbsDev, 1, wxALL | wxEXPAND);
	gridSizer->AddStretchSpacer();

	gridSizer->Add(lblStdRelDev, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtStdRelDev, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblMADRelDev, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtMADRelDev, 1, wxALL | wxEXPAND);
	gridSizer->AddStretchSpacer();

	gridSizer->Add(lblMaxAbsDev, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtMaxAbsDev, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblMaxAbsDevAt, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtMaxAbsDevX, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblMaxAbsDevK, 0, wxALL | wxEXPAND);

	gridSizer->Add(lblMaxRelDev, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtMaxRelDev, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblMaxRelDevAt, 0, wxALL | wxEXPAND);
	gridSizer->Add(txtMaxRelDevX, 1, wxALL | wxEXPAND);
	gridSizer->Add(lblMaxRelDevK, 0, wxALL | wxEXPAND);

	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(gridSizer, 1, wxALL | wxEXPAND, borderWidth);
	mainSizer->Add(lblDimInfo);

	this->SetSizer(mainSizer);
	this->Layout();
}

void DevInfoPanel::LoadLSQDevInfo(LSQDevInfo &dev)
{
	if (dev.IsInitialized()) {
		txtMaxAbsDev->ChangeValue(wxString::Format("%.4g", dev.maxAbsDev));
		txtMaxAbsDevX->ChangeValue(wxString::Format("%.4g", dev.maxAbsDevX));
		txtMaxRelDev->ChangeValue(wxString::Format("%.4g", dev.maxRelDev * 100));
		txtMaxRelDevX->ChangeValue(wxString::Format("%.4g", dev.maxRelDevX));
		txtStdAbsDev->ChangeValue(wxString::Format("%.4g", dev.stdAbsDev));
		txtStdRelDev->ChangeValue(wxString::Format("%.4g", dev.stdRelDev * 100));
		txtMADAbsDev->ChangeValue(wxString::Format("%.4g", dev.medAbsDev));
		txtMADRelDev->ChangeValue(wxString::Format("%.4g", dev.medRelDev * 100));
	} else {
		txtMaxAbsDev->ChangeValue("");
		txtMaxAbsDevX->ChangeValue("");
		txtMaxRelDev->ChangeValue("");
		txtMaxRelDevX->ChangeValue("");
		txtStdAbsDev->ChangeValue("");
		txtStdRelDev->ChangeValue("");
		txtMADAbsDev->ChangeValue("");
		txtMADRelDev->ChangeValue("");
	}
}


////////////////////////////////////////////////////////////
#define ID_CMB_POINTSFILTER 2000
wxBEGIN_EVENT_TABLE(DevInfoListPanel, wxPanel)
	EVT_SIZE(DevInfoListPanel::OnSize)
	EVT_COMBOBOX(ID_CMB_POINTSFILTER, DevInfoListPanel::OnCmbPointsFilter)
wxEND_EVENT_TABLE()

DevInfoListPanel::DevInfoListPanel(wxWindow *parent, wxPoint pos, wxSize size)
	: wxPanel(parent, wxID_ANY, pos, size)
{
	wxString strFilters[3] = {wxT("All points"), wxT("Baseline points"), wxT("Anomalies points")};
	cbPointsFilter = new wxComboBox(this, ID_CMB_POINTSFILTER, strFilters[0], wxPoint(5, 5), wxSize(150, 30), 3, strFilters, wxCB_READONLY);

	grDevs = new wxGrid(this, wxID_ANY, wxPoint(5, 40), wxSize(230,200));
	grDevs->CreateGrid(0, 9);
	grDevs->SetColLabelValue(0, wxT("Data series"));
	grDevs->SetColLabelValue(1, wxT("Data type"));
	grDevs->SetColLabelValue(2, wxT("Points"));
	grDevs->SetColLabelValue(3, wxT("Abs.std.dev."));
	grDevs->SetColLabelValue(4, wxT("Rel.std.dev.(%)"));
	grDevs->SetColLabelValue(5, wxT("Max.abs.dev."));
	grDevs->SetColLabelValue(6, wxT("Max.rel.dev.(%)"));
	grDevs->SetColLabelValue(7, wxT("Med.abs.dev."));
	grDevs->SetColLabelValue(8, wxT("Med.rel.dev.(%)"));
	grDevs->SetRowLabelSize(40);
	grDevs->SetColSize(0, 80);
}

void DevInfoListPanel::OnCmbPointsFilter(wxCommandEvent &event)
{
	int ind = cbPointsFilter->GetSelection();
	switch (ind)
	{
	case 0: LoadLSQDevsInfo(devs); break;
	case 1: LoadLSQDevsInfo(devsBaseLine); break;
	case 2: LoadLSQDevsInfo(devsAnomaly); break;
	}
}



void DevInfoListPanel::LoadLSQDevsInfo(LSQResult &result)
{
	devs = result.GetYDeviations();
	devsBaseLine = result.GetYDeviations(POINTSFILTER_BASELINE);
	devsAnomaly = result.GetYDeviations(POINTSFILTER_ANOMALY);

	LoadLSQDevsInfo(devs);
}

void DevInfoListPanel::LoadLSQDevsInfo(std::vector <LSQDevInfo> &devs)
{
	// Fill the table
	int nrows = grDevs->GetNumberRows();
	if (nrows > 0){
		grDevs->DeleteRows(0, nrows);
	}
	for (size_t i = 0; i < devs.size(); i++)
	{
		LSQDevInfo &dev = devs[i];
		grDevs->InsertRows(i, 1);
		grDevs->SetCellValue(i, 0, dev.seriesID);
		switch (dev.dataType) {
		case DATATYPE_CP:
			grDevs->SetCellValue(i, 1, "Cp, J/(mol*K)");
			break;
		case DATATYPE_H:
			grDevs->SetCellValue(i, 1, "H, J/mol");
			break;
		case DATATYPE_S:
			grDevs->SetCellValue(i, 1, "S, J/(mol*K)");
			break;
		case DATATYPE_UNKNOWN:
			grDevs->SetCellValue(i, 1, "?");
			grDevs->SetCellValue(i, 2, "?");
			break;
		}
		grDevs->SetCellValue(i, 2, wxString::Format("%d\n", (int) dev.npoints));
		grDevs->SetCellValue(i, 3, wxString::Format("%.3g\n", dev.stdAbsDev));
		grDevs->SetCellValue(i, 4, wxString::Format("%.3g\n", dev.stdRelDev * 100));
		grDevs->SetCellValue(i, 5, wxString::Format("%.3g (at %.2f K)\n", dev.maxAbsDev, dev.maxAbsDevX));
		grDevs->SetCellValue(i, 6, wxString::Format("%.3g (at %.2f K)\n", dev.maxRelDev * 100, dev.maxRelDevX));
		grDevs->SetCellValue(i, 7, wxString::Format("%.3g\n", dev.medAbsDev));
		grDevs->SetCellValue(i, 8, wxString::Format("%.3g\n", dev.medRelDev * 100));
		for (int j = 0; j < 9; j++) {
			grDevs->SetReadOnly(i, j, true);
		}
	}
	grDevs->AutoSizeColumns();
}

void DevInfoListPanel::OnSize(wxSizeEvent &event)
{
	wxSize s = event.GetSize();
	grDevs->SetSize(s.x - 20, s.y - 60);
	grDevs->AutoSizeColumns();
}


////////////////////////////////////////////////////////////
wxBEGIN_EVENT_TABLE(EinsteinParametersPanel, wxPanel)
wxEND_EVENT_TABLE()

EinsteinParametersPanel::EinsteinParametersPanel(wxWindow *parent)
	: wxPanel(parent, wxID_ANY, wxPoint(0, 0), wxSize(420,230))
{
	grPar = new wxGridAutoColSize(this, wxID_ANY);
	grPar->CreateGrid(3, 6);
	grPar->SetColLabelValue(0, wxT("\u03B1")); // alpha
	grPar->SetColLabelValue(1, wxT("\u0394(\u03B1)")); // DELTA(alpha)
	grPar->SetColLabelValue(2, wxT("s(\u03B1)")); // s(alpha)
	grPar->SetColLabelValue(3, wxT("\u03B8 / K")); // theta
	grPar->SetColLabelValue(4, wxT("\u0394(\u03B8) / K")); // DELTA(theta)
	grPar->SetColLabelValue(5, wxT("s(\u03B8) / K")); // s(theta)
	grPar->SetRowLabelSize(40);

	txtAlphaSum = new wxTextCtrl(this, wxID_ANY, "", wxDefaultPosition, wxDefaultSize, wxTE_READONLY);
	wxBoxSizer *sizerAlphaSum = new wxBoxSizer(wxHORIZONTAL);
	sizerAlphaSum->Add(new wxStaticText(this, wxID_ANY, wxT("\u03B1 sum")), 1, wxALL | wxEXPAND, 0);
	sizerAlphaSum->Add(txtAlphaSum, 1, wxALL | wxEXPAND, 0);

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	sizer->Add(grPar, 1, wxALL | wxEXPAND, 5);
	sizer->Add(sizerAlphaSum, 0, wxALL | wxEXPAND, 5);

	this->SetSizer(sizer);
	this->Layout();
}

void EinsteinParametersPanel::SetParameters(TDBaseLineFuncParams &param,
	TDBaseLineFuncParams &sparam,
	TDBaseLineFuncParams &dparam)
{
	size_t nterms = param.NumAlphaThetaTerms();
	size_t nrows = grPar->GetNumberRows();
	double alphaSum = 0.0;
	if (nrows > 0) {
		grPar->DeleteRows(0, nrows);
	}
	grPar->InsertRows(0, nterms);
	for (size_t i = 0; i < nterms; i++) {
		// alpha parameters
		alphaSum += param.alpha[i];
		grPar->SetCellValue(i, 0, wxString::Format("%g", param.alpha[i]));
		grPar->SetCellValue(i, 1, wxString::Format("%.4g", dparam.alpha[i]));
		grPar->SetCellValue(i, 2, wxString::Format("%.4g", sparam.alpha[i]));
		if (fabs(param.alpha[i] <= dparam.alpha[i])) {
			for (int j = 0; j < 3; j++) {
				grPar->SetCellTextColour(i, j, wxColour(255, 0, 0));
			}
		} else {
			for (int j = 0; j < 3; j++) {
				grPar->SetCellTextColour(i, j, wxColour(0, 0, 0));
			}
		}
		// theta parameters
		grPar->SetCellValue(i, 3, wxString::Format("%g", param.theta[i]));
		grPar->SetCellValue(i, 4, wxString::Format("%.4g", dparam.theta[i]));
		grPar->SetCellValue(i, 5, wxString::Format("%.4g", sparam.theta[i]));
		if (fabs(param.theta[i] <= dparam.theta[i])) {
			for (int j = 3; j < 6; j++) {
				grPar->SetCellTextColour(i, j, wxColour(255, 0, 0));
			}
		} else {
			for (int j = 3; j < 6; j++) {
				grPar->SetCellTextColour(i, j, wxColour(0, 0, 0));
			}
		}
		// Set read only cells for alpha and theta parameters
		for (size_t j = 0; j < 6; j++) {
			grPar->SetReadOnly(i, j);
		}
	}
	// Show alpha sum
	txtAlphaSum->SetValue(wxString::Format("%g", alphaSum));
}

////////////////////////////////////////////////////////////
wxBEGIN_EVENT_TABLE(PolynomialParametersPanel, wxPanel)
wxEND_EVENT_TABLE()

PolynomialParametersPanel::PolynomialParametersPanel(wxWindow *parent)
	: wxPanel(parent, wxID_ANY, wxPoint(0, 0), wxSize(420,230))
{
	grParPoly = new wxGridAutoColSize(this, wxID_ANY, wxPoint(20, 20), wxSize(380, 160));
	grParPoly->CreateGrid(0, 4);
	grParPoly->SetColLabelValue(0, wxT("a")); // a
	grParPoly->SetColLabelValue(1, wxT("\u0394(a)")); // DELTA(a)
	grParPoly->SetColLabelValue(2, wxT("s(a)")); // s(a)
	grParPoly->SetColLabelValue(3, wxT("n")); // n
	grParPoly->SetRowLabelSize(40);

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	sizer->Add(grParPoly, 1, wxALL | wxEXPAND, 5);
	
	this->SetSizer(sizer);
	this->Layout();
}

void PolynomialParametersPanel::SetParameters(TDBaseLineFuncParams &param,
	TDBaseLineFuncParams &sparam,
	TDBaseLineFuncParams &dparam)
{
	size_t npolyterms = param.polyCoeff.size();
	size_t nrows = grParPoly->GetNumberRows();
	if (nrows > 0) {
		grParPoly->DeleteRows(0, nrows);
	}
	grParPoly->InsertRows(0, npolyterms);
	for (size_t i = 0; i < npolyterms; i++) {
		// polynomial coefficients
		grParPoly->SetCellValue(i, 0, wxString::Format("%g", param.polyCoeff[i]));
		grParPoly->SetCellValue(i, 1, wxString::Format("%.4g", dparam.polyCoeff[i]));
		grParPoly->SetCellValue(i, 2, wxString::Format("%.4g", sparam.polyCoeff[i]));
		if (fabs(param.polyCoeff[i] <= dparam.polyCoeff[i])) {
			for (int j = 0; j < 3; j++) {
				grParPoly->SetCellTextColour(i, j, wxColour(255, 0, 0));
			}
		} else {
			for (int j = 0; j < 3; j++) {
				grParPoly->SetCellTextColour(i, j, wxColour(0, 0, 0));
			}
		}
		// Polynomial degrees
		grParPoly->SetCellValue(i, 3, wxString::Format("%g", param.polyDegree[i]));
		// Set read only cells for all parameters
		for (size_t j = 0; j < 4; j++) {
			grParPoly->SetReadOnly(i, j);
		}
	}
}

////////////////////////////////////////////////////////////
wxBEGIN_EVENT_TABLE(ExtraTermsParametersPanel, wxPanel)
	EVT_SIZE(ExtraTermsParametersPanel::OnSize)
wxEND_EVENT_TABLE()

ExtraTermsParametersPanel::ExtraTermsParametersPanel(wxWindow *parent)
	: wxPanel(parent, wxID_ANY, wxPoint(0, 0), wxSize(420,230))
{
	wxFont *fntCourier = new wxFont(10, wxFONTFAMILY_TELETYPE, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_NORMAL);	
	txtExParamsReport = new wxTextCtrl(this, wxID_ANY, "Report\nReport\n",
		wxPoint(20, 30), wxSize(10, 10), wxTE_READONLY | wxTE_MULTILINE);
	txtExParamsReport->SetFont(*fntCourier);

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	sizer->Add(txtExParamsReport, 1, wxALL | wxEXPAND, 5);
	
	this->SetSizer(sizer);
	this->Layout();
}

void ExtraTermsParametersPanel::SetParameters(LSQResult &lsqRes)
{
	wxString strExParams;
	size_t nexterms = lsqRes.extraTerms.size();
	if (nexterms == 0) {
		strExParams = "No extra terms are present";
	} else {
		strExParams = wxString::Format("%d extra terms are present\n", (int) nexterms);
		for (size_t i = 0; i < nexterms; i++) {
			ExtraTerm *term = lsqRes.extraTerms[i].term;
			std::vector <double> fixparam = term->GetFixParam();
			std::vector <double> lsqparam = term->GetLSQParam();
			std::vector <double> slsqparam = term->GetLSQParamStd();
			std::vector <double> dlsqparam = term->GetLSQParamCi();
			std::vector <double> Tv = term->GetTVector();
			std::vector <double> Yv = term->GetYVector();
			std::vector <double> resv = term->GetResVector();
			std::vector <bool> isrelv = term->GetIsRelDevVector();
			// General information
			strExParams += wxString::Format("TERM %d\n", (int) (i + 1));
			strExParams += "Type: " + term->GetType() + "\n";
			strExParams += wxString::Format("Tmin = %g K, Tmax = %g K\n", term->Tmin, term->Tmax);
			// Fixed parameters info
			strExParams += wxString::Format("Fixed parameters:\n");
			strExParams += wxString::Format("  %3s %12s\n", "#", "A");
			for (size_t j = 0; j < fixparam.size(); j++) {
				strExParams += wxString::Format("  %3d %12g\n", (int) (j + 1), fixparam[j]);
			}
			// LSQ parameters info
			strExParams += wxString::Format("LSQ parameters:\n");
			strExParams += wxString::Format("  %3s %12s %12s %12s\n", "#", "A", "S(A)", "D(A)");
			for (size_t j = 0; j < lsqparam.size(); j++) {
				strExParams += wxString::Format("  %3d %12g %12g %12g\n",
					(int) (j + 1), lsqparam[j], slsqparam[j], dlsqparam[j]);
			}
			// Data series information
			strExParams += wxString::Format("Number of points: %d\n", (int) Tv.size());
			strExParams += wxString::Format("  %3s %12s %5s %12s %12s %12s\n",
				"#", "T", "DEV", "Cp(e)-Cp(bl)", "Cp(c)-Cp(bl)", "RES(Cp)");
			for (size_t j = 0; j < Tv.size(); j++) {
				strExParams += wxString::Format("  %3d %12g %5s %12g %12g %12g\n",
					(int) (j + 1), Tv[j], isrelv[j] ? "REL" : "ABS", Yv[j], term->CpFuncEval(Tv[j]), resv[j]);
			}
			// Excess functions information
			strExParams += wxString::Format("Sex(Tmax) = %g J/(mol*K)\n", term->SFuncEval(term->Tmax));
			strExParams += wxString::Format("Hex(Tmax) = %g J/mol\n", term->HFuncEval(term->Tmax));
			// Some extra statistical information
			strExParams += wxString::Format("Regression uncertainties\n");
			strExParams += wxString::Format("  RSS = %g, sigma2 = %g\n", term->sumsqr, term->sigma2);
			// Final empty line
			strExParams += "\n";
		}
	}
	txtExParamsReport->SetValue(strExParams);
}

void ExtraTermsParametersPanel::OnSize(wxSizeEvent &event)
{
	event.Skip();
}



////////////////////////////////////////////////////////////
enum {
	ID_GR_PARAM = 100,
	ID_GR_PARAM_POLY,
	ID_BTN_COPYPARAMS,

	ID_MNU_CPPLOT,
	ID_MNU_CPTPLOT,
	ID_MNU_CP_LOWT_PLOT,
	ID_MNU_CPT_T2_LOWT_PLOT,
	ID_MNU_CPT3_T2_LOWT_PLOT,
	ID_MNU_HPLOT,
	ID_MNU_SPLOT,
	ID_MNU_DHPLOT,

	ID_MNU_CPABSRESPLOT,
	ID_MNU_CPRELRESPLOT,

	ID_MNU_HABSRESPLOT,
	ID_MNU_HRELRESPLOT,

	ID_MNU_ALPHATHETA_SPECTRUM_PLOT,

	ID_MNU_TDTABLE,
	ID_MNU_CPRESTABLE,
	ID_MNU_HRESTABLE,

	ID_MNU_SAVEASCSV,
	ID_MNU_SAVEASHTML,

	ID_MNU_DEBYECORR
};

wxBEGIN_EVENT_TABLE(GraphWindow, wxFrame)
	EVT_SIZE(GraphWindow::OnSize)
	EVT_CLOSE(GraphWindow::OnClose)
	EVT_BUTTON(ID_BTN_COPYPARAMS, GraphWindow::OnCopyParams)

	EVT_MENU(ID_MNU_SAVEASCSV, GraphWindow::OnSaveAsCSV)
	EVT_MENU(ID_MNU_SAVEASHTML, GraphWindow::OnSaveAsHTML)

	EVT_MENU(ID_MNU_CPPLOT, GraphWindow::OnCpPlot)
	EVT_MENU(ID_MNU_CPTPLOT, GraphWindow::OnCpTPlot)
	EVT_MENU(ID_MNU_CP_LOWT_PLOT, GraphWindow::OnCpLowTPlot)
	EVT_MENU(ID_MNU_CPT_T2_LOWT_PLOT, GraphWindow::OnCpTT2LowTPlot)
	EVT_MENU(ID_MNU_CPT3_T2_LOWT_PLOT, GraphWindow::OnCpT3T2LowTPlot)

	EVT_MENU(ID_MNU_HPLOT, GraphWindow::OnHPlot)
	EVT_MENU(ID_MNU_SPLOT, GraphWindow::OnSPlot)
	EVT_MENU(ID_MNU_DHPLOT, GraphWindow::OnDHPlot)

	EVT_MENU(ID_MNU_ALPHATHETA_SPECTRUM_PLOT, GraphWindow::OnAlphaThetaSpectrumPlot)

	EVT_MENU(ID_MNU_TDTABLE, GraphWindow::OnShowTdTable)
	EVT_MENU(ID_MNU_CPRESTABLE, GraphWindow::OnShowCpResTable)
	EVT_MENU(ID_MNU_HRESTABLE, GraphWindow::OnShowHResTable)

	EVT_MENU(ID_MNU_CPABSRESPLOT, GraphWindow::OnCpAbsResPlot)
	EVT_MENU(ID_MNU_CPRELRESPLOT, GraphWindow::OnCpRelResPlot)

	EVT_MENU(ID_MNU_HABSRESPLOT, GraphWindow::OnHAbsResPlot)
	EVT_MENU(ID_MNU_HRELRESPLOT, GraphWindow::OnHRelResPlot)

	EVT_MENU(ID_MNU_DEBYECORR, GraphWindow::OnDebyeCorrection)
wxEND_EVENT_TABLE()


/**
 * \brief Initializes all controls and menus
 *
 * Note: complex interaction with GraphWindow::OnSize function by means of some sizes
 * initialized in this function.
 */
GraphWindow::GraphWindow(MainWindow *parent, const wxString &title)
	: wxFrame(parent, wxID_ANY, title, wxPoint(50, 50), wxSize(600, 650))
{
	boxParams = NULL; boxLsq = NULL;
	wxPanel *panel = new wxPanel(this, -1);

	wxBoxSizer *sizer = new wxBoxSizer(wxVERTICAL);
	// Create all objects in the parameters static box
	boxParams = new wxStaticBox(panel, wxID_ANY, "Model parameters", wxPoint(10, 10), wxSize(420, 230));
	wxStaticBoxSizer *sizerParams = new wxStaticBoxSizer(boxParams, wxVERTICAL);
	nbkParams = new wxNotebook(boxParams, wxID_ANY, wxPoint(10, 10), wxSize(420, 230));
	ctlBlParams = new EinsteinParametersPanel(nbkParams);
	ctlBlPolyParams = new PolynomialParametersPanel(nbkParams);
	ctlExParams = new ExtraTermsParametersPanel(nbkParams);

	nbkParams->AddPage(ctlBlParams, "Base line (Einstein)");
	nbkParams->AddPage(ctlBlPolyParams, "Base line (polynomial)");
	nbkParams->AddPage(ctlExParams, "Extra terms");

	btnCopyParams = new wxButton(boxParams, ID_BTN_COPYPARAMS, "Use as initial approximation", wxPoint(20,190), wxSize(380,30));

	sizerParams->Add(nbkParams, 1, wxALL | wxEXPAND, 5);
	sizerParams->Add(btnCopyParams, 0, wxALL | wxEXPAND, 5);

	sizer->Add(sizerParams, 3, wxALL | wxEXPAND, 5);
	// Create all objects for the Cp model errors static box
	boxLsq = new wxStaticBox(panel, wxID_ANY, "Model errors", wxPoint(10, 250), wxSize(420, 260));
	wxStaticBoxSizer *sizerLsq = new wxStaticBoxSizer(boxLsq, wxVERTICAL);
	nbkTabsDev = new wxNotebook(boxLsq, wxID_ANY, wxPoint(10, 25), wxSize(400, 210));
	CpDevPanel = new DevInfoPanel(nbkTabsDev, wxPoint(0, 0), wxSize(400, 180));
	HDevPanel = new DevInfoPanel(nbkTabsDev, wxPoint(0, 0), wxSize(400, 180), "J / mol");
	SeriesDevPanel = new DevInfoListPanel(nbkTabsDev, wxPoint(0, 0), wxSize(400, 180));
	nbkTabsDev->AddPage(CpDevPanel, "Cp (heat capacity)");
	nbkTabsDev->AddPage(HDevPanel, "H(T2)-H(T1) (enthalpy)");
	nbkTabsDev->AddPage(SeriesDevPanel, "Data series");

	sizerLsq->Add(nbkTabsDev, 1, wxALL | wxEXPAND, 5);

	sizer->Add(sizerLsq, 2, wxALL | wxEXPAND, 5);
	// Set window icon
	SetIcon(wxICON(sample));
	// Create top menu
	CreateMenu();
	// Set some default optimization parameters
	Natoms = -1; // Default option is "no fixed Natoms value"
	// Set sizer for layout
	panel->SetSizer(sizer);
	panel->Layout();
	// Set window coordinates and sizes
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetMinSize(wxSize(600, 700));
}


void GraphWindow::CreateMenu()
{
	// Create user menu
	wxMenuBar *menuBar = new wxMenuBar();
	// Create file menu
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_MNU_SAVEASCSV, "Save as CSV", "Save as CSV");
	menuFile->Append(ID_MNU_SAVEASHTML, "Save as HTML", "Save as HTML");
	menuBar->Append(menuFile, "File");
	// Create plots menu
	wxMenu *menuPlots = new wxMenu();
	menuBar->Append(menuPlots, "Plots");
	// Plots menu: Cp plots submenu
	wxMenu *menuCpPlots = new wxMenu();
	menuCpPlots->Append(ID_MNU_CPPLOT, "Function", "Function");
	menuCpPlots->Append(ID_MNU_CPABSRESPLOT, "Absolute residuals", "Absolute residuals");
	menuCpPlots->Append(ID_MNU_CPRELRESPLOT, "Relative residuals", "Relative residuals");
	menuPlots->AppendSubMenu(menuCpPlots, "Cp function");
	// Plots menu: Cp low-T submenu
	wxMenu *menuCpLTPlots = new wxMenu();
	menuCpLTPlots->Append(ID_MNU_CP_LOWT_PLOT, "Cp-T low-temperature plot", "Cp-T low-temperature plot");
	menuCpLTPlots->Append(ID_MNU_CPT_T2_LOWT_PLOT, "Cp/T-T2 low-temperature plot", "Cp/T-T2 low-temperature plot");
	menuCpLTPlots->Append(ID_MNU_CPT3_T2_LOWT_PLOT, "Cp/T3-T2 low-temperature plot", "Cp/T3-T2 low-temperature plot");
	menuPlots->AppendSubMenu(menuCpLTPlots, "Cp low T function");
	// Plots menu: Cp/T plots submenu
	wxMenu *menuCpTPlots = new wxMenu();
	menuCpTPlots->Append(ID_MNU_CPTPLOT, "Cp/T plot", "Cp/T function");
	menuPlots->AppendSubMenu(menuCpTPlots, "Cp/T function");
	// Plots menu: H plots submenu
	wxMenu *menuHPlots = new wxMenu();
	menuHPlots->Append(ID_MNU_HPLOT, "H plot", "H plot");
	menuHPlots->Append(ID_MNU_HABSRESPLOT, "Absolute residuals", "Absolute residuals");
	menuHPlots->Append(ID_MNU_HRELRESPLOT, "Relative residuals", "Relative residuals");
	menuHPlots->Append(ID_MNU_DHPLOT, "H(T)-H(T0) plot", "H(T)-H(T0) plot");
	menuPlots->AppendSubMenu(menuHPlots, "H function");
	// Plots menu: S plots submenu
	wxMenu *menuSPlots = new wxMenu();
	menuSPlots->Append(ID_MNU_SPLOT, "S plot", "S plot");
	menuPlots->AppendSubMenu(menuSPlots, "S function");
	// Plots menu: alpha-theta spectrum item
	menuPlots->Append(ID_MNU_ALPHATHETA_SPECTRUM_PLOT, wxT("\u03B1-\u03B8 spectrum"), wxT("\u03B1-\u03B8 spectrum"));
	// Plots menu: Tables menu
	wxMenu *menuTables = new wxMenu();
	menuTables->Append(ID_MNU_TDTABLE, "Thermodynamic functions", "Thermodynamic functions");
	wxMenu *menuResiduals = new wxMenu();
	menuResiduals->Append(ID_MNU_CPRESTABLE, "Cp residuals", "Cp residuals");
	menuResiduals->Append(ID_MNU_HRESTABLE, "H(T2)-H(T1) residuals", "H(T2)-H(T1) residuals");
	menuTables->AppendSubMenu(menuResiduals, "Residuals");
	menuBar->Append(menuTables, "Tables");
	// Evaluate menu: different numeric estimations
	wxMenu *menuEvals = new wxMenu();
	menuEvals->Append(ID_MNU_DEBYECORR, "Low-T S correction", "Low-T S correction");
	menuBar->Append(menuEvals, "Evaluate");
	// Append the whole menu to the form
	SetMenuBar(menuBar);
}

/**
 * \brief Set experimental data
 */
void GraphWindow::AddData(LSQContext lsq)
{
	this->lsq = lsq;
}

/**
 * \brief Set initial approximation
 */
void GraphWindow::AddParam0(const TDBaseLineFuncParams &param)
{
	param0 = param;
}

/**
 * \brief Sets fixed number of Natoms
 */
void GraphWindow::AddFixedNatoms(double Natoms)
{
	this->Natoms = Natoms;
}

/**
 * \brief Sets varied number of natoms
 */
void GraphWindow::DeleteFixedNatoms()
{
	this->Natoms = -1;
}


/**
 * \brief Run the least squares method
 *
 * Starts the least squares method (LSQContext::dataFit) using initial approximation
 * given by a user and loads all results to the controls of the class. The main results
 * are coefficients, their standard deviations, and absolute and relative deviations
 * of Cp data description.
 */
void GraphWindow::CurveFit(GraphWindowFitMode fitMode)
{
	TDBaseLineFuncParams param;
	// Run data fitting	
	// a) initialize initial approximation (if required)
	if (fitMode != FIT_VALNTERMS_AUTO) {
		param = param0;
	}
	// b) select an appropriate method
	TDPEFunc baseLineFuncs;
	lsq.Natoms = Natoms;
	lsq.useConstraints = false;
	switch (fitMode) {
	case FIT_VALNTERMS_AUTO:
		lsqRes = lsq.DataFitWithPoly(param0.polyCoeff, param0.polyDegree);
		if (!lsqRes.isParamPositive()) {
			lsq.useConstraints = true;
			lsqRes = lsq.DataFitWithPoly(param0.polyCoeff, param0.polyDegree);
		}
		break;
	case FIT_VALS:
		lsqRes = lsq.DataFit(param);
		if (!lsqRes.isParamPositive()) {
			lsq.useConstraints = true;
			lsqRes = lsq.DataFit(param);
		}
		break;
	case FIT_VALNTERMS_INITAPPROX:
		lsqRes = lsq.DataFitAddTerms(param);
		if (!lsqRes.isParamPositive()) {
			lsq.useConstraints = true;
			lsqRes = lsq.DataFitAddTerms(param);
		}
		break;
	case FIT_EXTERMS_ONLY:
		lsqRes = lsq.DataFit(param, true, false); // Optimize extra terms only
		break;
	case FIT_NOOPT:
		lsqRes = lsq.NoDataFit(param);
		break;
	}
	// Get parameters
	param = lsqRes.GetParam();
	TDBaseLineFuncParams sparam = lsqRes.GetParamStd();
	TDBaseLineFuncParams dparam = lsqRes.GetParamConfInt();
	// Warning in the case of bad optimization results
	if (!lsqRes.IsExitCodeGood()) {
		wxMessageBox(lsqRes.GetExitMessage().c_str(), "CpFit", wxICON_WARNING);
	}
	// Visualize parameters: alpha/theta, polynomial, extra terms
	ctlBlParams->SetParameters(param, sparam, dparam);
	ctlBlPolyParams->SetParameters(param, sparam, dparam);
	ctlExParams->SetParameters(lsqRes);
	// Get and visualize Cp and H deviations
	LSQDevInfo devCp = lsqRes.GetYDeviations(DATATYPE_CP);
	CpDevPanel->LoadLSQDevInfo(devCp);
	LSQDevInfo devH = lsqRes.GetYDeviations(DATATYPE_H);
	HDevPanel->LoadLSQDevInfo(devH);
	SeriesDevPanel->LoadLSQDevsInfo(lsqRes);
}

/**
 * \brief Resizing some controls
 */
void GraphWindow::OnSize(wxSizeEvent &event)
{
	event.Skip();
}

void GraphWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

/**
 * \brief Sends optimized parameters to the parent window
 *
 * Sends optimized parameters to the parent window. They will be used
 * as an initial approximation and will appear in the corresponding table
 * of the CpFit main window.
 */
void GraphWindow::OnCopyParams(wxCommandEvent &event)
{
	TDBaseLineFuncParams param = lsqRes.GetParam();
	MainWindow *parent = (MainWindow *) this->GetParent();
	parent->WriteAlpha(param.alpha);
	parent->WriteTheta(param.theta);
	parent->WritePolyCoeff(param.polyCoeff);
	parent->WritePolyDegree(param.polyDegree);
	if (!parent->WriteExtraTermsParams(lsqRes.extraTerms)) {
		wxMessageBox("Cannot copy extra terms: sizes or types are not matching");
	}
}

/**
 * \brief Shows a window with a table of tabulated thermodynamic functions
 */
void GraphWindow::OnShowTdTable(wxCommandEvent &event)
{
	TdTableWindow *tdtab = new TdTableWindow(this, &lsqRes, "CpFit thermodynamic functions table");
	tdtab->Show();
}

/**
 * \brief Shows a window with residuals (experimental Cp vs calculated Cp)
 */
void GraphWindow::OnShowCpResTable(wxCommandEvent &event)
{
	ResTableWindow *restab = new ResTableWindow(this, DATATYPE_CP, lsq, lsqRes, wxString("CpFit residuals table"));
	restab->Show();
}

/**
 * \brief Shows a window with residuals (experimental H(T2)-H(T1) vs calculated H(T2)-H(T1))
 */
void GraphWindow::OnShowHResTable(wxCommandEvent &event)
{
	ResTableWindow *restab = new ResTableWindow(this, DATATYPE_H, lsq, lsqRes, wxString("CpFit residuals table"));
	restab->Show();
}

/**
 * \brief Shows Cp(T) plot with experimental points and calculated curve
 */
void GraphWindow::OnCpPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit Cp Data", lsq, lsqRes, DATATYPE_CP);
	cpwnd->Show();
}

/**
 * \brief Shows Cp(T)/T plot with experimental points and calculated curve
 *
 * Cp(T)/T terms graphs are also present (colored dashed lines)
 */
void GraphWindow::OnCpTPlot(wxCommandEvent &event)
{
	CpTDataWindow *cpwnd = new CpTDataWindow(this, "CpFit Cp/T Data", lsq, lsqRes);
	cpwnd->Show();
}

/**
 * \brief Shows the low-temperature part of Cp(T) - T plot with experimental
 * points and calculated curve.
 */
void GraphWindow::OnCpLowTPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit Cp Low-T Data", lsq, lsqRes,
		DATATYPE_CP, YDATAWINDOW_LOWT_T_CP);
	cpwnd->Show();
}

/**
 * \brief Shows the low-temperature part of Cp(T)/T - T2 plot with experimental
 * points and calculated curve. It is useful for Cp(T) = AT + BT3 asymptotic
 * behaviour visualization.
 */
void GraphWindow::OnCpTT2LowTPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit Cp/T2-T2 Data", lsq, lsqRes,
		DATATYPE_CP, YDATAWINDOW_LOWT_T2_CPT);
	cpwnd->Show();
}

/**
 * \brief Shows the low-temperature part of Cp(T)/T3 - T2 plot with experimental
 * points and calculated curve. It is useful for Cp(T) = AT3 + BT5 asymptotic
 * behaviour visualization.
 */
void GraphWindow::OnCpT3T2LowTPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit Cp/T3-T2 Data", lsq, lsqRes,
		DATATYPE_CP, YDATAWINDOW_LOWT_T2_CPT3);
	cpwnd->Show();
}


/**
 * \brief Shows H(T)-H(0) plot with calculated curve
 */
void GraphWindow::OnHPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit H Data", lsq, lsqRes, DATATYPE_H);
	cpwnd->Show();
}

/**
 * \brief Shows H(T)-H(T0) plot with calculated curve and experimental points
 */
void GraphWindow::OnDHPlot(wxCommandEvent &event)
{
	DHPlotDialog *dlg = new DHPlotDialog(this, lsq, lsqRes);
	dlg->Show(true);
}

/**
 * \brief Shows S(T)-S(0) plot with calculated curve
 */
void GraphWindow::OnSPlot(wxCommandEvent &event)
{
	YDataWindow *cpwnd = new YDataWindow(this, "CpFit S Data", lsq, lsqRes, DATATYPE_S);
	cpwnd->Show();
}


/**
 * \brief Shows Cp absolute residuals
 */
void GraphWindow::OnCpAbsResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit Cp absolute residuals", lsq, lsqRes, DATATYPE_CP, false);
	reswnd->Show();
}

/**
 * \brief Shows Cp relative residuals
 */
void GraphWindow::OnCpRelResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit Cp relative residuals", lsq, lsqRes, DATATYPE_CP, true);
	reswnd->Show();
}

/**
 * \brief Shows H absolute residuals
 */
void GraphWindow::OnHAbsResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit H absolute residuals", lsq, lsqRes, DATATYPE_H, false);
	reswnd->Show();
}

/**
 * \brief Shows H relative residuals
 */
void GraphWindow::OnHRelResPlot(wxCommandEvent &event)
{
	ResDataWindow *reswnd = new ResDataWindow(this, "CpFit H relative residuals", lsq, lsqRes, DATATYPE_H, true);
	reswnd->Show();
}

void GraphWindow::OnAlphaThetaSpectrumPlot(wxCommandEvent &event)
{
	AlphaThetaSpectrumWindow *althWnd = new AlphaThetaSpectrumWindow(this,
		wxT("CpFit \u03B1-\u03B8 spectrum"), lsq, lsqRes); // alpha-theta
	althWnd->Show();
}



/**
 * \brief Exports parameters and their confidence intervals into CSV file
 */
void GraphWindow::OnSaveAsCSV(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Save image"), "", "",
                       "csv files (*.csv)|*.csv", wxFD_SAVE);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		wxTextFile csvFile;
		if (csvFile.Create(fileName)) {
			TDBaseLineFuncParams param = lsqRes.GetParam();
			TDBaseLineFuncParams sparam = lsqRes.GetParamStd();
			TDBaseLineFuncParams dparam = lsqRes.GetParamConfInt();
			csvFile.AddLine("alpha;dalpha;salpha;theta;dtheta;stheta");
			for (size_t i = 0; i < param.NumAlphaThetaTerms(); i++) {
				csvFile.AddLine(wxString::Format("%.15g;%g;%g;%.15g;%g;%g",
					param.alpha[i], dparam.alpha[i], sparam.alpha[i],
					param.theta[i+1], dparam.theta[i+1], sparam.theta[i+1]));
			}
			csvFile.Write();
			csvFile.Close();
		} else {
			wxMessageBox("Cannot create CSV file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
}


/**
 * \brief Converts baseline (Einstein and polynomial parts ) model
 * parameters to the HTML report.
 */
void GraphWindow::SaveBaselineParametersToHTML(wxTextFile &htmlFile)
{
	// --- Baseline parameters
	htmlFile.AddLine("<h2>Optimized baseline parameters set</h2>");
	// Get data
	TDBaseLineFuncParams param = lsqRes.GetParam();
	TDBaseLineFuncParams sparam = lsqRes.GetParamStd();
	TDBaseLineFuncParams dparam = lsqRes.GetParamConfInt();
	// alpha/theta table header
	htmlFile.AddLine("<h3>Einstein functions sum part</h3>");
	htmlFile.AddLine("<p><i>C<sub>p</sub></i><sup>Ein</sup> = &Sigma;<sub><i>i</i></sub>");
	htmlFile.AddLine("&alpha;<sub><i>i</i></sub><i>C</i><sub>E</sub>(&theta;<sub><i>i</i></sub> / <i>T</i>); ");
	htmlFile.AddLine("C</i><sub>E</sub>(<i>x</i>) / <i>R</i> = 3<i>x</i><sup>2</sup><i>e<sup>x</sup></i> / (<i>e<sup>x</sup></i> - 1)<sup>2<sup>");

	htmlFile.AddLine("<table>");
	htmlFile.AddLine("<style>");
	htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
	htmlFile.AddLine("td {border: 1px solid black;}");
	htmlFile.AddLine("</style>");
	htmlFile.AddLine("<tr><td>&alpha;</td> <td>&Delta;(&alpha;)</td> <td><i>s</i>(&alpha;)</td>");
	htmlFile.AddLine("<td>&theta;&nbsp;/&nbsp;K</td> <td>&Delta;(&theta;)&nbsp;/&nbsp;K</td> <td><i>s</i>(&theta;)&nbsp;/&nbsp;K</td> </tr>");
	// Baseline parameters table content
	for (size_t i = 0; i < param.NumAlphaThetaTerms(); i++) {
		htmlFile.AddLine(wxString::Format("<tr><td>%.15g</td><td>%g</td><td>%g</td>",
			param.alpha[i], dparam.alpha[i], sparam.alpha[i]));
		htmlFile.AddLine(wxString::Format("    <td>%.15g</td><td>%g</td><td>%g</td></tr>",
			param.theta[i], dparam.theta[i], sparam.theta[i]));
	}
	// Baseline parameters table ending
	htmlFile.AddLine("</table>\n");

	// Polynomial table header
	htmlFile.AddLine("<h3>Polynomial part</h3>");
	htmlFile.AddLine("<p><i>C<sub>p</sub></i><sup>poly</sup> / <i>R</i> = &Sigma;<sub><i>i</i></sub>");
	htmlFile.AddLine("<i>a</i><sub><i>i</i></sub> (<i>T</i>/<i>T</i><sub>0</sub>)<sup><i>n</i></sup>");
	htmlFile.AddLine("<i>T</i><sub>0</sub> = 298.15&nbsp;K</p>");


	htmlFile.AddLine("<table>");
	htmlFile.AddLine("<style>");
	htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
	htmlFile.AddLine("td {border: 1px solid black;}");
	htmlFile.AddLine("</style>");
	htmlFile.AddLine("<tr><td>a</td> <td>&Delta;(a)</td> <td><i>s</i>(a)</td> <td><i>n</i></td></tr>");
	// Polynomial table contents
	for (size_t i = 0; i < param.polyCoeff.size(); i++) {
		htmlFile.AddLine(wxString::Format("<tr><td>%.15g</td> <td>%g</td> <td>%g</td> <td>%g</td></tr>",
			param.polyCoeff[i], dparam.polyCoeff[i], sparam.polyCoeff[i], param.polyDegree[i]));
	}
	// Polynomial table ending
	htmlFile.AddLine("</table>");
}

/**
 * \brief Converts extra terms parameters to HTML report.
 */
void GraphWindow::SaveExtraParametersToHTML(wxTextFile &htmlFile)
{
	// Extra terms parameters table header
	htmlFile.AddLine("<h2>Optimized extra terms parameters set</h2>");
	// Extra terms parameters tables
	for (unsigned int i = 0; i < lsqRes.extraTerms.size(); i++) {
		ExtraTerm *term = lsqRes.extraTerms[i].term;
		htmlFile.AddLine(wxString::Format("<h3>Term %d; Type: %s</h3>\n", (int) (i + 1), term->GetType()));
		// a) Tmin and Tmax (constants)
		htmlFile.AddLine(wxString::Format("<p><i>T</i><sub>min</sub> = %g &nbsp;K</p>\n", term->Tmin));
		htmlFile.AddLine(wxString::Format("<p><i>T</i><sub>min</sub> = %g &nbsp;K</p>\n", term->Tmax));
		// b) Constant parameters
		htmlFile.AddLine("<p><b>Constant parameters</b></p>\n");
		htmlFile.AddLine("<table>");
		htmlFile.AddLine("<style>");
		htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
		htmlFile.AddLine("td {border: 1px solid black;}");
		htmlFile.AddLine("</style>");
		htmlFile.AddLine("<tr><td><i>i</i></td> <td>Name</td> <td><i>c<sub>i</sub></i></td></tr>\n");
		std::vector <double> fixparam = term->GetFixParam();
		std::vector <std::string> fixparam_names = term->GetFixParamNames();
		for (unsigned int j = 0; j < fixparam.size(); j++) {
			htmlFile.AddLine(wxString::Format(
				"<tr><td>%d</td> <td>%s</td> <td>%.15g</td></tr>\n",
				(int) (j + 1), fixparam_names[j], fixparam[j]
			));
		}
		htmlFile.AddLine("</table>");
		// c) LSQ parameters (optimized)
		htmlFile.AddLine("<p><b>LSQ parameters</b></p>\n");
		htmlFile.AddLine("<table>");
		htmlFile.AddLine("<style>");
		htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
		htmlFile.AddLine("td {border: 1px solid black;}");
		htmlFile.AddLine("</style>");
		htmlFile.AddLine("<tr><td><i>i</i></td> <td>Name</td> <td><i>b<sub>i</sub></i></td> <td><i>s</i>(<i>b<sub>i</sub></i>)</td> <td>&Delta;(<i>b<sub>i</sub></i>)</td></tr>\n");
		std::vector <double> lsqparam = term->GetLSQParam();
		std::vector <double> lsqparam_std = term->GetLSQParamStd();
		std::vector <double> lsqparam_ci = term->GetLSQParamCi();
		std::vector <std::string> lsqparam_names = term->GetLSQParamNames();
		for (unsigned int j = 0; j < lsqparam.size(); j++) {
			htmlFile.AddLine(wxString::Format(
				"<tr><td>%d</td> <td>%s</td> <td>%.15g</td> <td>%.15g</td> <td>%.15g</td></tr>\n",
				(int) (j + 1), lsqparam_names[j], lsqparam[j], lsqparam_std[j], lsqparam_ci[j]
			));
		}
		htmlFile.AddLine("</table>\n");
	}
}


void GraphWindow::SaveSeriesDeviationsToHTML(wxTextFile &htmlFile, std::vector <LSQDevInfo> &devs)
{
	// Table header
	htmlFile.AddLine("<table>");
	htmlFile.AddLine("<style>");
	htmlFile.AddLine("table {border-collapse: collapse; border: 4px double black;}");
	htmlFile.AddLine("td {border: 1px solid black;}");
	htmlFile.AddLine("</style>");

	htmlFile.AddLine("<tr>");
	htmlFile.AddLine("  <td>Data series</td> <td>Data type</td>");
	htmlFile.AddLine("  <td>Points</td>");
	htmlFile.AddLine("  <td>Abs.std.dev.</td> <td>Rel.std.dev.(%)</td>");
	htmlFile.AddLine("  <td>Max.abs.dev.</td> <td>Max.rel.dev.(%)</td>");
	htmlFile.AddLine("  <td>Med.abs.dev.</td> <td>Med.rel.dev.(%)</td>");
	htmlFile.AddLine("</tr>");


	// Data series
	for (size_t i = 0; i < devs.size(); i++)
	{
		LSQDevInfo &dev = devs[i];
		if (dev.npoints <= 0) { // Entry is absent for this data series
			continue; // (e.g. Cp entry for H data or anomalies data 
		}                 // for data without anomalies)
		htmlFile.AddLine("<tr>");
		htmlFile.AddLine("  <td>" + wxString(dev.seriesID) + "</td>");
		switch (dev.dataType) {
		case DATATYPE_CP:
			htmlFile.AddLine("  <td><i>C<sub>p</sub></i>, J/(mol*K)</td>");
			break;
		case DATATYPE_H:
			htmlFile.AddLine("  <td><i>H</i><sub><i>T</i></sub> - <i>H</i><sub>0</sub>, J/mol</td>");
			break;
		case DATATYPE_S:
			htmlFile.AddLine("  <td><i>S</i>, J/(mol*K)</td>");
			break;
		case DATATYPE_UNKNOWN:
			htmlFile.AddLine("  <td>?</td>");
			break;
		}
		htmlFile.AddLine(wxString::Format("  <td>%d</td>", (int) dev.npoints));
		htmlFile.AddLine(wxString::Format("  <td>%.3g</td>", dev.stdAbsDev));
		htmlFile.AddLine(wxString::Format("  <td>%.3g</td>", dev.stdRelDev * 100));
		htmlFile.AddLine(wxString::Format("  <td>%.3g (at %.2f K)</td>", dev.maxAbsDev, dev.maxAbsDevX));
		htmlFile.AddLine(wxString::Format("  <td>%.3g (at %.2f K)</td>", dev.maxRelDev * 100, dev.maxRelDevX));
		htmlFile.AddLine(wxString::Format("  <td>%.3g\n</td>", dev.medAbsDev));
		htmlFile.AddLine(wxString::Format("  <td>%.3g\n</td>", dev.medRelDev * 100));
		htmlFile.AddLine("</tr>");
	}
	// Table ending
	htmlFile.AddLine("</table>\n");
}


/**
 * \biref Exports deviations of experimental points from the obtained model
 * to the HTML report.
 */
void GraphWindow::SaveDeviationsToHTML(wxTextFile &htmlFile)
{
	wxString fmtDevHdr = "<h2>%s</h2><table><style>"
		"table {border-collapse: collapse; border: 4px double black;}"
		"td {border: 1px solid black;}</style>";

	// a) heat capacities (Cp)
	//   header
	htmlFile.AddLine(wxString::Format(fmtDevHdr, "<i>C<sub>p</sub></i> deviations"));
	//   body
	LSQDevInfo devCp = lsqRes.GetYDeviations(DATATYPE_CP);
	htmlFile.AddLine(wxString::Format("<tr><td>Standard absolute deviation, J/(mol*K)</td>    <td>%g</td><td></td></tr>", devCp.stdAbsDev));
	htmlFile.AddLine(wxString::Format("<tr><td>Standard relative deviation, %%</td> <td>%g</td><td></td></tr>", devCp.stdRelDev * 100));
	htmlFile.AddLine(wxString::Format("<tr><td>Maximal absolute deviation, J/(mol*K)</td>     <td>%g</td><td>(at %g K)</td></tr>",
		devCp.maxAbsDev, devCp.maxAbsDevX));
	htmlFile.AddLine(wxString::Format("<tr><td>Maximal relative deviation, %%</td>  <td>%g</td><td>(at %g K)</td></tr>",
		devCp.maxRelDev * 100, devCp.maxRelDevX));
	//   ending
	htmlFile.AddLine("</table>");
	// b) enthalpies (H)
	//   header
	htmlFile.AddLine(wxString::Format(fmtDevHdr, "<h2><i>H</i>(<i>T<sub>2</sub></i>)-<i>H</i>(<i>T</i><sub>1</sub>) deviations</h2>"));
	//   body
	LSQDevInfo devH = lsqRes.GetYDeviations(DATATYPE_H);
	htmlFile.AddLine(wxString::Format("<tr><td>Standard absolute deviation, J/mol</td>    <td>%g</td><td></td></tr>", devH.stdAbsDev));
	htmlFile.AddLine(wxString::Format("<tr><td>Standard relative deviation, %%</td> <td>%g</td><td></td></tr>", devH.stdRelDev * 100));
	htmlFile.AddLine(wxString::Format("<tr><td>Maximal absolute deviation, J/mol</td>     <td>%g</td><td>(at %g K)</td></tr>",
		devH.maxAbsDev, devH.maxAbsDevX));
	htmlFile.AddLine(wxString::Format("<tr><td>Maximal relative deviation, %%</td>  <td>%g</td><td>(at %g K)</td></tr>",
		devH.maxRelDev * 100, devH.maxRelDevX));
	//   ending
	htmlFile.AddLine("</table>");
	
	// Deviations for individual data series
	htmlFile.AddLine("<h2>Deviations for individual data series</h2>");
	htmlFile.AddLine("<h3>All points</h3>");
	SaveSeriesDeviationsToHTML(htmlFile, lsqRes.devs);
	htmlFile.AddLine("<h3>Points outside anomalies regions</h3>");
	SaveSeriesDeviationsToHTML(htmlFile, lsqRes.devsBaseLine);
	htmlFile.AddLine("<h3>Points inside anomalies regions</h3>");
	SaveSeriesDeviationsToHTML(htmlFile, lsqRes.devsAnomaly);
}




/**
 * \brief Exports parameters, their confidence intervals and
 * some another data into HTML file (report)
 */
void GraphWindow::OnSaveAsHTML(wxCommandEvent &event)
{
	wxFileDialog openFileDialog(this, wxT("Save as HTML report"), "", "",
                       "html files (*.html)|*.html", wxFD_SAVE);
	if (openFileDialog.ShowModal() != wxID_CANCEL) {
		wxString fileName = openFileDialog.GetPath();
		wxTextFile htmlFile;
		if (htmlFile.Create(fileName)) {
			wxDateTime dtNow = wxDateTime(wxDateTime::GetTimeNow());
			// --- General header
			htmlFile.AddLine("<html><head><meta charset = \"utf-8\"/></head><body>");
			htmlFile.AddLine("<h1>Report</h1>");
			htmlFile.AddLine("<h2>General info</h2>");
			htmlFile.AddLine("<p>Date: " + dtNow.FormatDate() + " </p>");
			htmlFile.AddLine("<p>Time: " + dtNow.FormatTime() + " </p>");
			htmlFile.AddLine(wxString::Format("<p>Total number of points: %d", (int) lsq.data.size()));
			// --- Baseline and extra terms parameters
			SaveBaselineParametersToHTML(htmlFile);
			if (lsqRes.extraTerms.size() > 0) {
				SaveExtraParametersToHTML(htmlFile);
			}
			SaveDeviationsToHTML(htmlFile);
			// Final of the report
			htmlFile.AddLine("</body></html>");
			// Closing the file
			htmlFile.Write();
			htmlFile.Close();
		} else {
			wxMessageBox("Cannot create HTML file", "CpFit", wxICON_ERROR | wxOK);
		}
	}
}

/**
 * \brief Opens window for estimation of differences between obtained
 * approximation and Debye model.
 */
void GraphWindow::OnDebyeCorrection(wxCommandEvent &event)
{
	TDPEFunc *f = &lsqRes;
	MainWindow *mainWnd = (MainWindow *) this->GetParent();
	DebyeCorrWindow *wnd = new DebyeCorrWindow(this, f, mainWnd);
	wnd->Show(true);
}


