/**
 * \file 
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <wx/wx.h>
#include <wx/grid.h>
#include <wx/bookctrl.h>
#include <wx/textfile.h>
#include <wx/tokenzr.h>
#include "paramedit.h"
#include "misc.h"

#include "cp_ids.h"

wxBEGIN_EVENT_TABLE(ParametersWindow, wxPanel)	
	EVT_BUTTON(ID_BTN_RESETPAR, ParametersWindow::OnResetPar)
	EVT_BUTTON(ID_BTN_CALC, ParametersWindow::OnEvalPar)
	
	EVT_COMBOBOX(ID_CMB_PARNUM, ParametersWindow::OnCmbParNum)
	EVT_COMBOBOX(ID_CMB_PARNUM_POLY, ParametersWindow::OnCmbParNumPoly)
	EVT_CHECKBOX(ID_CHK_FIXED_NATOMS, ParametersWindow::OnChkFixedNatoms)

	EVT_GRID_CELL_CHANGING(ParametersWindow::OnParamGridCellChanging)
	
	EVT_TEXT(ID_TXT_FIXED_NATOMS, ParametersWindow::OnTextEvent)

	EVT_SIZE(ParametersWindow::OnSize)
wxEND_EVENT_TABLE()


const wxString ParametersWindow::polyDegreeChoices[] = {"-4", "-3", "-2", "-1", "-0.5", "0", "0.5", "1", "2", "3", "4"};
const wxString ParametersWindow::parNumChoices[] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9", "10", "11", "12"};


ParametersWindow::ParametersWindow(wxWindow *parent, std::function <void()> cbfEvalParam)
	: wxPanel(parent, wxID_ANY)
{
	this->cbfEvalParam = cbfEvalParam;

	box = new wxStaticBox(this, wxID_ANY, "Initial approximation");
	wxStaticBoxSizer *boxSizer = new wxStaticBoxSizer(box, wxVERTICAL);


	wxNotebook *nbParams = new wxNotebook(box, wxID_ANY);
	wxPanel *panelEinstein = new wxPanel(nbParams, wxID_ANY);
	wxPanel *panelPoly = new wxPanel(nbParams, wxID_ANY);
	nbParams->AddPage(panelEinstein, "Einstein functions");
	nbParams->AddPage(panelPoly, "Polynomial part");
	boxSizer->Add(nbParams, 1, wxALL | wxEXPAND, 5);

	
	wxBoxSizer *sizerEinstein = new wxBoxSizer(wxVERTICAL);
	panelEinstein->SetSizer(sizerEinstein);

	wxBoxSizer *sizerPoly = new wxBoxSizer(wxVERTICAL);
	panelPoly->SetSizer(sizerPoly);

	// --- Einstein functions panel	
	// a) grid with parameters
	grPar = new wxGridAutoColSize(panelEinstein, ID_GR_PARAM);
	grPar->CreateGrid(3, 2);
	grPar->SetColLabelValue(0, wxT("\u03B1")); // alpha
	grPar->SetColLabelValue(1, wxT("\u03B8 / K")); // theta
	grPar->SetRowLabelSize(40);
	grPar->SetColSize(0, 80);
	for (int i = 0; i < 3; i++) {
		grPar->SetCellValue(i, 0, "1e-6");
		grPar->SetCellValue(i, 1, "1");
	}
	sizerEinstein->Add(grPar, 1, wxALL | wxEXPAND, 5);
	sizerEinstein->SetItemMinSize(grPar, 230, 200);

	// b) Selector of parameters number
	wxBoxSizer *parNumSizer = new wxBoxSizer(wxHORIZONTAL);
	lblParNum = new wxStaticText(panelEinstein, wxID_ANY, "Number of terms");
	parNumSizer->Add(lblParNum, 1, wxALL | wxEXPAND, 5);
	cmbParNum = new wxComboBox(panelEinstein, ID_CMB_PARNUM, "3",
		wxPoint(-1, -1), wxSize(-1, -1),
		WXSIZEOF(parNumChoices), parNumChoices, wxCB_READONLY);
	parNumSizer->Add(cmbParNum, 1, wxALL | wxEXPAND, 5);
	sizerEinstein->Add(parNumSizer, 0, wxALL | wxEXPAND, 5);

	// c) Selector of fixed/varied natoms
	wxBoxSizer *fixNatSizer = new wxBoxSizer(wxHORIZONTAL);
	chkFixedNatoms = new wxCheckBox(panelEinstein, ID_CHK_FIXED_NATOMS,
		wxT("Fixed \u03A3\u03B1")); // Fixed \sum\alpha
	fixNatSizer->Add(chkFixedNatoms, 1, wxALL | wxEXPAND, 5);
	txtFixedNatoms = NULL;
	txtFixedNatoms = new wxTextCtrl(panelEinstein, ID_TXT_FIXED_NATOMS, wxT("1"));
	txtFixedNatoms->Enable(false);
	strFixedNatomsOld = "1";
	fixNatSizer->Add(txtFixedNatoms, 1, wxALL | wxEXPAND, 5);
	sizerEinstein->Add(fixNatSizer, 0, wxALL | wxEXPAND, 5);

	//--- Polynomial parameters
	// a) Grid with parameters and degrees
	grParPoly = new wxGrid(panelPoly, ID_GR_PARAM_POLY);
	grParPoly->CreateGrid(0, 2);
	grParPoly->SetColLabelValue(0, wxT("a"));
	grParPoly->SetColLabelValue(1, wxT("n"));
	grParPoly->SetRowLabelSize(40);
	grParPoly->SetColSize(0, 80);
	sizerPoly->Add(grParPoly, 1, wxALL | wxEXPAND, 5);
	sizerPoly->SetItemMinSize(grParPoly, 230, 200);
	// b) Selector of parameters number
	wxBoxSizer *parNumSizerPoly = new wxBoxSizer(wxHORIZONTAL);
	lblParNumPoly = new wxStaticText(panelPoly, wxID_ANY, "Number of terms");
	parNumSizerPoly->Add(lblParNumPoly, 1, wxALL | wxEXPAND, 5);
	cmbParNumPoly = new wxComboBox(panelPoly, ID_CMB_PARNUM_POLY, "0",
		wxPoint(-1, -1), wxSize(-1, -1),
		WXSIZEOF(parNumChoices), parNumChoices, wxCB_READONLY);
	parNumSizerPoly->Add(cmbParNumPoly, 1, wxALL | wxEXPAND, 5);
	sizerPoly->Add(parNumSizerPoly, 0, wxALL | wxEXPAND, 5);


	//--- Calculate/reset parameters buttons
	wxBoxSizer *parBtnsSizer = new wxBoxSizer(wxHORIZONTAL);
	btnCalc = new wxButton(box, ID_BTN_CALC, "Evaluate parameters");
	btnReset = new wxButton(box, ID_BTN_RESETPAR, "Reset parameters");
	parBtnsSizer->Add(btnCalc, 1, wxALL | wxEXPAND, 5);
	parBtnsSizer->Add(btnReset, 1, wxALL | wxEXPAND, 5);
	parBtnsSizer->SetItemMinSize(btnCalc, 60, 30);
	parBtnsSizer->SetItemMinSize(btnReset, 60, 30);
	boxSizer->Add(parBtnsSizer, 0, wxALL | wxEXPAND, 5);

	// Main sizer
	wxBoxSizer *mainSizer = new wxBoxSizer(wxHORIZONTAL);
	mainSizer->Add(boxSizer, 1, wxALL | wxEXPAND);
	this->SetSizer(mainSizer);
	this->Layout();
	ResizeWxGridCols(grPar);
}

/**
 * \brief Gets the fixed Natoms value or returns -1 if it is varied
 */
double ParametersWindow::GetFixedNatoms(bool *isOk)
{
	if (chkFixedNatoms->IsChecked()) {
		double Natoms = -1;
		bool isOkBuf = true;
		wxString NatomsTxt = txtFixedNatoms->GetValue();
		if (!NatomsTxt.ToDouble(&Natoms)) {
			Natoms = -1;
			wxMessageBox(wxT("Natoms string is not numeric"), "CpFit", wxOK | wxICON_EXCLAMATION);
			isOkBuf = false;
		}
		if (isOk != NULL)
			*isOk = isOkBuf;
		return Natoms;
		
	} else {
		*isOk = true;
		return -1;
	}
}


/**
 * \brief  Reset all parameters values to default (i.e. default initial approximation)
 */
void ParametersWindow::OnResetPar(wxCommandEvent &event)
{
	ResetParameters();
}

/**
 * \brief  Call callback for parameters evaluation
 */
void ParametersWindow::OnEvalPar(wxCommandEvent &event)
{
	cbfEvalParam();
}


/**
 * \brief  Reset all parameters values to default (i.e. default initial approximation)
 */
void ParametersWindow::ResetParameters()
{
	for (int i = 0; i < grPar->GetNumberRows(); i++) {
		grPar->SetCellValue(i, 0, "1e-6");
		grPar->SetCellValue(i, 1, "1");
	}
}

/**
 * \brief Mainly to control editing text in metadata as changing the file
 */
void ParametersWindow::OnTextEvent(wxCommandEvent& event)
{
	wxString NatomsTxt;
	double Natoms;
	switch (event.GetId()) {
	case ID_TXT_FIXED_NATOMS:
		if (txtFixedNatoms != NULL) {
			NatomsTxt = txtFixedNatoms->GetValue();
			if (!NatomsTxt.ToDouble(&Natoms) || Natoms <= 0.01) {
				wxMessageBox(
					wxT("Natoms string is not positive numeric"),
					"CpFit", wxOK | wxICON_EXCLAMATION);
				txtFixedNatoms->SetValue(strFixedNatomsOld);
			} else {
				strFixedNatomsOld = NatomsTxt;
			}
		}
		break;
	}
	event.Skip();
}

/**
 * \brief Parameters numbers (i.e. number of terms) selection (by means of combo box)
 */
void ParametersWindow::OnCmbParNum(wxCommandEvent &event)
{
  	ResizeParamGrid(cmbParNum->GetSelection());
	grPar->SetFocus();
}

/**
 * \brief Parameters numbers in the polynomial (i.e. number of terms) selection (by means of combo box)
 */
void ParametersWindow::OnCmbParNumPoly(wxCommandEvent &event)
{
	ResizePolyParamGrid(cmbParNumPoly->GetSelection());
	grParPoly->SetFocus();
}


/**
 * \brief Enables/disables text fields for Natoms (\sum\alpha)
 */
void ParametersWindow::OnChkFixedNatoms(wxCommandEvent &event)
{
	if (chkFixedNatoms->IsChecked()) {
		txtFixedNatoms->Enable(true);
	} else {
		txtFixedNatoms->Enable(false);
	}
}

/**
 * \brief  Change number of parameters (terms) inside of the grid
 * \param  Number of terms (table rows)
 */
void ParametersWindow::ResizeParamGrid(int numOfRows)
{
	int numOfRows_old = grPar->GetNumberRows();
	if (numOfRows > numOfRows_old) {
		grPar->InsertRows(numOfRows_old, numOfRows - numOfRows_old);
		for (int i = numOfRows_old; i < numOfRows; i++) {
			grPar->SetCellValue(i, 0, "1e-6");
			grPar->SetCellValue(i, 1, "1");
		}
	} else if (numOfRows < numOfRows_old) {
		grPar->DeleteRows(numOfRows, numOfRows_old - numOfRows);
	}
	cmbParNum->SetSelection(numOfRows);
}

/**
 * \brief  Change number of parameters (terms) inside of the polynomial parameters grid
 * \param  Number of terms (table rows)
 */
void ParametersWindow::ResizePolyParamGrid(int numOfRows)
{
	int numOfRows_old = grParPoly->GetNumberRows();
	if (numOfRows > numOfRows_old) {
		grParPoly->InsertRows(numOfRows_old, numOfRows - numOfRows_old);
		for (int i = numOfRows_old; i < numOfRows; i++) {
			grParPoly->SetCellValue(i, 0, "1e-2");
			grParPoly->SetCellValue(i, 1, "1");

			grParPoly->SetCellEditor(i, 1,
				new wxGridCellChoiceEditor(
					WXSIZEOF(ParametersWindow::polyDegreeChoices),
					ParametersWindow::polyDegreeChoices, false));

		}
	} else if (numOfRows < numOfRows_old) {
		grParPoly->DeleteRows(numOfRows, numOfRows_old - numOfRows);
	}
	cmbParNumPoly->SetSelection(numOfRows);
}



/**
 * \brief  Controls validity of user data editing
 */
void ParametersWindow::OnParamGridCellChanging(wxGridEvent& event)
{
	if (event.GetId() == ID_GR_PARAM || event.GetId() == ID_GR_PARAM_POLY) {
		double t;
		if (!event.GetString().ToDouble(&t)) {
			wxMessageBox(strNotNumeric, "CpFit", wxICON_ERROR);
			event.Veto();
		}
	}
	event.Skip();
}

/**
 * \brief  Collect all alpha values from the grid into a vector and return them
 */
std::vector <double> ParametersWindow::ReadAlpha()
{
	int numOfRows = grPar->GetNumberRows();
	std::vector <double> alpha;
	for (int i = 0; i < numOfRows; i++) {
		double a;
		grPar->GetCellValue(i, 0).ToDouble(&a); // TODO:ERROR CHECK
		alpha.push_back(a);
	}
	return alpha;
}

/**
 * \brief  Collect all theta values from the grid into a vector and return them
 */
std::vector <double> ParametersWindow::ReadTheta()
{
	int numOfRows = grPar->GetNumberRows();
	std::vector <double> theta;
	for (int i = 0; i < numOfRows; i++) {
		double t;
		grPar->GetCellValue(i, 1).ToDouble(&t); // TODO:ERROR CHECK
		theta.push_back(t);
	}
	return theta;
}

/**
 * \brief  Collect all polynomial coefficients (a in aT^n) from the grid
 * into a vector and return them.
 */
std::vector <double> ParametersWindow::ReadPolyCoeff()
{
	int numOfRows = grParPoly->GetNumberRows();
	std::vector <double> coeff;
	for (int i = 0; i < numOfRows; i++) {
		double t;
		grParPoly->GetCellValue(i, 0).ToDouble(&t); // TODO:ERROR CHECK
		coeff.push_back(t);
	}
	return coeff;
}

/**
 * \brief  Collect all polynomial degrees (n in aT^n) from the grid
 * into a vector and return them.
 */
std::vector <double> ParametersWindow::ReadPolyDegree()
{
	int numOfRows = grParPoly->GetNumberRows();
	std::vector <double> degree;
	for (int i = 0; i < numOfRows; i++) {
		double t;
		grParPoly->GetCellValue(i, 1).ToDouble(&t); // TODO:ERROR CHECK
		degree.push_back(t);
	}
	return degree;
}



/**
 * \brief  Write alpha values to the table
 *
 * The table will be resized to make sizes of alpha and table grid consistent.
 */
bool ParametersWindow::WriteAlpha(std::vector <double> alpha)
{
	int numOfRows = alpha.size();
	ResizeParamGrid(numOfRows);
	for (int i = 0; i < numOfRows; i++) {
		grPar->SetCellValue(i, 0, wxString::Format("%.14g", alpha[i]));
	}
	return true;
}

/**
 * \brief  Write theta values to the table
 *
 * The table will be resized to make sizes of alpha and table grid consistent.
 */
bool ParametersWindow::WriteTheta(std::vector <double> theta)
{
	int numOfRows = theta.size();
	ResizeParamGrid(numOfRows);
	for (int i = 0; i < numOfRows; i++) {
		grPar->SetCellValue(i, 1, wxString::Format("%.14g", theta[i]));
	}
	return true;
}

bool ParametersWindow::WritePolyCoeff(std::vector <double> coeff)
{
	int numOfRows = coeff.size();
	ResizePolyParamGrid(numOfRows);
	for (int i = 0; i < numOfRows; i++) {
		grParPoly->SetCellValue(i, 0, wxString::Format("%.14g", coeff[i]));
	}
	return true;
}

bool ParametersWindow::WritePolyDegree(std::vector <double> degree)
{
	int numOfRows = degree.size();
	ResizePolyParamGrid(numOfRows);
	for (int i = 0; i < numOfRows; i++) {
		grParPoly->SetCellValue(i, 1, wxString::Format("%.14g", degree[i]));
	}
	return true;
}


/**
 * \brief  Main window resizing
 */
void ParametersWindow::OnSize(wxSizeEvent &event)
{
	ResizeWxGridCols(grPar);
	ResizeWxGridCols(grParPoly);
	event.Skip();
}
