/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <wx/wx.h>
#include <wx/textfile.h>
#include <wx/filefn.h>
#include <wx/grid.h>
#include <wx/dcsvg.h>
#include <wx/dcps.h>
#include <wx/dcbuffer.h>
#include <wx/metafile.h> // !!! WIN32-SPECIFIC
#include <wx/cmndata.h>
#include "datafit.h"
#include "plotcanvas.h"

wxBEGIN_EVENT_TABLE(PlotCanvas, wxWindow)
	EVT_PAINT(PlotCanvas::OnPaint)
	EVT_SIZE(PlotCanvas::OnSize)
	EVT_MOUSEWHEEL(PlotCanvas::OnMouseWheel)
	EVT_LEFT_DCLICK(PlotCanvas::OnLeftMouseDClick)
	EVT_LEFT_DOWN(PlotCanvas::OnLeftMouseButton)
	EVT_LEFT_UP(PlotCanvas::OnLeftMouseButton)
wxEND_EVENT_TABLE()

/**
 * \brief Creates a marker with default settings.
 *
 * Default settings are:
 * * Solid black line (between points) with width 2
 * * Circle yellow marker (with black edge) and radius 5
 */
PlotMarker::PlotMarker()
{
	markerType = MARKER_CIRCLE;
	lineType = LINE_SOLID;

	radius = 5;
	lineWidth = 2;
	lineColor = wxColor(0, 0, 0);
	fillColor = wxColor(255, 255, 0);
}

/**
 * \brief Creates a marker with user-defined settings
 */
PlotMarker::PlotMarker(wxColor lineColor, wxColor fillColor,
	line_t lineType, marker_t markerType,
	int lineWidth, int radius)
{
	this->lineColor = lineColor;
	this->fillColor = fillColor;
	this->lineType = lineType;
	this->markerType = markerType;
	this->lineWidth = lineWidth;
	this->radius = radius;
}

void PlotMarker::Copy(PlotMarker &m)
{
	this->lineColor = m.lineColor;
	this->fillColor = m.fillColor;
	this->lineType = m.lineType;
	this->markerType = m.markerType;
	this->lineWidth = m.lineWidth;
	this->radius = m.radius;
}

void PlotMarker::Draw(wxDC &dc, wxPoint &xyScr, bool initPensAndBrushes)
{
	wxPoint diamondPoly[4] = {
		{-radius, 0},
		{0, radius},
		{radius, 0},
		{0, -radius}};

	if (initPensAndBrushes) {
		dc.SetPen(GetWxPen());
		dc.SetBrush(GetWxBrush());
	}
	switch (markerType) {
	case MARKER_NONE:
		break;
	case MARKER_CIRCLE:
		dc.DrawCircle(xyScr, radius);
		break;
	case MARKER_SQUARE:
		dc.DrawRectangle(xyScr.x - radius, xyScr.y - radius, 2*radius, 2*radius);
		break;
	case MARKER_DIAMOND:
		dc.DrawPolygon(4, diamondPoly, xyScr.x, xyScr.y);
		break;
	}
}


/**
 * \brief  Returns wxWidgets wxPenStyle class example that
 * corresponds to the marker line style
 */
wxPen PlotMarker::GetWxPen()
{
	wxPenStyle penStyle = wxPENSTYLE_INVALID;

	switch(lineType) {
	case LINE_NONE:
		penStyle = wxPENSTYLE_SOLID;
		break;
	case LINE_SOLID:
		penStyle = wxPENSTYLE_SOLID;
		break;
	case LINE_DOT:
		penStyle = wxPENSTYLE_DOT;
		break;
	case LINE_LONG_DASH:
		penStyle = wxPENSTYLE_LONG_DASH;
		break;
	case LINE_SHORT_DASH:
		penStyle = wxPENSTYLE_SHORT_DASH;
		break;
	}

	if (penStyle == wxPENSTYLE_INVALID) {
		return *wxTRANSPARENT_PEN;
	} else {
		return wxPen(lineColor, lineWidth, penStyle);
	}
}

/**
 * \brief Return wxWidgets wxBrush class example that
 * corresponds to the marker filling
 */
wxBrush PlotMarker::GetWxBrush()
{
	return wxBrush(fillColor);
}



//----------------------------------------------------------------------

/**
 * \brief Initialize empty data series
 */
DataSeries::DataSeries() : PlotMarker()
{
	// Nothing to do
}


/**
 * \brief Initializes X-Y data series with user data
 */
DataSeries::DataSeries(const double *x, const double *y, int n) : PlotMarker()
{
	for (int i = 0; i < n; i++) {
		wxRealPoint point(x[i], y[i]);
		xy.push_back(point);
	}
}

/**
 * \brief  Draws dataseries using given PlotCanvas and wxDC classes
 * examples (mainly for internal use)
 */
void DataSeries::Draw(PlotCanvas *pl, wxDC *dc)
{
	if (lineType != LINE_NONE) {
		DrawLines(pl, *dc);
	}
	if (markerType != MARKER_NONE) {
		DrawMarkers(pl, *dc);
	}
}


/**
 * \brief  Low-level drawing of circles (mainly for internal use)
 */
void DataSeries::DrawMarkers(PlotCanvas *pl, wxDC &dc)
{
	dc.SetPen(GetWxPen());
	dc.SetBrush(GetWxBrush());
	std::vector <wxPoint> xyScr = pl->xyToScrXY(xy);
	for (size_t i = 0; i < xy.size(); i++) {
		if (pl->IsPointInsideAxes(xy[i])) {
			PlotMarker::Draw(dc, xyScr[i], false);
		}
	}
}


/**
 * \brief  Low-level drawing of lines (mainly for internal use)
 *
 * The function automatically excludes parts of the line that are outside
 * the visible axes box. If the resolution of the original lines is not enough
 * to cross the boundaries -- autocompletion will be applied (i.e. no white gaps
 * between lines and boundaries at high zooms)
 */
void DataSeries::DrawLines(PlotCanvas *pl, wxDC &dc)
{
	std::vector < std::vector <wxPoint> > lineFragments;
	std::vector <wxPoint> currentFragment;
	std::vector <wxPoint> xyScr = pl->xyToScrXY(xy);
	bool insideFragment = false;
	// Find all visible line fragments (i.e. that are inside
	// the current axes box)
	for (size_t i = 0; i < xy.size(); i++) {
		bool isVisible = pl->IsPointInsideAxes(xy[i]);
		if (!insideFragment && isVisible) {
			// Add initial point (crossing with axes) to make line cross
			// with axes box
			if (i > 0) {
				wxRealPoint crossing = pl->FindLineAxesCrossing(xy[i-1], xy[i]);
				if (crossing.x != INFINITY && crossing.y != INFINITY) {
					currentFragment.push_back(pl->xyToScrXY(crossing));
				}
			}
			// Open the fragment
			insideFragment = true;
			currentFragment.push_back(xyScr[i]);
		} else if (insideFragment) {
			if (isVisible) {
				currentFragment.push_back(xyScr[i]);
			} else {
				// Add final point (crossing with axes) to make line cross
				// with axes box
				wxRealPoint crossing = pl->FindLineAxesCrossing(xy[i-1], xy[i]);
				if (crossing.x != INFINITY && crossing.y != INFINITY) {
					currentFragment.push_back(pl->xyToScrXY(crossing));
				}
				// Close the fragment
				insideFragment = false;
				lineFragments.push_back(currentFragment);
				currentFragment.clear();
			}
		}		
	}
	// Close the fragment that includes the last point (if it exists)
	if (currentFragment.size() > 0) {
		lineFragments.push_back(currentFragment);
		currentFragment.clear();
	}
	// Draw all fragments
	dc.SetPen(GetWxPen());
	for (std::vector< std::vector<wxPoint> >::iterator lf = lineFragments.begin();
	     lf < lineFragments.end(); lf++) {
		std::vector <wxPoint> &vec = lf[0];
		dc.DrawLines(lf->size(), &vec[0]);
	}
}


/**
 * \brief Returns minimal X values
 */
double DataSeries::GetMinX()
{
	double minx = xy[0].x;
	for (size_t i = 1; i < xy.size(); i++) {
		if (xy[i].x < minx) {
			minx = xy[i].x;
		}
	}
	return minx;
}

/**
 * \brief Returns maximal X values
 */
double DataSeries::GetMaxX()
{
	double maxx = xy[0].x;
	for (size_t i = 1; i < xy.size(); i++) {
		if (xy[i].x > maxx) {
			maxx = xy[i].x;
		}
	}
	return maxx;
}

/**
 * \brief Returns minimal Y values
 */
double DataSeries::GetMinY()
{
	double miny = xy[0].y;
	for (size_t i = 1; i < xy.size(); i++) {
		if (xy[i].y < miny) {
			miny = xy[i].y;
		}
	}
	return miny;
}

/**
 * \brief Returns maximal Y values
 */
double DataSeries::GetMaxY()
{
	double maxy = xy[0].y;
	for (size_t i = 1; i < xy.size(); i++) {
		if (xy[i].y > maxy) {
			maxy = xy[i].y;
		}
	}
	return maxy;
}



//-------------------------------------------------------------------------------------------------

/**
 * \brief Initializes canvas with some basic settings (fonts, pens, axes etc.)
 */
PlotCanvas::PlotCanvas(wxWindow *parent, wxWindowID id, const wxPoint &pos, const wxSize &size) : wxWindow(parent, id, pos, size, wxFULL_REPAINT_ON_RESIZE)
{

	gridEnabled = true;
	legendEnabled = false;
	isDragging = false;
	isLegendOnTop = true;
	isLegendOnLeft = true;
	logScaleX = false;

	marginTop = 20;
	marginLeft = 100;
	
	marginBottom = 100;
	marginRight = 20;

	bgColor = wxColor(255,255,255); // White
	axisPen = wxPen(wxColor(0, 0, 0)); // Black thin line
	axisBoxPos = wxPoint(marginLeft, marginTop);
	axisBoxSize = wxSize(size.x - marginLeft - marginRight, size.y-marginTop-marginBottom);
	canvasSize = size;

	axisFont = wxFont(10, wxFONTFAMILY_DEFAULT, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD);
	labelFont = wxFont(12, wxFONTFAMILY_SWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD);
	labelFontItalic = wxFont(12, wxFONTFAMILY_SWISS, wxFONTSTYLE_ITALIC, wxFONTWEIGHT_BOLD);
	legendFont = wxFont(12, wxFONTFAMILY_SWISS, wxFONTSTYLE_NORMAL, wxFONTWEIGHT_BOLD);
	
	gridPen = wxPen(wxColor(0, 0, 0), 1, wxPENSTYLE_LONG_DASH);

	// Axes settings
	xy_min = wxPoint(0, 0); xy_min_default = xy_min;
	xy_max = wxPoint(10, 10); xy_max_default = xy_max;
	tickLength = 10;
}

/**
 * \brief Canvas resizing
 */
void PlotCanvas::OnSize( wxSizeEvent& event )
{
	Resize(event.GetSize());
	event.Skip();
}

/**
 * \brief Mouse wheel scrolling
 */
void PlotCanvas::OnMouseWheel( wxMouseEvent& event )
{
	// Get mouse wheel event data
	int rot = event.GetWheelRotation();
	wxWindowDC dc(this);
	wxPoint pos = event.GetLogicalPosition(dc);
	// If the cursor is inside the axes -- run zoom in/zoom out (scaling) procedure
	if (axisBoxPos.x <= pos.x && pos.x <= axisBoxPos.x + axisBoxSize.x &&
	    axisBoxPos.y <= pos.y && pos.y <= axisBoxPos.y + axisBoxSize.y) {
		// Convert mouse pointer coordinates to logical coordinates
		// a) logical width and height
		double width = xy_max.x - xy_min.x;
		double height = xy_max.y - xy_min.y;
		// b) logical coordiantes		    
		double x = xy_min.x + width * (pos.x - axisBoxPos.x) / axisBoxSize.x;
		double y = xy_max.y - height * (pos.y - axisBoxPos.y) / axisBoxSize.y;
		// Some coefficients for scaling
		double wx = (x - xy_min.x) / width;
		double wy = (y - xy_min.y) / height;
		// Scaling procedure. The main idea of this procedure: point under
		// the mouse cursor must save its screen coordinates during the process
		if (rot > 0) {
			// Zoom in
			xy_min.x = x - wx*(width * 0.95);
			xy_min.y = y - wy*(height * 0.95);
			xy_max.x = x + (1 - wx)*(width * 0.95);
			xy_max.y = y + (1 - wy)*(height * 0.95);
		} else {
			// Zoom out
			xy_min.x = x - wx*(width * 1.05);
			xy_min.y = y - wy*(height * 1.05);
			xy_max.x = x + (1 - wx)*(width * 1.05);
			xy_max.y = y + (1 - wy)*(height * 1.05);
			// If we are out of box -- fit in into box
			if (xy_min.x < xy_min_default.x) {
				xy_min.x = xy_min_default.x;
			}
			if (xy_min.y < xy_min_default.y) {
				xy_min.y = xy_min_default.y;
			}

			if (xy_max.x > xy_max_default.x) {
				xy_max.x = xy_max_default.x;
			}
			if (xy_max.y > xy_max_default.y) {
				xy_max.y = xy_max_default.y;
			}
		}
		Refresh();
	}
	event.Skip();
}

/**
 * \brief  Drag and drop implementation (required for navigation
 * through the plot)
 */
void PlotCanvas::OnLeftMouseButton( wxMouseEvent& event )
{
	wxWindowDC dc(this);

	if (event.ButtonDown() && !isDragging) {
		// "Drag" part
		isDragging = true;
		posDrag = event.GetLogicalPosition(dc);
	} else if (event.ButtonUp() && isDragging) {
		// "Drop" part
		isDragging = false;
		double width = xy_max.x - xy_min.x;
		double height = xy_max.y - xy_min.y;
		// Get drag-n-drop offset and convert it to logical coordinates
		wxPoint posDrop = event.GetLogicalPosition(dc);
		int dx_scr = posDrop.x - posDrag.x;
		int dy_scr = posDrop.y - posDrag.y;
		double dx = -((double) dx_scr) * width / axisBoxSize.x;
		double dy =  ((double) dy_scr) * height / axisBoxSize.y;
		// Change axis box coordinates
		xy_min.x += dx; xy_max.x += dx;
		xy_min.y += dy; xy_max.y += dy;
		// If we are out of box -- fit into box
		// a) x axis
		if (xy_min.x < xy_min_default.x) {
			xy_min.x = xy_min_default.x;
			xy_max.x = xy_min.x + width;
		} else if (xy_max.x > xy_max_default.x) {
			xy_max.x = xy_max_default.x;
			xy_min.x = xy_max.x - width;
		}
		// b) y axis
		if (xy_min.y < xy_min_default.y) {
			xy_min.y = xy_min_default.y;
			xy_max.y = xy_min.y + height;
		} else if (xy_max.y > xy_max_default.y) {
			xy_max.y = xy_max_default.y;
			xy_min.y = xy_max.y - height;
		}
		// Update drawing
		Refresh();
		
	}
	event.Skip();
}

/**
 * @brief  Shows current coordinates of the mouse cursor on double click
 */
void PlotCanvas::OnLeftMouseDClick( wxMouseEvent& event )
{
	wxWindowDC dc(this);
	wxPoint scrXY = event.GetLogicalPosition(dc);	
	wxRealPoint xy = scrXYToxy(scrXY);
	wxString msg = wxString::Format("x = %g, y = %g", xy.x, xy.y);
	wxMessageBox(msg, wxT("CpFit"), wxICON_INFORMATION);
	event.Skip();
}



/**
 * \brief Changes size of axes used during drawing into wxDC.
 *
 * Will be useful for exporting canvas into hi-res bitmaps
 * (not implemented yet).
 */
void PlotCanvas::Resize(wxSize size)
{
	axisBoxSize = wxSize(size.x - marginLeft - marginRight, size.y-marginTop-marginBottom);
	canvasSize = size;
}

/**
 * \brief Redrawing canvas using double buffering.
 */
void PlotCanvas::OnPaint( wxPaintEvent& WXUNUSED(event) )
{
	wxBufferedPaintDC dc(this);
	PaintToDC(dc);
}

/**
 * \brief  Exports canvas content into bitmap file
 * \param fileName  Full file name
 * \param bmpType File type (see wxWidgets documentation)
 * \return true/false -- success/failure
 *
 * Note: hi-resolution DPI in the file is not implemeneted yet!
 * Screen resolution will be used!
 */
bool PlotCanvas::SaveToBitmap(wxString fileName, wxBitmapType bmpType)
{
// TODO: HI-RES IMAGE
//	wxSize oldSize = canvasSize;
//	canvasSize = wxSize(4*600, 4*600);
//	Resize(canvasSize); // TODO: CUSTOMIZABLE SIZE
	wxBitmap bmp(canvasSize);
	wxMemoryDC dc;
	// Draw current plot into a bitmap (in memory)
	dc.SelectObject(bmp);
//	dc.SetLogicalScale(0.01, 0.01);
	PaintToDC(dc);
	dc.SelectObject(wxNullBitmap);
	// Restore the size of axis box
//	canvasSize = oldSize;
//	Resize(canvasSize);
	// Save the bitmap to a file
	return bmp.SaveFile(fileName, bmpType);	
}


/**
 * \brief Export to SVG file
 *
 * Note: buggy on some things in canvas
 */
bool PlotCanvas::SaveToSVG(wxString fileName)
{
	wxSVGFileDC svgDC(fileName, canvasSize.x, canvasSize.y);
	PaintToDC(svgDC);
	return svgDC.IsOk();
}

/**
 * \brief Export to WMF file
 *
 * Note: works only in MS Windows
 */
bool PlotCanvas::SaveToWMF(wxString fileName)
{
	// TODO: HERE IS WIN32-SPECIFIC CODE!
#ifdef __MINGW32__
	wxMetafileDC wmfDC(fileName);
	if (!wmfDC.IsOk()) {
		return false;
	}
	PaintToDC(wmfDC);
	wmfDC.Close();
#else
	wxMessageBox("WMF format is supported only in MS Windows!");
#endif
	return true;
}


/**
 * \brief Export to MATLAB M file (script)
 */
bool PlotCanvas::SaveToM(wxString fileName)
{
	wxTextFile file;
	// Open file for writing (if it exists -- 
	// it will be erased)
	if (wxFileExists(fileName)) {
		if (!wxRemoveFile(fileName)) {
			return false;
		}
	}
	if (!file.Create(fileName)) {
		return false;
	}
	// Generate MATLAB script
	file.AddLine("% CpFit 2D plot export");
	file.AddLine(wxString::Format("%% Number of data series: %d", (int) dataSeries.size()));
	// a) for data series
	file.AddLine("dataSeries = {");
	for (size_t i = 0; i < dataSeries.size(); i++) {
		DataSeries &ser = dataSeries[i];
		file.AddLine(wxString::Format("    %% Series %d", (int) i));
		wxString markerTypeStr = "", markerLineStr = "";
		switch (ser.markerType) {
			case MARKER_NONE:
				markerTypeStr = ""; break;
			case MARKER_CIRCLE:
				markerTypeStr = "o"; break;
			case MARKER_SQUARE:
				markerTypeStr = "s"; break;
			case MARKER_DIAMOND:
				markerTypeStr = "d"; break;
				
		}
		switch (ser.lineType) {
			case LINE_NONE:
				markerLineStr = ""; break;
			case LINE_SOLID:
				markerLineStr = "-"; break;
			case LINE_DOT:
				markerLineStr = ":"; break;
			case LINE_LONG_DASH:
				markerLineStr = "--"; break;
			case LINE_SHORT_DASH:
				markerLineStr = "--"; break;
		}
		file.AddLine(wxString::Format("    struct('Marker', '%s',...", markerTypeStr));
		if (i < legendEntries.size()) {
			file.AddLine(wxString::Format("        'Name', '%s',...", legendEntries[i].name));
		}
		file.AddLine(wxString::Format("        'Line', '%s',...", markerLineStr));
		file.AddLine(wxString::Format("        'LineColor', [%d %d %d],...",
			(int) ser.lineColor.Red(), (int) ser.lineColor.Green(), (int) ser.lineColor.Blue()));
		file.AddLine(wxString::Format("        'FillColor', [%d %d %d],...",
			(int) ser.fillColor.Red(), (int) ser.fillColor.Green(), (int) ser.fillColor.Blue()));
		// Save data
		file.AddLine("    'xy',[...");
		for (size_t j = 0; j < ser.xy.size(); j++) {
			wxRealPoint p = ser.xy[j];
			file.AddLine(wxString::Format("        %g %g", p.x, p.y));
		}
		// Close structure and (if neccessary) cell array
		if (i != dataSeries.size() - 1) {
			file.AddLine("    ]),..."); // struct only
		} else {
			file.AddLine("    ])};"); // struct + cell array
		}
	}	
	// b) for legend entries
	file.AddLine("legendEntries = {...");
	for (size_t i = 0; i < legendEntries.size(); i++) {
		PlotCanvasLegendEntry &le = legendEntries[i];
		file.AddLine(wxString::Format("    struct('Name', '%s', 'Marker', '', 'Line', '')%s", // TODO: EXTRA OUTPUT!
			le.name, (i != legendEntries.size() - 1) ? ",..." : "};"));
	}
	// Write data to file and close it
	file.Write();
	file.Close();
	return true;
}

/**
 * \brief Export to PS (postscript) file
 *
 * Note: buggy on some things in canvas
 */

bool PlotCanvas::SaveToPS(wxString fileName)
{
	wxPrintData pd;
	pd.SetFilename(fileName);
 	pd.SetPaperId(wxPAPER_A4);
	wxPostScriptDC psDC(pd);
//	psDC.SetResolution(600);
	wxSize ppi = psDC.GetPPI();
	psDC.StartDoc(wxEmptyString);
	wxMessageBox(wxString::Format("%d %d", ppi.x, ppi.y), "", 0);
	PaintToDC(psDC);
	wxMessageBox(wxString::Format("%d %d", ppi.x, ppi.y), "", 0);
	psDC.EndDoc();
	return true;
//	return psDC.IsOk();
}

/**
 * \brief Copies content of the canvas into WIN32 Clipboard
 * using WMF/EMF-style vector format.
 *
 * Note: this is Win32-specific code!
 */
bool PlotCanvas::ToClipboard()
{
	// TODO: HERE IS WIN32-SPECIFIC CODE!
#ifdef __MINGW32__
	wxMetafileDC wmfDC;
	if (!wmfDC.IsOk()) {
		return false;
	}
	PaintToDC(wmfDC);
	wxMetafile *wmf = wmfDC.Close();
	wmf->SetClipboard();
#endif
	return true;
}


/**
 * \brief Paints content of the canvas into wxDC context
 * (this is a universal way to draw it at screen, clipboard, file etc.)
 */
void PlotCanvas::PaintToDC(wxDC &dc)
{
	int scrX, scrY;
	dc.GetSize(&scrX, &scrY); // This is the size of the visible area only!
	if (scrX == 0 || scrY == 0) { // For the case of WMF-FILE export (to clipboard)
		scrX = canvasSize.x;
		scrY = canvasSize.y;
	}

	// Background fill
	dc.SetPen( *wxTRANSPARENT_PEN );
	dc.SetBrush(wxBrush(bgColor));
	dc.DrawRectangle(0,0,scrX,scrY);
	// Draw axes box
	dc.SetPen(axisPen);
	dc.DrawRectangle(axisBoxPos, axisBoxSize);
	// Draw ticks
	for (size_t i = 0; i < xTicks.size(); i++) {
		wxString label = wxString::Format("%g", xTicks[i]);
		DrawXTick(dc, xTicks[i], label);
	}
	for (size_t i = 0; i < yTicks.size(); i++) {
		wxString label = wxString::Format("%g", yTicks[i]);
		DrawYTick(dc, yTicks[i], label);
	}
	// Draw data series
	for (size_t i = 0; i < dataSeries.size(); i++) {
		dataSeries[i].Draw(this, &dc);
	}
	// Draw grid
	if (gridEnabled) {
		DrawGrid(dc);
	}
	if (legendEnabled) {
		DrawLegend(dc);
	}
	// Draw labels
	// a) xlabel
	dc.SetFont(labelFont);	
	wxSize strSize = dc.GetTextExtent(xLabel);
	double x0 = axisBoxPos.x + (axisBoxSize.x / 2);
	double y0 = (axisBoxPos.y + axisBoxSize.y);
	dc.DrawLabel(xLabel, wxRect(x0 - strSize.x/2, y0 + strSize.y * 2, strSize.x, strSize.y));
	// b) ylabel
	strSize = dc.GetTextExtent(yLabel);
	x0 = axisBoxPos.x / 2 - strSize.y * 0.75;
	y0 = axisBoxPos.y + axisBoxSize.y / 2 + strSize.x / 2;
	dc.DrawRotatedText(yLabel, x0, y0, 90);
}

/**
 * \brief  Adds new data set to the canvas
 */
void PlotCanvas::Plot(double *x, double *y, int n, PlotMarker marker)
{
	if (n > 0) {
		DataSeries ser = DataSeries(x, y, n);
		ser.Copy(marker);
		dataSeries.push_back(ser);
	}
}

/**
 * \brief  Adds new data set to the canvas
 */
void PlotCanvas::Plot(std::vector <double> x, std::vector <double> y, PlotMarker marker)
{
	if (x.size() == 0 || x.size() != y.size()) {
		return;
	}
	this->Plot(&x[0], &y[0], x.size(), marker);
}

/**
 * \brief  Check if the point is inside axes box (logical coordinates
 * are used)
 */
bool PlotCanvas::IsPointInsideAxes(wxRealPoint xy)
{
	if (xy.x < xy_min.x - 1e-15 || xy.x > xy_max.x + 1e-15) {
		return false;
	}
	if (xy.y < xy_min.y - 1e-15 || xy.y > xy_max.y + 1e-15) {
		return false;
	}
	return true;
}

wxRealPoint PlotCanvas::FindLineAxesCrossing(wxRealPoint pointA, wxRealPoint pointB)
{
	// Point with crossing
	wxRealPoint crossing;
	// Find Ax + By + C = 0 equation of the line
	double A = pointA.y - pointB.y;
	double B = pointB.x - pointA.x;
	double C = pointA.x * pointB.y - pointB.x * pointA.y;
	// Crossing with x axes
	if (fabs(A) > 1e-12) { // Line mustn't be parallel to x
		// Boundaries for the line
		double x1 = std::min(pointA.x, pointB.x);
		double x2 = std::max(pointA.x, pointB.x);
		// Left x axis
		if (x1 <= xy_min.x && xy_min.x <= x2) {
			crossing.x = xy_min.x;
			crossing.y = -(A * xy_min.x + C) / B;
			if (IsPointInsideAxes(crossing)) {
				return crossing;
			}
		}
		// Right x axis
		if (x1 <= xy_max.x && xy_max.x <= x2) {
			crossing.x = xy_max.x;
			crossing.y = -(A * xy_max.x + C) / B;
			if (IsPointInsideAxes(crossing)) {
				return crossing;
			}
		}
	}	
	// Crossing with y axes
	if (fabs(B) > 1e-12) { // Line mustn't be parallel to y
		// Boundaries for the line
		double y1 = std::min(pointA.y, pointB.y);
		double y2 = std::max(pointA.y, pointB.y);
		// Lower y axis
		if (y1 <= xy_min.y && xy_min.y <= y2) {
			crossing.y = xy_min.y;
			crossing.x = -(B * xy_min.y + C) / A;
			if (IsPointInsideAxes(crossing)) {
				return crossing;
			}
		}
		// Upper y axis
		if (y1 <= xy_max.y && xy_max.y <= y2) {
			crossing.y = xy_max.y;
			crossing.x = -(B * xy_max.y + C) / A;
			if (IsPointInsideAxes(crossing)) {
				return crossing;
			}
		}
	}
	// Degenerate case: no crossings
	return wxRealPoint(INFINITY, INFINITY);
}

/**
 * \brief  Converts logical coordinates to physical
 */
wxPoint PlotCanvas::xyToScrXY(wxRealPoint xy)
{
	double xmax = xy_max.x, xmin = xy_min.x;
	double ymax = xy_max.y, ymin = xy_min.y;
	double x = xy.x, y = xy.y;

	if (logScaleX) {
		xmin = (xmin < 1e-16) ? -16 : log10(xmin);
		xmax = log10(xmax);
		x = log10(x);
	}

	wxPoint xyScr;
	double xScaling = (double) axisBoxSize.x / (xmax - xmin);
	double yScaling = (double) axisBoxSize.y / (ymax - ymin);
	xyScr.x = axisBoxPos.x + (int) ((x - xmin) * xScaling);
	xyScr.y = axisBoxPos.y + axisBoxSize.y - (int) ((y - ymin) * yScaling);
	return xyScr;
}

/**
 * \brief  Converts logical coordinates to physical
 */
std::vector <wxPoint> PlotCanvas::xyToScrXY(std::vector <wxRealPoint> xy)
{
	std::vector <wxPoint> xyScr(xy.size());
	for (size_t i = 0; i < xy.size(); i++) {
		xyScr[i] = xyToScrXY(xy[i]);
	}	
	return xyScr;	
}

/**
 * \brief  Converts physical coordinates to logical
 */
wxRealPoint PlotCanvas::scrXYToxy(wxPoint xyScr)
{
	double xmax = xy_max.x, xmin = xy_min.x;
	double ymax = xy_max.y, ymin = xy_min.y;

	if (logScaleX) {
		xmin = (xmin < 1e-16) ? -16 : log10(xmin);
		xmax = log10(xmax);
	}

	wxRealPoint xy;
	double xScaling = (xmax - xmin) / axisBoxSize.x;
	double yScaling = (ymax - ymin) / axisBoxSize.y;
	xy.x = xmin + (xyScr.x - axisBoxPos.x) * xScaling;
	xy.y = ymin + (axisBoxPos.y + axisBoxSize.y - xyScr.y) * yScaling;

	if (logScaleX) {
		xy.x = pow(10, xy.x);
	}

	return xy;
}

/**
 * \brief Draws X tick using given coordinate and label
 * \param x  x logical coordinate
 * \param label text label near the tick
 */
void PlotCanvas::DrawXTick(wxDC &dc, double x, wxString label)
{
	wxRealPoint xy_tmp(x, 0);
	wxPoint xyscr_tmp = xyToScrXY(xy_tmp);
	double x0 = xyscr_tmp.x;
	
	// a) Tick
	if (xy_min.x + 1e-15 < x && x < xy_max.x - 1e-15) {
		dc.SetPen(axisPen);
		double y0 = axisBoxPos.y;
		dc.DrawLine(wxPoint(x0, y0), wxPoint(x0, y0 + tickLength));
		y0 = axisBoxPos.y + axisBoxSize.y - 1;
		dc.DrawLine(wxPoint(x0, y0), wxPoint(x0, y0 - tickLength));
	}
	// b) Label
	if (xy_min.x - 1e-15 < x && x < xy_max.x + 1e-15) {
		wxSize strSize = dc.GetTextExtent(label);
		dc.SetFont(axisFont);
		dc.DrawLabel(label, wxRect(x0 - strSize.x/2, axisBoxPos.y + axisBoxSize.y - 1 + tickLength, strSize.x, strSize.y));
	}	
}

/**
 * \brief Draws Y tick using given coordinate and label
 * \param y  y logical coordinate
 * \param label text label near the tick
 */
void PlotCanvas::DrawYTick(wxDC &dc, double y, wxString label)
{
	wxRealPoint xy_tmp(0, y);
	wxPoint xyscr_tmp = xyToScrXY(xy_tmp);
	double y0 = xyscr_tmp.y;
	// Tick
	if (xy_min.y + 1e-15 < y && y < xy_max.y - 1e-15) {
		dc.SetPen(axisPen);
		double x0 = axisBoxPos.x;
		dc.DrawLine(wxPoint(x0, y0), wxPoint(x0 + tickLength, y0));
		x0 = axisBoxPos.x + axisBoxSize.x - 1;
		dc.DrawLine(wxPoint(x0, y0), wxPoint(x0 - tickLength, y0));
	}
	// Label from the left side of the left Y axis
	if (xy_min.y - 1e-15 < y && y < xy_max.y + 1e-15) {
		wxSize strSize = dc.GetTextExtent(label);
		dc.SetFont(axisFont);
		dc.DrawLabel(label, wxRect(axisBoxPos.x - tickLength - strSize.x, y0 - strSize.y/2, strSize.x, strSize.y));
	}
}

/**
 * \brief  Enables the grid on 2D plot
 */
void PlotCanvas::EnableGrid()
{
	gridEnabled = true;
	Refresh();
}

/**
 * \brief  Disables the grid on 2D plot
 */
void PlotCanvas::DisableGrid()
{
	gridEnabled = false;
	Refresh();
}

/**
 * \brief Enables (makes visible) the legend on 2D plot
 */
void PlotCanvas::EnableLegend()
{
	legendEnabled = true;
	Refresh();
}

/**
 * \brief Disables (makes invisible) the legend on 2D plot
 */
void PlotCanvas::DisableLegend()
{
	legendEnabled = false;
	Refresh();
}

/**
 * \brief  Draws the grid on 2D plot
 *
 * Precalculated X and Y ticks values are used
 */
void PlotCanvas::DrawGrid(wxDC &dc)
{
	dc.SetPen(gridPen);
	// Horizontal lines
	double y1 = axisBoxPos.y;
	double y2 = y1 + axisBoxSize.y - 1;
	for (size_t i = 0; i < xTicks.size(); i++) {
		double x = xTicks[i];
		if (xy_min.x + 1e-15 < x && x < xy_max.x - 1e-15) {
			wxPoint xyscr_tmp = xyToScrXY(wxRealPoint(x, 0));
			dc.DrawLine(wxPoint(xyscr_tmp.x, y1), wxPoint(xyscr_tmp.x, y2));
		}
	}
	// Vertical lines
	double x1 = axisBoxPos.x;
	double x2 = x1 + axisBoxSize.x - 1;
	for (size_t i = 0; i < yTicks.size(); i++) {
		double y = yTicks[i];
		if (xy_min.y + 1e-15 < y && y < xy_max.y - 1e-15) {
			wxPoint xyscr_tmp = xyToScrXY(wxRealPoint(0, y));
			dc.DrawLine(wxPoint(x1, xyscr_tmp.y), wxPoint(x2, xyscr_tmp.y));
		}
	}	
}


double PlotCanvas::DecimalRound(double val, int ndigits, int nmax = -10000, int isFloor = 1)
{
	if (val == 0)
		return 0;
	
	int n = (nmax == -10000) ? floor(log10(val)) : nmax;
	double m = val / pow(10, n);
	double m_rnd;

	if (isFloor) {		
		m_rnd = floor(m * pow(10, ndigits - 1)) / pow(10, ndigits - 1);
	} else {
		m_rnd = ceil(m * pow(10, ndigits - 1)) / pow(10, ndigits - 1);
	}
	
	return m_rnd * pow(10, n);
}

/**
 * \brief Checks if val/10^n can be presented as 1.0, 1.5, 2.0, 2.5, 5.0
 */
bool PlotCanvas::IsGoodAxisStep(double val)
{
	double logm = log10(val) - floor(log10(val));
	double logm0[] = {log10(1.0), log10(1.5), log10(2.0), log10(2.5), log10(5.0), -1e10};
	
	for (double *plogm0 = logm0; *plogm0 != -1e10; plogm0++) {
		if (fabs(*plogm0 - logm) < 1e-6) {
			return true;
		}
	}
	return false;
}

/**
 * \brief Sets axes limits manually and without refitting axes ticks, labels etc.
 *
 */
bool PlotCanvas::SetAxesBox(double xmin, double xmax, double ymin, double ymax)
{
	if (xmin >= xmax || ymin >= ymax) {
		return false;
	}
	xy_min.x = xmin;
	xy_max.x = xmax;
	xy_min.y = ymin;
	xy_max.y = ymax;
	xy_min_default = xy_min;
	xy_max_default = xy_max;
	return true;
}


void PlotCanvas::FitXAxis()
{
	double xmin = 1e200, xmax = -1e200;
	for (size_t i = 0; i < dataSeries.size(); i++)
	{		
		double xmin_i = dataSeries[i].GetMinX();
		double xmax_i = dataSeries[i].GetMaxX();
		xmin = (xmin_i < xmin) ? xmin_i : xmin;
		xmax = (xmax_i > xmax) ? xmax_i : xmax;
	}
	// Process situation with the equal values of borders
	if (fabs(xmax - xmin) < 1e-15) {
		xmax++;
		xmin--;
	}
	// Round axes limits
	int ndig_x = floor(log10( std::max(fabs(xmax), fabs(xmin)) ));
	xy_min.x = DecimalRound(xmin, 1, ndig_x, 1);
	xy_max.x = DecimalRound(xmax, 1, ndig_x, 0);
	// Find optimal number of ticks	for x axis
	int nticks_x;
	for (nticks_x = 5; nticks_x < 20; nticks_x++) {
		double x = fabs((xy_max.x - xy_min.x) / (nticks_x - 1.0));
		if (IsGoodAxisStep(x)) {
			break;
		}
	}
	// Ticks
	xTicks.clear();
	for (int i = 0; i < nticks_x; i++) {
		double x = xy_min.x + i / (nticks_x - 1.0) * (xy_max.x - xy_min.x);
		xTicks.push_back(x);
	}
}

void PlotCanvas::FitXAxisLogScale()
{
	double xmin = 1e200, xmax = -1e200;
	for (size_t i = 0; i < dataSeries.size(); i++)
	{		
		double xmin_i = dataSeries[i].GetMinX();
		double xmax_i = dataSeries[i].GetMaxX();
		xmin = (xmin_i < xmin) ? xmin_i : xmin;
		xmax = (xmax_i > xmax) ? xmax_i : xmax;
	}
	// Process situation with the equal values of borders
	if (fabs(xmax - xmin) < 1e-15) {
		xmax++;
		xmin--;
	}
	// Round axes limits using logarithmic scale
	if (xmin < 1e-16)
		xmin = 1e-16;
    if (xmax < 1e-16)
    	xmax = 1e-16;
    int xmin_lg = floor(log10(xmin));
	xy_min.x = pow(10, xmin_lg);
	int xmax_lg = ceil(log10(xmax));
	xy_max.x = pow(10, xmax_lg);
	// Find optimal number of ticks	for x axis
	int lg_step = 1;
	// Ticks
	xTicks.clear();
	for (int i = xmin_lg; i <= xmax_lg; i += lg_step) {
		double x = pow(10, i);
		xTicks.push_back(x);
	}
}

void PlotCanvas::FitYAxis()
{
	double ymin = 1e200, ymax = -1e200;
	for (size_t i = 0; i < dataSeries.size(); i++)
	{		
		double ymin_i = dataSeries[i].GetMinY();
		double ymax_i = dataSeries[i].GetMaxY();
		ymin = (ymin_i < ymin) ? ymin_i : ymin;
		ymax = (ymax_i > ymax) ? ymax_i : ymax;
	}
	// Process situation with the equal values of borders
	if (fabs(ymax - ymin) < 1e-15) {
		ymax++;
		ymin--;
	}
	// Round axes limits
	int ndig_y = floor(log10( std::max(fabs(ymax), fabs(ymin)) ));
	xy_min.y = DecimalRound(ymin, 1, ndig_y, 1);
	xy_max.y = DecimalRound(ymax, 1, ndig_y, 0);
	// Find optimal number of ticks	for y axis
	int nticks_y;
	for (nticks_y = 5; nticks_y < 20; nticks_y++) {
		double y = fabs((xy_max.y - xy_min.y) / (nticks_y - 1.0));
		if (IsGoodAxisStep(y)) {
			break;
		}
	}
	// Ticks
	yTicks.clear();
	for (int i = 0; i < nticks_y; i++) {
		double y = xy_min.y + i / (nticks_y - 1.0) * (xy_max.y - xy_min.y);
		yTicks.push_back(y);
	}
}

/**
 * \brief Finds optimal axes boundaries, X and Y ticks steps.
 *
 * It initializes some internal class variables: xy_min, xy_max,
 * xTicks, yTicks that will be used by another member functions.
 */
void PlotCanvas::FitAxes()
{
	if (dataSeries.size() == 0) {
		xy_min.x = 0; xy_max.x = 1;
		xy_min.y = 0; xy_max.y = 1;
		xy_min_default = xy_min;
		xy_max_default = xy_max;
		return;
	}

	if (logScaleX) {
		FitXAxisLogScale();
	} else {
		FitXAxis();
	}
	FitYAxis();

	xy_min_default = xy_min;
	xy_max_default = xy_max;
}

/**
 * \brief Calculates label rectangle
 */
wxRect CalcLabelRect(wxDC &dc, wxString text)
{
	wxRect ans; // TODO: DO SOMETHING USEFUL!!!
	return ans;
}


/**
 * \brief Returns minimal X on logical axes
 */
double PlotCanvas::GetAxesMinX()
{
	return xy_min.x;
}

/**
 * \brief Returns maximal X on logical axes
 */
double PlotCanvas::GetAxesMaxX()
{
	return xy_max.x;
}

/**
 * \brief Enables/disables logarithmic scaling for X axis
 */
void PlotCanvas::SetLogScaleX(bool enable)
{
	logScaleX = enable;
	FitAxes();
	Refresh();
}


/**
 * \brief  Sets X axis text label
 */
void PlotCanvas::SetXLabel(wxString xLabel)
{
	this->xLabel = xLabel;
}

/**
 * \brief  Sets Y axis text label
 */
void PlotCanvas::SetYLabel(wxString yLabel)
{
	this->yLabel = yLabel;
}

/**
 * \brief  Returns current X axis text label
 */
wxString PlotCanvas::GetXLabel()
{
	return xLabel;
}

/**
 * \brief  Returns current Y axis text label
 */
wxString PlotCanvas::GetYLabel()
{
	return yLabel;
}

int PlotCanvas::GetNumOfDataSeries()
{
	return dataSeries.size();
}

DataSeries PlotCanvas::GetDataSeries(int ind)
{
	int nseries = dataSeries.size();
	if (ind >= 0 && ind < nseries) {
		return dataSeries[ind];
	} else {
		DataSeries ans;
		return ans;		
	}
}



template <class S>
bool PlotCanvas::SetLegendTemplate(std::vector <S> &seriesNames, std::vector <PlotMarker> &seriesMarkers, PlotMarker &defaultMarker)
{
	size_t numOfMarkers = seriesMarkers.size();
	legendEntries.clear();
	for (size_t i = 0; i < seriesNames.size(); i++) {
		PlotMarker marker = (numOfMarkers == 0) ? defaultMarker : (seriesMarkers[i % numOfMarkers]);
		legendEntries.push_back(PlotCanvasLegendEntry(seriesNames[i], marker));
	}
	legendEnabled = true;
	return true;
}



/**
 * \brief Defines legend
 */
bool PlotCanvas::SetLegend(std::vector <std::string> &seriesNames, std::vector <PlotMarker> &seriesMarkers, PlotMarker &defaultMarker)
{
	return SetLegendTemplate <std::string>(seriesNames, seriesMarkers, defaultMarker);
}

/**
 * \brief Defines legend
 */
bool PlotCanvas::SetLegend(std::vector <wxString> &seriesNames, std::vector <PlotMarker> &seriesMarkers, PlotMarker &defaultMarker)
{
	return SetLegendTemplate <wxString>(seriesNames, seriesMarkers, defaultMarker);
}

/**
 * \brief Set legend position at the top
 */
bool PlotCanvas::SetLegendPosition(bool isOnTop, bool isOnLeft)
{
	isLegendOnTop = isOnTop;
	isLegendOnLeft = isOnLeft;
	Refresh();
	return true;
}

/**
 * \brief Draws legend defined by PlotCanvas::SetLegend member function
 */
void PlotCanvas::DrawLegend(wxDC &dc)
{
	// Define text size (for legend entries)
	int numOfEntries = legendEntries.size();
	int maxTextWidth = 0, maxTextHeight = 0;
	for (int i = 0; i < numOfEntries; i++) {
		wxCoord w, h;
		dc.GetTextExtent(legendEntries[i].name, &w, &h, NULL, NULL, &legendFont);
		if (w > maxTextWidth) {
			maxTextWidth = w;
		}
		if (h > maxTextHeight) {
			maxTextHeight = h;
		}
	}
	// Define legend box sizes
	int w = maxTextWidth + 50;
	int h = maxTextHeight * numOfEntries * 1.5 + 20;
	int x = (isLegendOnLeft) ? (axisBoxPos.x + 10) : (axisBoxPos.x + axisBoxSize.x - w - 10);
	int y = (isLegendOnTop) ? (axisBoxPos.y + 10) : (axisBoxPos.y + axisBoxSize.y - h - 10);

	// Draw legend box	
	// a) rectangle
	dc.SetPen(axisPen);
	dc.SetBrush(wxBrush(bgColor));	
	dc.DrawRectangle(x,y,w,h);
	// b) text and markers
	dc.SetFont(legendFont);
	for (int i = 0; i < numOfEntries; i++) {
		wxSize strSize = dc.GetTextExtent(legendEntries[i].name);
		double x0 = x + 30;
		double y0 = y + 10 + maxTextHeight * i * 3 / 2;
		dc.DrawLabel(legendEntries[i].name, wxRect(x0, y0, strSize.x, strSize.y));
		wxPoint mxy = wxPoint(x + 15, y0 + maxTextHeight / 2);
		legendEntries[i].marker.Draw(dc, mxy);
	}
}


PlotCanvasLegendEntry::PlotCanvasLegendEntry()
{
}

PlotCanvasLegendEntry::PlotCanvasLegendEntry(wxString name, PlotMarker marker)
{
	this->name = name;
	this->marker = marker;
}

