/**
 * \file 
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */


#include "tdmodel.h"
#include <math.h>
#include <algorithm>
#include <float.h>

TDBaseLineFuncParams::TDBaseLineFuncParams()
{
	isNatomsFixed = false;
}

TDBaseLineFuncParams::TDBaseLineFuncParams(size_t nAlphaTheta, size_t nPolyCoeffs)
{
	alpha.resize(nAlphaTheta, 0);
	theta.resize(nAlphaTheta, 0);
	polyCoeff.resize(nPolyCoeffs, 0);
	polyDegree.resize(nPolyCoeffs, 0);
}



bool TDBaseLineFuncParams::SetAlphaTheta(const std::vector <double> &alpha, const std::vector <double> &theta)
{
	if (alpha.size() != theta.size()) {
		return false;
	} else {
		this->alpha = alpha;
		this->theta = theta;
		return true;
	}
}

bool TDBaseLineFuncParams::SetPolyCoeffs(const std::vector <double> &coeff, const std::vector <double> &degree)
{
	if (coeff.size() != degree.size()) {
		return false;
	} else {
		this->polyCoeff = coeff;
		this->polyDegree = degree;
		return true;
	}
}

/**
 * \brief Converts optimizable model parameters to the vector.
 * \return [alpha_1, theta_1, ..., alpha_m, theta_m, a_1, ..., a_n]
 */
std::vector <double> TDBaseLineFuncParams::Serialize()
{
	size_t nterms = alpha.size();
	size_t npolyterms = polyCoeff.size();
	size_t pos = 0;

	std::vector <double> param(nterms*2 + npolyterms);
	for (size_t i = 0; i < nterms; i++) {
		param[pos++] = alpha[i];
		param[pos++] = theta[i];
	}
	for (size_t i = 0; i < npolyterms; i++) {
		param[pos++] = polyCoeff[i];
	}
	return param;
}

/**
 * \brief Load alpha, theta and polyCoeff parameters value from the double vector,
 * i.e. deserialize it.
 * \param param [alpha_1, theta_1, ..., alpha_m, theta_m, a_1, ..., a_n]
 */
bool TDBaseLineFuncParams::Deserialize(const std::vector <double> &param)
{
	size_t nterms = alpha.size();
	size_t npolyterms = polyCoeff.size();
	if (param.size() != 2*nterms + npolyterms) {
		return false;
	}
	for (size_t i = 0; i < nterms; i++) {
		alpha[i] = param[2*i];
		theta[i] = param[2*i + 1];
	}
	for (size_t i = 0; i < npolyterms; i++) {
		polyCoeff[i] = param[2*nterms + i];
	}
	return true;
}


/**
 * \brief Converts optimizable model parameters to the vector.
 * \return [theta_1, ksi_1, theta_2, ksi_2, ..., theta_{m-1}, ksi_{m-1}, theta_m,  a_1, ..., a_n]
 */
std::vector <double> TDBaseLineFuncParams::SerializeFixNatoms()
{
	size_t nterms = alpha.size();
	size_t npolyterms = polyCoeff.size();	
	size_t nparam = 2*nterms + npolyterms;
	// Convert alpha to ksi
	std::vector <double> ksi = AlphaToKsi(alpha);
	// Make parameters
	std::vector <double> ksiparam(nparam - 1);
	// Fill ksi/theta parameters
	for (size_t i = 0, j = 1; i < nterms - 1; i++, j += 2) {
		ksiparam[j] = ksi[i];
	}
	for (size_t i = 0, j = 0; i < nterms; i++, j += 2) {
		ksiparam[j] = theta[i];
	}
	// Fill poly parameters
	for (size_t i = 0; i < npolyterms; i++) {
		ksiparam[2*nterms - 1 + i] = polyCoeff[i];
	}
	// Return parameters
	return ksiparam;
}

bool TDBaseLineFuncParams::DeserializeFixNatoms(const std::vector <double> &ksiparam)
{
	size_t nterms = alpha.size();
	size_t npolyterms = polyCoeff.size();	
	size_t nparam = 2*nterms + npolyterms;

	if (ksiparam.size() != nparam - 1) {
		return false;
	}
	// Deserialize ksi and theta
	std::vector <double> ksi(nterms - 1);
	for (size_t i = 1, j = 0; j < nterms - 1; i += 2, j++) {
		ksi[j] = ksiparam[i];
	}
	alpha = KsiToAlpha(ksi);
	for (size_t i = 0, j = 0; j < nterms; i += 2, j++) {
		theta[j] = ksiparam[i];
	}
	// Deserialize polyCoeff
	for (size_t i = 0; i < npolyterms; i++) {
		polyCoeff[i] = ksiparam[2*nterms - 1 + i];
	}
	return true;
}



TDBaseLineFuncParams TDBaseLineFuncParams::GetLowerBoundaries()
{
	TDBaseLineFuncParams lb = *this;
	for (size_t i = 0; i < alpha.size(); i++) {
		lb.alpha[i] = 0;
		lb.theta[i] = 0;
	}
	for (size_t i = 0; i < polyCoeff.size(); i++) {
		lb.polyCoeff[i] = 0;
	}
	return lb;
}

TDBaseLineFuncParams TDBaseLineFuncParams::GetUpperBoundaries()
{
	TDBaseLineFuncParams ub = *this;
	for (size_t i = 0; i < alpha.size(); i++) {
		ub.alpha[i] = 5000;
		ub.theta[i] = 1e6;
	}
	for (size_t i = 0; i < polyCoeff.size(); i++) {
		ub.polyCoeff[i] = 1e10;
	}
	return ub;
}



size_t TDBaseLineFuncParams::NumParams()
{
	size_t nterms = alpha.size();
	size_t npolyterms = polyCoeff.size();
	return 2*nterms + npolyterms;
}

size_t TDBaseLineFuncParams::NumAlphaThetaTerms()
{
	return alpha.size();
}


	
/**
 * \brief  Checks if all parameters values (param vector) are positive
 */
bool TDBaseLineFuncParams::IsAllAlphaThetaPositive()
{
	int isPositive = true;
	for (size_t i = 0; i < alpha.size(); i++) {
		if (alpha[i] < 0 || theta[i] < 0) {
			isPositive = false;
			break;
		}
	}
	return isPositive;
}

void TDBaseLineFuncParams::AddTerm(double alpha, double theta)
{
	this->alpha.push_back(alpha);
	this->theta.push_back(theta);
}

/**
 * \brief Sets to zero all varied parameters, i.e. alpha, theta and polyCoeff
 */
void TDBaseLineFuncParams::SetZeros()
{
	for (size_t i = 0; i < alpha.size(); i++) {
		alpha[i] = 0.0;
		theta[i] = 0.0;
	}
	for (size_t i = 0; i < polyCoeff.size(); i++) {
		polyCoeff[i] = 0.0;
	}		
}

/**
 * \brief  Check if all parameters are statistically significant
 * (t-distribution based criteria)
 */
bool TDBaseLineFuncParams::IsAllParamsSignificant(const TDBaseLineFuncParams &dparam)
{
	for (size_t i = 0; i < alpha.size(); i++) {
		if (fabs(alpha[i]) <= dparam.alpha[i] || fabs(theta[i]) <= dparam.theta[i]) {
			return false; 
		}
	}
	for (size_t i = 0; i < polyCoeff.size(); i++) {
		if (fabs(polyCoeff[i]) <= dparam.polyCoeff[i]) {
			return false;
		}
	}
	return true;
}


/**
 * \brief Converts &ksi; vector to &alpha; vector.
 * \details It uses the next formula:
 * - \f$\alpha_1 = N_{atoms}(1-\xi_1)\f$
 * - \f$\alpha_2 = N_{atoms}(1-\xi_2)\xi_1\f$
 * - \f$\alpha_i = N_{atoms}(1-\xi_i)\prod_{k=1}^{i-1} \xi_k \f$ where \f$1<i<m\f$
 * - \f$\alpha_m = N_{atoms}\prod_{k=1}^m \xi_k \f$
 */
std::vector <double> TDBaseLineFuncParams::KsiToAlpha(const std::vector <double> &ksi)
{
	int nksi = ksi.size();
	std::vector <double> alpha(nksi + 1);
	if (nksi == 0) {
		alpha[0] = Natoms;
	} else {
		// alpha_1 ... alpha_(m-1)
		for (int i = 0; i < nksi; i++) {
			double alphai = Natoms;
			for (int j = 0; j < i; j++) {
				alphai *= ksi[j];
			}
			alphai *= (1 - ksi[i]);
			alpha[i] = alphai;
		}
		// alpha_m
		double alphai = Natoms;
		for (int j = 0; j < nksi; j++) {
			alphai *= ksi[j];
		}
		alpha[nksi] = alphai;
	}
	return alpha;
}

/**
 * \brief Converts &alpha; vector to &ksi; vector.
 * \details It uses the next formula:
 * - \f$\xi_1 = 1 - \frac{\alpha_1}{N_{atoms}} \f$
 * - \f$\xi_2 = 1 - \frac{\alpha_2}{N_{atoms}\xi_1}\f$
 * - \f$\xi_i = 1 - \frac{\alpha_i}{N_{atoms}\prod_{k=1}^{i-1}\xi_k} \f$
 */
std::vector <double> TDBaseLineFuncParams::AlphaToKsi(const std::vector <double> &alpha)
{
	size_t m = alpha.size();
	std::vector <double> ksi(m - 1);
	if (ksi.size() != 0) {
		// Find sum of alphas to renorm them
		double alpha_sum = 0.0;
		for (size_t i = 0; i < alpha.size(); i++) {
			alpha_sum += alpha[i];
		}		
		// Recalculate alpha to ksi
		for (size_t i = 0; i < m - 1; i++) {
			double c = alpha_sum; // Instead of Natoms
			for (size_t j = 0; j < i; j++) {
				c *= ksi[j];
			}
			ksi[i] = 1 - alpha[i] / c;
		}
	}
	return ksi;
}

std::vector <double> TDBaseLineFuncParams::GetKsi()
{
	return AlphaToKsi(alpha);
}

/**
 * \brief Sorts &alpha;-&theta; parameters pairs by &theta; values.
 */
void TDBaseLineFuncParams::SortParameters()
{
	size_t nterms = alpha.size();
	std::vector <std::pair <double, double>> alth(nterms);
	for (size_t i = 0; i < nterms; i++) {
		std::pair <double, double> alth_i;
		alth_i.first = alpha[i];
		alth_i.second = theta[i];
		alth[i] = alth_i;
	}
	auto cmpfunc = [] (std::pair <double, double> a, std::pair <double, double> b) -> bool {
		return a.second > b.second;
	};
	std::sort(alth.begin(), alth.end(), cmpfunc);

	for (size_t i = 0; i < nterms; i++) {
		alpha[i] = alth[i].first;
		theta[i] = alth[i].second;
	}
}




/**
 * \brief Signum function
 */
double sign(double x)
{
	if (x == 0) {
		return 0;
	} else if (x > 0) {
		return 1;
	} else {
		return -1;
	}
}


double bisection_method(double a, double b, double (*func)(double, void*), void *fdata)
{
	// Initialize variables	
	double tA = a, tB = b, tC = (a + b) / 2;
	double FA, FB;
	// Find boundaries for bisection method
	do {	
		tA *= 2;
		tB *= 2;
		FA = func(tA, fdata);
		FB = func(tB, fdata);
		if (FA == 0) {
			return tA;
		}
		if (FB == 0) {
			return tB;
		}
	} while (sign(FA) * sign(FB) > 0);
	// Bisection method iterations
	int iter = 0;
	while (fabs(FB - FA) > DBL_EPSILON && iter < 10000) {
		iter++;
		tC = tA + 0.5*(tB - tA);
		double FC = func(tC, fdata);

		if (FC == 0) {
			return tC;
		}

		if (sign(FA)*sign(FC) < 0) {
			tB = tC; 
			FB = FC;
		}
		else if (sign(FC)*sign(FB) < 0) {
			tA = tC;
			FA = FC;
		}
	}
	return tC;
}

static double erf_env(double t, void *fdata)
{
	double x = *((double *) fdata);	
	return erf(t) - x;
}



/**
 * \brief Inverse error function
 */
double erf_inv(double x)
{
	if (x <= -10) {
		return -1.0;
	} else if (x >= 10) {
		return 1.0;
	} else {
		return bisection_method(-10, 10, erf_env, &x);
	}
}


/**
 * \brief Cumulative t-distribution function
 * \param x Cumulative distribution function argument
 * \param f Degrees of freedom
 *
 * The function implementation is based on the next article:
 * Hubert J Chen. On approximations to the inverse student's t-distribution function //
 * // Journal of Statistical Computation and Simulation, 1978, 7:3-4, 167-180
 * DOI: http://dx.doi.org/10.1080/00949657808810225
 */
double tdistr_cdf(double x, int f)
{
	// Check input arguments
	if (f <= 0) { // Invalid value
		return -1;
	}
	// Intermediate variables
	double t = atan(x / sqrt(f)); // theta
	double cos_t = cos(t); // cos(theta)
	double F = 0.5;
	// Calculate series
	if (f == 1) {
		F += t / M_PI;
	} else if (f % 2 == 1) { // Odd
		double cossum = cos_t, coeff = 1.0;
		for (int i = 3; i <= f - 2; i += 2) {
			coeff *= (double) (i - 1) / i;
			cossum += coeff*pow(cos_t, i);
		}
		F += (t + sin(t)*cossum) / M_PI;		
	} else if (f % 2 == 0) { // Even
		double cossum = 1.0, coeff = 1.0;
		for (int i = 2; i <= f - 2; i += 2) {
			coeff *= (double) (i - 1) / i;
			cossum += coeff*pow(cos_t, i);
		}
		F += sin(t) / 2 * cossum;
	}
	// Return the function value
	return F;
}

static double tdistr_cdf_env(double x, void *fdata)
{
	double p = ((double *) fdata)[0];
	double f = ((double *) fdata)[1];
	return tdistr_cdf(x, f) - p;
}

/**
 * \brief Inverse t-distribution function
 * \param p Probability (0 - 1)
 * \param f Degrees of freedom
 *
 * Uses bisection method and the next functions:
 * - tdistr_cdf function for f < 1000
 * - erf_inv function and asymptotic decomposition from Abramowitz et al (26.7.5) for large f >= 1000 
 */
double tdistr_inv(double p, double f)
{
	// Input arguments control
	if ( (p < 0) | (p > 1) ) {
		return -1;
	}
	if (p < 1e-12) {
		return -INFINITY;
	}
	if (p > 1 - 1e-12) {
		return INFINITY;
	}
	if (f < 1) {
		return NAN;
	}

	if (f < 1000) { // Variant 1: series decomposition
		double fdata[2] = {p, f};	
		return bisection_method(-50.0, 50.0, tdistr_cdf_env, (void *) fdata);
	} else { // Variant 2: asymptotic decomposition from Abramowitz (26.7.5)
		double x = sqrt(2)*erf_inv(2*p - 1);
		double t = x;
		double xsqr = x*x;
		t += x*(xsqr + 1) / 4 / f; // g1(x)/v
		t += x*(xsqr*(5*xsqr + 16) + 3) / 96 / (f*f); // g2(x)/v^2
		t += x*(xsqr*(xsqr*(3*xsqr + 19) + 17) - 15) / 384 / pow(f, 3); // g3(x)/v^3
		t += x*(xsqr*(xsqr*(xsqr*(79*xsqr + 776) + 1482) - 1920) - 945) / 92160 / pow(f, 4); // g4(x)/v^4

		return t;
	}
}

/**
 * \brief exp(-x)*Ei(x) function
 */
double eminusx_ei(double x)
{
	double val = 1, fact = 1, d = 1;
	if (x <= -2) { // Continued fraction
		for (int i = 100; i >= 1; i--) {
			val = (-x) + i / (1 + i / val);
		}
		val = -1.0 / val;
	} else if (fabs(x) < fabs(log(DBL_EPSILON))) { // Taylor series
		val = exp(-x) * (EULER_GAMMA + log(fabs(x)));
		for (int i = 1; i < 500 && fabs(d)/fabs(val) > DBL_EPSILON; i++) {
			fact *= i;
			d = exp(-x) * pow(x, i) / (i * fact);
			val += d;
		}
	} else { // Asymptotic formula
		double d_prev = 1e100;
		for (int i = 1; i < 500 && fabs(d)/fabs(val) > DBL_EPSILON && fabs(d) < fabs(d_prev); i++) {
			fact *= i;
			d_prev = d;
			d = fact / pow(x, i);
			val += d;
		}
		val /= x;
	}
	return val;
}

