/**
 * \file 
 * \brief Some subroutines for implementation of thermodynamic models
 * and statistical tests in the datafit.cpp module.
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#ifndef __TDMODEL_H
#define __TDMODEL_H
#include <string>
#include <vector>
#include <map>
#include <functional>

/** CODATA 2018 molar gas constant */
#define CONST_R 8.314462618

/** Thermochemical calorie (in J) */
#define CONST_CAL 4.184

/** Mathematical constants */
#define EULER_GAMMA 0.577215664901532860606512
#define MAD_TO_STD 1.482602218505602

class TDPEFunc;

/**
 * \brief Contains parameters for the baseline Cp(T) and H(T)-H(0) model;
 * includes Einstein-Planck and polynomial part. Supports serialization
 * and deserialization.
 */
class TDBaseLineFuncParams {
public:
	TDBaseLineFuncParams();
	TDBaseLineFuncParams(size_t nAlphaTheta, size_t nPolyCoeffs);
	bool SetAlphaTheta(const std::vector <double> &alpha, const std::vector <double> &theta);
	bool SetPolyCoeffs(const std::vector <double> &coeff, const std::vector <double> &degree);
	bool IsAllAlphaThetaPositive();
	bool IsAllParamsSignificant(const TDBaseLineFuncParams &dparam);
	void AddTerm(double alpha, double theta);

	TDBaseLineFuncParams GetLowerBoundaries();
	TDBaseLineFuncParams GetUpperBoundaries();
	void SortParameters();

	std::vector <double> Serialize();
	bool Deserialize(const std::vector <double> &param);

	size_t NumParams();
	size_t NumAlphaThetaTerms();
	std::vector <double> GetKsi();
	void SetZeros();

	std::vector <double> SerializeFixNatoms();
	bool DeserializeFixNatoms(const std::vector <double> &ksiparam);


	
	std::vector <double> alpha;
	std::vector <double> theta;
	std::vector <double> polyCoeff;
	std::vector <double> polyDegree;
	double Natoms;

	bool isNatomsFixed;

private:
	std::vector <double> KsiToAlpha(const std::vector <double> &ksi);
	std::vector <double> AlphaToKsi(const std::vector <double> &alpha);
};

double sign(double x);
double bisection_method(double a, double b, double (*func)(double, void*), void *fdata);
double erf_inv(double x);
double tdistr_cdf(double x, int f);
double tdistr_inv(double p, double f);
double eminusx_ei(double x);

#endif
