/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <algorithm>
#include "datafit.h"
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/textfile.h"
#include "wx/filefn.h"
#include "tdtable.h"
#include "misc.h"


/**
 * \brief Dialog for requesting T range from the user.
 * This range also contains temperature step.
 *
 * Applied for addition of values into tables with tabulated
 * thermodynamic values.
 */
class TRangeDialog : public wxDialog
{
public:
	TRangeDialog(wxWindow *parent);

	double GetTMin();
	double GetTMax();
	double GetTStep();
private:
	void OnOk(wxCommandEvent &event);
	void OnCancel(wxCommandEvent &event);

	wxTextCtrl *txtTmin;
	wxTextCtrl *txtTmax;
	wxTextCtrl *txtTstep;

	wxDECLARE_EVENT_TABLE();
};

enum {
	ID_BTN_OK = 101,
	ID_BTN_CANCEL
};

wxBEGIN_EVENT_TABLE(TRangeDialog, wxDialog)
	EVT_BUTTON(ID_BTN_OK, TRangeDialog::OnOk)
	EVT_BUTTON(ID_BTN_CANCEL, TRangeDialog::OnCancel)
wxEND_EVENT_TABLE()


/**
 * \brief Initializes all controls of the temperature range dialog
 */
TRangeDialog::TRangeDialog(wxWindow *parent) : wxDialog(parent, wxID_ANY, wxString("T range addition"), wxPoint(10, 10), wxSize(250, 200))
{
	wxPanel *panel = new wxPanel(this, wxID_ANY);

	wxStaticText *lblTmin = new wxStaticText(panel, wxID_ANY, "Tmin, K");
	wxStaticText *lblTmax = new wxStaticText(panel, wxID_ANY, "Tmax, K");
	wxStaticText *lblTstep = new wxStaticText(panel, wxID_ANY, "Tstep, K");

	txtTmin = new wxTextCtrl(panel, wxID_ANY, "100");
	txtTmax = new wxTextCtrl(panel, wxID_ANY, "1000");
	txtTstep = new wxTextCtrl(panel, wxID_ANY, "250");

	wxButton *btnOk = new wxButton(panel, ID_BTN_OK, "Ok");
	wxButton *btnCancel = new wxButton(panel, ID_BTN_CANCEL, "Cancel");
	
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);

	// Sizers
	int borderWidth = 5;
	wxFlexGridSizer *gridSizer = new wxFlexGridSizer(3, 2, 5, 5);
	gridSizer->Add(lblTmin, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(txtTmin, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(lblTmax, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(txtTmax, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(lblTstep, 0, wxALL | wxEXPAND, borderWidth);
	gridSizer->Add(txtTstep, 0, wxALL | wxEXPAND, borderWidth);

	wxBoxSizer *okCancelSizer = new wxBoxSizer(wxHORIZONTAL);
	okCancelSizer->AddStretchSpacer(1);
	okCancelSizer->Add(btnOk, 0, wxALL | wxEXPAND, borderWidth);
	okCancelSizer->Add(btnCancel, 0, wxALL | wxEXPAND, borderWidth);

	wxBoxSizer *mainSizer = new wxBoxSizer(wxVERTICAL);
	mainSizer->Add(gridSizer, 0, wxALL | wxEXPAND, borderWidth);
	mainSizer->Add(okCancelSizer, 1, wxALL | wxEXPAND, borderWidth);	
	panel->SetSizer(mainSizer);
	panel->Layout();
}

/**
 * \brief Ok button click handler (checks user data and closes the modal dialog
 * with wxID_OK status)
 */
void TRangeDialog::OnOk(wxCommandEvent &event)
{
	double Tmin = GetTMin();
	double Tmax = GetTMax();
	double Tstep = GetTStep();
	// Check inputs
	if (Tmin <= 0) {
		wxMessageBox("Invalid Tmin value", "", wxICON_ERROR);
		return;
	}
	if (Tmax <= 0) {
		wxMessageBox("Invalid Tmax value", "", wxICON_ERROR);
		return;
	}
	if (Tstep <= 0) {
		wxMessageBox("Invalid Tstep value", "", wxICON_ERROR);
		return;
	}
	if (Tmin >= Tmax) {
		wxMessageBox("Tmax must be greater than Tmin", "", wxICON_ERROR);
		return;
	}
	if ((Tmax - Tmin) / Tstep > 1000) {
		wxMessageBox("Tstep is too small");
		return;
	}
	// If input is ok -- end it as OK
	EndModal(wxID_OK);
}

/**
 * \brief Ok button click handler (closes the modal dialog with wxID_CANCE status)
 */
void TRangeDialog::OnCancel(wxCommandEvent &event)
{
	EndModal(wxID_CANCEL);
}


/**
 * \brief Returns Tmin (minimal temperature, K) entered by a user
 */
double TRangeDialog::GetTMin()
{
	double Tmin;
	return (txtTmin->GetValue().ToDouble(&Tmin)) ? Tmin : -1;
}

/**
 * \brief Returns Tmax (maximal temperature, K) entered by a user
 */
double TRangeDialog::GetTMax()
{
	double Tmax;
	return (txtTmax->GetValue().ToDouble(&Tmax)) ? Tmax : -1;
}

/**
 * \brief Returns Tstep (temperature step, K) entered by a user
 */
double TRangeDialog::GetTStep()
{
	double Tstep;
	return (txtTstep->GetValue().ToDouble(&Tstep)) ? Tstep : -1;
}


//////////////////////////////////////////////
enum {
	ID_GR_DATA = 501,
	ID_MNU_TBLRESET,
	ID_MNU_TBLADDRNG,
	ID_MNU_SAVEAS,

	ID_MNU_ERRMODE,
	ID_MNU_ERRSTD,
	ID_MNU_ERRCONFINT
};


wxBEGIN_EVENT_TABLE(TdTableWindow, wxFrame)
	EVT_SIZE(TdTableWindow::OnSize)
	EVT_CLOSE(TdTableWindow::OnClose)
	EVT_MENU(ID_MNU_TBLRESET, TdTableWindow::OnResetTable)
	EVT_MENU(ID_MNU_TBLADDRNG, TdTableWindow::OnAddTableRange)
	EVT_MENU(ID_MNU_SAVEAS, TdTableWindow::OnSaveAs)
	EVT_MENU(ID_MNU_ERRSTD, TdTableWindow::OnSetErrMode)
	EVT_MENU(ID_MNU_ERRCONFINT, TdTableWindow::OnSetErrMode)
wxEND_EVENT_TABLE()

TdTableWindow::TdTableWindow(wxWindow *parent, LSQResult *res, const wxString &title)
	: wxFrame(parent, wxID_ANY, title, wxPoint(50, 50), wxSize(800, 600) )
{
	useConfIntervals = false;

	wxPanel *panel = new wxPanel(this, wxID_ANY);
	lsqRes = *res;

	grData = new wxGrid(panel, ID_GR_DATA, wxPoint(20, 20), wxSize(230,200));
	grData->CreateGrid(0, 7);
	RenderHeader();

	TVector.push_back(298.15);
	TVector.push_back(0.001);
	TVector.push_back(0.01);
	TVector.push_back(0.1);
	TVector.push_back(1);
	TVector.push_back(5);
	for (double Ti = 10; Ti < 51; Ti += 51) {
		TVector.push_back(Ti);
	}
	for (double Ti = 50; Ti < 1001; Ti += 50) {
		TVector.push_back(Ti);
	}
	std::sort(TVector.begin(), TVector.end());
	RenderTVector();

	// Window menu
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_MNU_SAVEAS, "Save as");
	
	wxMenu *menuEdit = new wxMenu();
	menuEdit->Append(ID_MNU_TBLRESET, "Reset the table");
	menuEdit->Append(ID_MNU_TBLADDRNG, "Add T range to the table");

	wxMenu *menuErrMode = new wxMenu();
	menuErrMode->Append(ID_MNU_ERRSTD, "Standard deviation", "Standard deviation", wxITEM_RADIO);
	menuErrMode->Append(ID_MNU_ERRCONFINT, "95% confidence interval", "95% confidence interval", wxITEM_RADIO);	
	menuErrMode->FindItem(ID_MNU_ERRSTD)->Check(true);
	
	menuEdit->AppendSubMenu(menuErrMode, "Errors estimation");

	wxMenuBar *menuBar = new wxMenuBar();
	menuBar->Append(menuFile, "&File");
	menuBar->Append(menuEdit, "&Edit");
	SetMenuBar(menuBar);
	// Window icon
	SetIcon(wxICON(sample));	
	// Set window coordinates
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetSize(wxSize(800, 650));
	SetMinSize(wxSize(800, 300));
}

void TdTableWindow::OnSize(wxSizeEvent &event)
{
        wxSize s = event.GetSize();
	s.SetWidth(s.GetWidth() - 50);
	s.SetHeight(s.GetHeight() - 80);
	grData->SetSize(s);
	grData->Move(10, 10);
	ResizeWxGridCols(grData);
	event.Skip();
}

void TdTableWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}

void TdTableWindow::RenderHeader()
{
	// Header texts
	grData->SetColLabelValue(0, wxT("T / K"));
	grData->SetColLabelValue(1, wxT("Cp\nJ/(mol\u00B7K)"));
	grData->SetColLabelValue(2, wxT("S-S(0)\nJ/(mol*K)"));
	grData->SetColLabelValue(3, wxT("H-H(0)\nJ/mol"));
	if (useConfIntervals) {
		grData->SetColLabelValue(4, wxT("\u0394(Cp)\nJ/(mol\u00B7K)")); // Delta
		grData->SetColLabelValue(5, wxT("\u0394(S-S(0))\nJ/(mol\u00B7K)")); // Delta
		grData->SetColLabelValue(6, wxT("\u0394(H-H(0))\nJ/mol")); // Delta
	} else {
		grData->SetColLabelValue(4, wxT("s(Cp)\nJ/(mol\u00B7K)"));
		grData->SetColLabelValue(5, wxT("s(S-S(0))\nJ/(mol\u00B7K)"));
		grData->SetColLabelValue(6, wxT("s(H-H(0))\nJ/mol"));
	}
	// Header sizes	
	grData->SetColSize(0, 60);
	grData->SetColSize(1, 100);
	grData->SetColSize(2, 100);
	grData->SetColSize(3, 100);
	grData->SetColSize(4, 100);
	grData->SetColSize(5, 100);
	grData->SetColSize(6, 100);
	grData->SetColLabelSize(50);
}


void TdTableWindow::RenderTVector()
{
	std::vector <double> &T = TVector;
	int numOfRows_old = grData->GetNumberRows();
	int numOfRows = T.size();
	if (numOfRows_old > 0) {
		grData->DeleteRows(0, numOfRows_old);
	}
	grData->InsertRows(0, numOfRows);

	std::vector <double> Cp = lsqRes.CpFunc(T);
	std::vector <double> S = lsqRes.SFunc(T);
	std::vector <double> H = lsqRes.HFunc(T);

	double tcoeff = useConfIntervals ? tdistr_inv(0.975, lsqRes.GetNumOfPoints() - lsqRes.GetNumOfParams()) : 1;	
	for (int i = 0; i < numOfRows; i++) {
		// a) Temperature
		grData->SetCellValue(i, 0, wxString::Format("%g", T[i]));
		// b) Values
		grData->SetCellValue(i, 1, wxString::Format("%g", Cp[i]));
		grData->SetCellValue(i, 2, wxString::Format("%g", S[i]));
		grData->SetCellValue(i, 3, wxString::Format("%g", H[i]));
		// c) Confidence intervals
		grData->SetCellValue(i, 4, wxString::Format("%g", tcoeff * lsqRes.FuncConfIntStd(T[i], DATATYPE_CP)));
		grData->SetCellValue(i, 5, wxString::Format("%g", tcoeff * lsqRes.FuncConfIntStd(T[i], DATATYPE_S)));
		grData->SetCellValue(i, 6, wxString::Format("%g", tcoeff * lsqRes.FuncConfIntStd(T[i], DATATYPE_H)));
		// Set all cells read-only
		for (int j = 0; j < 7; j++) {
			grData->SetReadOnly(i, j);
		}
	}
}

void TdTableWindow::OnResetTable(wxCommandEvent &event)
{
	TVector.clear();
	RenderTVector();
}

void TdTableWindow::OnAddTableRange(wxCommandEvent &event)
{
	TRangeDialog *tdlg = new TRangeDialog(this);
	if (tdlg->ShowModal() == wxID_OK) {
		double Tmax = tdlg->GetTMax();
		double Tstep = tdlg->GetTStep();
		for (double Ti = tdlg->GetTMin(); Ti <= Tmax + 0.5*Tstep; Ti += Tstep) {
			TVector.push_back(Ti);
		}
		std::sort(TVector.begin(), TVector.end());
		std::vector <double>::iterator it = std::unique(TVector.begin(), TVector.end());
		TVector.resize(std::distance(TVector.begin(), it));
		RenderTVector();
	}
	tdlg->Destroy();	
}

void TdTableWindow::OnSetErrMode(wxCommandEvent &event)
{
	int id = event.GetId();
	bool checked = event.IsChecked();
	if (id == ID_MNU_ERRSTD && checked) {
		useConfIntervals = false;
	} else if (id == ID_MNU_ERRCONFINT && checked) {
		useConfIntervals = true;
	} else {
		return; // Something unknown
	}
	RenderHeader();
	RenderTVector();	
}


/**
 * \brief  Export the table to .csv file
 */
void TdTableWindow::OnSaveAs(wxCommandEvent &event)
{
	// Header texts
	static const wxString strHeader[] = {"T;", "Cp;", "S;", "H-H0;",
		"s(Cp);", "s(S-S0);", "H(H-H0);",
		"D(Cp);", "D(S-S0);", "D(H-H0)"};

	wxFileDialog saveFileDialog(this, wxT("Save as .csv file"), "", "",
                       "csv files (*.csv)|*.csv", wxFD_SAVE);
	if (saveFileDialog.ShowModal() != wxID_CANCEL) {
		wxTextFile file;
		wxString fileName = saveFileDialog.GetPath();
		// Open file for writing (if it exists -- 
		// it will be erased)
		if (wxFileExists(fileName)) {
			wxString msg = wxString::Format("File %s already exists. Overwrite it?", fileName);
			if (wxMessageBox(msg, "CpFit", wxICON_EXCLAMATION | wxYES_NO | wxNO_DEFAULT) == wxNO) {
				return;
			}			
			if (!wxRemoveFile(fileName)) {
				wxMessageBox("Cannot save the file", "CpFit", wxICON_ERROR | wxOK);
				return;
			}
		}
		if (!file.Create(fileName)) {
			wxMessageBox("Cannot save the file", "CpFit", wxICON_ERROR | wxOK);
			return;
		}
		// Write caption
		int numOfCols = sizeof(strHeader) / sizeof(strHeader[0]);
		wxString caption;
		for (int i = 0; i < numOfCols; i++) {
			caption += strHeader[i];
		}
		file.AddLine(caption);
		// Write data
		std::vector <double> &T = TVector;
		std::vector <double> Cp = lsqRes.CpFunc(T);
		std::vector <double> S = lsqRes.SFunc(T);
		std::vector <double> H = lsqRes.HFunc(T);
		double tcoeff = tdistr_inv(0.975, T.size() - lsqRes.GetNumOfParams());
		for (size_t i = 0; i < T.size(); i++) {
			wxString dataLine;
			double sCp = lsqRes.FuncConfIntStd(T[i], DATATYPE_CP);
			double sS = lsqRes.FuncConfIntStd(T[i], DATATYPE_S);
			double sH = lsqRes.FuncConfIntStd(T[i], DATATYPE_H);
			// a) Temperature
			dataLine += wxString::Format("%g;", T[i]);
			// b) Values
			dataLine += wxString::Format("%.15g;%.15g;%.15g;", Cp[i], S[i], H[i]);
			// c) Confidence intervals: standard deviations
			dataLine += wxString::Format("%g;%g;%g;", sCp, sS, sH);
			// d) Confidence intervals: t-distribution intervals
			dataLine += wxString::Format("%g;%g;%g", tcoeff*sCp, tcoeff*sS, tcoeff*sH);
			// e) write to the file
			file.AddLine(dataLine);
		}
		// Save everything to the file and close it
		file.Write();
		file.Close();
	}
}

////////////////////////////////////////////////////
wxBEGIN_EVENT_TABLE(ResTableWindow, wxFrame)
	EVT_MENU(ID_MNU_SAVEAS, ResTableWindow::OnSaveAs)
	EVT_SIZE(ResTableWindow::OnSize)
	EVT_CLOSE(ResTableWindow::OnClose)
wxEND_EVENT_TABLE()


/**
 * \brief  Loads residuals into wxGrid table.
 */
ResTableWindow::ResTableWindow(wxWindow *parent, TDDataType dataType, LSQContext &lsqCont, LSQResult &lsqRes, const wxString &title)
	: wxFrame(parent, wxID_ANY, title, wxPoint(50, 50), wxSize(800, 600))
{
	// Save data sets
	this->dataType = dataType;
	this->lsqCont = lsqCont;
	this->lsqRes = lsqRes;
	// Create table 
	wxPanel *panel = new wxPanel(this, wxID_ANY);	
	grData = new wxGrid(panel, ID_GR_DATA, wxPoint(20, 20), wxSize(230,200));
	if (dataType != DATATYPE_H) {
		grData->CreateGrid(0, 5);
	} else {
		grData->CreateGrid(0, 6);
	}
	grData->SetColLabelSize(50);
	// Choose data type specific names of columns
	switch (dataType) {
	case DATATYPE_CP:
		grData->SetColLabelValue(0, wxT("T / K"));
		grData->SetColLabelValue(1, wxT("Cp(obsv)\nJ/(mol\u00B7K)")); // \cdot
		grData->SetColLabelValue(2, wxT("Cp(calc)\nJ/(mol\u00B7K)"));
		grData->SetColLabelValue(3, wxT("\u03B4Cp\nJ/(mol\u00B7K)")); // \delta and \cdot
		grData->SetColLabelValue(4, wxT("\u03B4Cp/Cp(obsv)\n%"));
		break;
	case DATATYPE_H:
		grData->SetColLabelValue(0, wxT("T1 / K"));
		grData->SetColLabelValue(1, wxT("T2 / K"));
		grData->SetColLabelValue(2, wxT("\u0394H(obsv)\nJ/mol")); // \Delta H(obsv)
		grData->SetColLabelValue(3, wxT("\u0394H(calc)\nJ/mol")); // \Delta H(calc)
		grData->SetColLabelValue(4, wxT("\u03B4(\u0394H)\nJ/mol")); // \delta (\Delta H)
		grData->SetColLabelValue(5, wxT("rel.err.%"));
		break;
	case DATATYPE_S:
		grData->SetColLabelValue(0, wxT("T / K"));
		grData->SetColLabelValue(1, wxT("(S-S(0))(obsv)\nJ/(mol\u00B7K)"));
		grData->SetColLabelValue(2, wxT("(S-S(0))(calc)\nJ/(mol\u00B7K)"));
		grData->SetColLabelValue(3, wxT("\u03B4(S-S(0))\nJ/(mol\u00B7K)"));
		grData->SetColLabelValue(4, wxT("rel.err.%"));
		break;
	case DATATYPE_UNKNOWN:
		grData->SetColLabelValue(0, wxT("T / K"));
		grData->SetColLabelValue(1, wxT("Unknown"));
		grData->SetColLabelValue(2, wxT("Unknown"));
		grData->SetColLabelValue(3, wxT("Unknown"));
		grData->SetColLabelValue(4, wxT("Unknown"));
	}
	// Add data to the table
	std::vector <double> T1, T2, ycalc;
	std::vector <double> T = lsqCont.GetXVector(dataType);
	std::vector <double> yobs = lsqCont.GetYVector(dataType);

	if (dataType != DATATYPE_H) {
		ycalc = lsqRes.YFunc(dataType, T);
	} else {
		T1 = lsqCont.GetX1Vector(dataType);
		T2 = lsqCont.GetX2Vector(dataType);
		ycalc = lsqRes.DHFunc(T1, T2);
	}

	grData->InsertRows(0, T.size());
	for (size_t i = 0; i < T.size(); i++) {
		double dy = yobs[i] - ycalc[i];
		double ey = fabs(dy / yobs[i]) * 100.0;
		int j;
		if (dataType != DATATYPE_H) {
			grData->SetCellValue(i, 0, wxString::Format("%g", T[i]));
			j = 1;
		} else {
			grData->SetCellValue(i, 0, wxString::Format("%g", T1[i]));
			grData->SetCellValue(i, 1, wxString::Format("%g", T2[i]));
			j = 2;
		}
		grData->SetCellValue(i, j++, wxString::Format("%g", yobs[i]));
		grData->SetCellValue(i, j++, wxString::Format("%g", ycalc[i]));
		grData->SetCellValue(i, j++, wxString::Format("%g", dy));
		grData->SetCellValue(i, j++, wxString::Format("%g", ey));
		
		
		for (int j = 0; j < grData->GetNumberCols(); j++) {
			grData->SetReadOnly(i, j);			
		}
	}
	// Window icon
	SetIcon(wxICON(sample));
	// Window menu
	wxMenuBar *menuBar = new wxMenuBar();	
	wxMenu *menuFile = new wxMenu();
	menuFile->Append(ID_MNU_SAVEAS, "Save as");
	menuBar->Append(menuFile, "&File");
	SetMenuBar(menuBar);
	// Set window coordinates
	wxPoint wndPos = parent->GetPosition();
	wndPos.x += 20;
	wndPos.y += 20;
	SetPosition(wndPos);
	SetSize(wxSize(800, 600));
	SetMinSize(wxSize(800, 300));
}


/**
 * \brief  Resizes table and its columns
 */
void ResTableWindow::OnSize(wxSizeEvent &event)
{
        wxSize s = event.GetSize();
	s.SetWidth(s.GetWidth() - 50);
	s.SetHeight(s.GetHeight() - 80);
	grData->SetSize(s);
	grData->Move(10, 10);
	ResizeWxGridCols(grData);
	event.Skip();
}

/**
 * \brief  Exports current residuals table to .csv file
 * 
 * Dot is used as decimal separator, ; as separator between fields.
 * Both heat capacities and enthalpies are supported
 */
void ResTableWindow::OnSaveAs(wxCommandEvent &event)
{
	wxFileDialog saveFileDialog(this, wxT("Save as .csv file"), "", "",
                       "csv files (*.csv)|*.csv", wxFD_SAVE);
	if (saveFileDialog.ShowModal() != wxID_CANCEL) {
		wxTextFile file;
		wxString fileName = saveFileDialog.GetPath();
		// Open file for writing (if it exists -- 
		// it will be erased)
		if (wxFileExists(fileName)) {
			// TODO: WARNING FOR USER!!!
			if (!wxRemoveFile(fileName)) {
				return; // TODO: ERROR
			}
		}
		if (!file.Create(fileName)) {
			return; // TODO:ERROR
		}
		// Write caption
		wxString hdr;
		int numCols = grData->GetNumberCols();
		for (int i = 0; i < numCols; i++) {
			hdr += grData->GetColLabelValue(i);
			if (i != numCols - 1) {
				hdr += ";";
			}
		}
		file.AddLine(hdr);
		// Write data
		for (int i = 0; i < grData->GetNumberRows(); i++) {
			wxString line;
			for (int j = 0; j < numCols; j++) {
				line += grData->GetCellValue(i, j);
				if (j != numCols - 1) {
					line += ";";
				}
			}
			file.AddLine(line);
		}
		file.Write();
		file.Close();
	}
	event.Skip();
}

void ResTableWindow::OnClose(wxCloseEvent &event)
{
	event.Skip();
}
