/**
 * \file
 * \brief Tests for TDPEFunc class
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#include <math.h>
#include <vector>
#include <algorithm>
#include <functional>
#include <string>
#include <stdio.h>
#include "datafit.h"


static TDBaseLineFuncParams makeParams()
{
	std::vector <double> param_vec = {0.9778,49.185, 12.36,164.19, 20.96,404.9, 30.3,1095};
	TDBaseLineFuncParams param(4, 0);
	param.Deserialize(param_vec);
	return param;
}

/**
 * \brief Shows a table of tabulated thermodynamic values using optimized parameters
 *
 * This test reproduces the table of tabulated thermodynamic values from
 * http://dx.doi.org/10.1007/s00269-013-0625-1. Note that currently its validity
 * must be checked manually by a user!
 */
static void test1()
{
	// Show the legend about the test
	printf("TEST: REPRODUCING THE TABLE OF TABULATED VALUES\n");
	printf("SOURCE OF PARAMETERS AND DATA: http://dx.doi.org/10.1007/s00269-013-0625-1\n");
	printf("  S.V. Kurdakova, R.O. Grishchenko, A.I. Druzhinina, L.P.Ogorodova\n");
	printf("  Thermodynamic properties of synthetic calcium-free carbonate cancrinite //\n");
	printf("  // Physics and Chemistry of Minerals. 2014. 41(1). P. 75-83\n");
	// Create object using existing parameters
	TDPEFunc tdf(makeParams());
	// Create T vector
	std::vector <double> T;	
	for (double Ti = 5.0; Ti < 81.0; Ti += 5.0) {
		T.push_back(Ti);
	}
	for (double Ti = 90; Ti <= 301; Ti += 10.0) {
		T.push_back(Ti);
	}
	T.push_back(273.15);
	T.push_back(298.15);
	std::sort(T.begin(), T.end());
	// Calculate Cp, S and H vectors for the table	
	std::vector <double> Cp = tdf.CpFunc(T);
	std::vector <double> S = tdf.SFunc(T);
	std::vector <double> H = tdf.HFunc(T);
	
	// Show the table	
	printf("%5s %10s %10s %10s\n", "T", "Cp", "S", "H");
	for (size_t i = 0; i < T.size(); i++) {
		printf("%6.2f %10.2f %10.2f %10.4f\n", T[i], Cp[i], S[i], H[i]/1000.0);
	}
	printf("\n\n");
	// Final print
	printf("\n\n");
}

/**
 * \brief Tests Cp(T), S(T)-S(0) and H(T)-H(0) derivatives
 *
 * This tests checks validity of Cp(T), S(T)-S(0) and H(T)-H(0) derivatives
 * using numerical methods. Note that currently its validity must be checked
 * manually by a user!
 */
static void test2()
{
	printf("TEST: TESTING CP, S and H DERIVATIVES\n");
	// Create object using existing parameters
	TDBaseLineFuncParams param = makeParams();
	TDBaseLineFuncParams param_tmp = param;
	TDPEFunc tdf(param);
	// Allocate memory for derivatives calculations	
	int nparams = tdf.GetNumOfParams();
	std::vector <double> param_tmp_vec(nparams);
	std::vector <double> JxCp(nparams);
	std::vector <double> JxS(nparams);
	std::vector <double> JxH(nparams);
	std::vector <double> JxCpNum(nparams);
	std::vector <double> JxSNum(nparams);
	std::vector <double> JxHNum(nparams);
	double T = 298.15;
	double dB = 1e-7;
	// Calculate derivatives using built-in functions
	tdf.dCpdBFunc(T, &JxCp[0]);
	tdf.dSdBFunc(T, &JxS[0]);
	tdf.dHdBFunc(T, &JxH[0]);
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxCp[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxS[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxH[i]);
	}
	printf("\n\n");


	// Calculate derivatives using numerical methods
	for (int i = 0; i < 8; i++) {
		param_tmp_vec = param.Serialize();
		param_tmp_vec[i] += dB;
		param_tmp.Deserialize(param_tmp_vec);
		TDPEFunc tdf2(param_tmp);
		double Cp2 = tdf2.CpFunc(T);
		double S2  = tdf2.SFunc(T);
		double H2  = tdf2.HFunc(T);
	
		param_tmp_vec = param.Serialize();
		param_tmp_vec[i] -= dB;
		param_tmp.Deserialize(param_tmp_vec);
		TDPEFunc tdf1(param_tmp);
		double Cp1 = tdf1.CpFunc(T);
		double S1  = tdf1.SFunc(T);
		double H1  = tdf1.HFunc(T);

		JxCpNum[i] = (Cp2 - Cp1) / (2*dB);
		JxSNum[i] = (S2 - S1) / (2*dB);
		JxHNum[i] = (H2 - H1) / (2*dB);
	}
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxCpNum[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxSNum[i]);
	}
	printf("\n");
	for (int i = 0; i < 8; i++) {
		printf("%15g ", JxHNum[i]);
	}
	printf("\n\n");
	// Final print
	printf("\n\n");		
}

/**
 * \brief Tests t-distribution implementation
 */
static void test3()
{
	printf("TEST: T-DISTRIBUTION\n");
	double f[] = {1, 2, 3, 5, 10, 100, 101, 1000, 10000, 1e6, 1e6+1, -1, NAN};
	double t[] = {12.706204736174696, 4.302652729749463, 3.182446305283710,
		      2.570581835636315,  2.228138851986273, 1.983971518523539,
		      1.983731002955605,  1.962339080826408, 1.960201239890626,
		      1.959966356814107,  1.959966356811734, NAN};
	double p = 0.975;
	printf("p = %g\n", p);
	printf("%10s %10s %10s %10s\n", "f", "tinv(calc)", "tinv(test)", "Dtinv");
	for (int i = 0; f[i] == f[i]; i++) {
		double tinv = tdistr_inv(p, f[i]);
		printf("%10.0f %10.5f %10.5f %10.2e\n", f[i], tinv, t[i], tinv - t[i]);
	}
	// Final print
	printf("\n\n");	
}

/**
 * \brief Runs two tests
 */
int main()
{
	test1();
	test2();
	test3();
	
	return 0;
}
