/**
 * \file
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */
#include <stdio.h>
#include <stdlib.h>
#include "wx/wx.h"
#include "wx/grid.h"
#include "wx/textfile.h"
#include "wx/tokenzr.h"
#include "wx/aboutdlg.h"
#include "wx/filefn.h"
#include "datafit.h"
#include "xydatafile.h"


int XYDataFile::MessageBoxStub(const wxString & message, const wxString & caption, int style)
{
#ifndef XYDATAFILE_CONSOLE
	return wxMessageBox(message, caption, style);
#else
	std::string msg = message.ToStdString();
	std::cout << msg << std::endl;
	return 0;
#endif
}


XYDataFile::XYDataFile()
{
	writeToStdio = false;
}

/**
 * \brief Returns current file name
 */
wxString XYDataFile::GetFileName()
{
	return fileName;
}


/**
 * \brief  Returns file info (3 strings of metadata) as wxString vector
 */
std::vector <wxString> XYDataFile::GetFileInfo()
{
	return fileInfo;
}

/**
 * \brief  Sets file info (3 strings of metadata) using wxString vector
 * given by a user
 */
bool XYDataFile::SetFileInfo(std::vector <wxString> fileInfo)
{
	if (fileInfo.size() < 3) {
		return false;
	} else {
		this->fileInfo = fileInfo;
		return true;
	}
}

/**
 * \brief Parses line that contains data
 * \return 0 -- empty/comment, 1 -- success, 2 -- failure
 */
int XYDataFile::ParsePoint(wxString &str, Point &point)
{
	str = str.Trim(false).Trim(true);
	str.Replace(",", "."); // Provides support , as decimal delimiter
	str.Replace(";", " "); // Provides support ; as delimiter of data
	// Skip empty/commented lines
	if (str.length() <= 2 || str[0] == '#') {
		return 0;
	}
	// Parse not empty line
	int fieldId = 0;
	bool isLineValid = true;
	if (str[0] == 'A' && str[1] == 'L') {
		// Load alpha values
		str = str.Mid(2).Trim();
		wxStringTokenizer tokenizer(str, " \t");
		while ( tokenizer.HasMoreTokens() ) {
			wxString token = tokenizer.GetNextToken();
			double val;
			if (!token.ToDouble(&val)) {
				isLineValid = false;
			}
			params_0.alpha.push_back(val);
		}
		return 0; // This line doesn't give any point
	} else if (str[0] == 'T' && str[1] == 'H') {
		// Load theta values
		str = str.Mid(2).Trim();
		wxStringTokenizer tokenizer(str, " \t");
		while ( tokenizer.HasMoreTokens() ) {
			wxString token = tokenizer.GetNextToken();
			double val;
			if (!token.ToDouble(&val)) {
				isLineValid = false;
			}
			params_0.theta.push_back(val);
		}
		return 0; // This line doesn't give any point
	} else if (str[0] == 'P' && str[1] == 'O') {
		// Load polynomial values
		str = str.Mid(2).Trim();
		wxStringTokenizer tokenizer(str, " \t");
		while ( tokenizer.HasMoreTokens() ) {
			double coeffVal, degreeVal;
			wxString tokenCoeff = tokenizer.GetNextToken();
			if (!tokenCoeff.ToDouble(&coeffVal)) { // Error: not numeric value
				isLineValid = false;
			}
			if (!tokenizer.HasMoreTokens()) { // Error: coeff/degree must be in pairs
				isLineValid = false;
			}

			wxString tokenDegree = tokenizer.GetNextToken();
			if (!tokenDegree.ToDouble(&degreeVal)) { // Error: not numeric value
				isLineValid = false;
			}
			params_0.polyCoeff.push_back(coeffVal);
			params_0.polyDegree.push_back(degreeVal);
		}
		return 0; // This line doesn't give any point
	} else if (str[0] == 'H') {
		point.dataType = DATATYPE_H;
		str = str.Mid(1).Trim();
		wxStringTokenizer tokenizer(str, " \t");
		while ( tokenizer.HasMoreTokens() )
		{
			wxString token = tokenizer.GetNextToken();
			double val, coeff = 1.0;
			long int ival;
			switch (fieldId) {
			case 0: // T1
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.x1 = val;
				break;
			case 1: // T2
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.x2 = val;
				break;
			case 2: // H(T2)-T(T1)
				if (token[token.size() - 1] == 'c') {
					token = token.Left(token.size() - 1);
					coeff = 4.184;
				}
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.y = val * coeff;
				break;
			case 3: // isRelDev
				if (!token.ToLong(&ival)) {
					isLineValid = false;
				}
				point.isRelDev = (ival) ? true : false;
				break;
			case 4: // w (weight)
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.w = val;
				break;
			case 5: // Data series ID
				point.setSeriesID(token);
				break;
			}
			fieldId++;
		}
		if (fieldId < 3) {
			isLineValid = false;
		}
	} else if (str[0] == 'X') {
		// Load extra term (X T1 T2 type fixparams | lsqparams)
		// Variables initialization
		// a) Intermediate variables
		str = str.Mid(1).Trim();
		wxStringTokenizer tokenizer(str, " \t");
		bool isFixParam = true;
		// b) Extra term data
		std::string termType;
		double Tmin = 0.0, Tmax = 0.0;
		std::vector <double> fixparam, lsqparam;
		// Parser loop
		while ( tokenizer.HasMoreTokens() )
		{
			wxString token = tokenizer.GetNextToken();
			double val;

			if (fieldId == 0) { // T1
				if (!token.ToDouble(&val)) {
					return 2; // Failure
				}
				Tmin = val;
			} else if (fieldId == 1) { // T2
				if (!token.ToDouble(&val)) {
					return 2; // Failure
				}
				Tmax = val;
			} else if (fieldId == 2) { // Type
				termType = std::string(token);
			} else if (fieldId > 2) { // Model parameters
				if (token == "|") {
					isFixParam = false;
				} else {
					if (!token.ToDouble(&val)) {
						return 2; // Failure
					}
					if (isFixParam) {
						fixparam.push_back(val);
					} else {
						lsqparam.push_back(val);
					}
				}
			}
			// TODO: IMPROVED ERROR CHECKING!!!
			fieldId++;
		}
		// Initialize term		
		ExtraTermContainer c(termType);
		if (!c.IsValid()) {
			return 2; // Unknown term type
		}
		c.term->Tmin = Tmin;
		c.term->Tmax = Tmax;		
		// Add parsed term (if it is valid)
		if (!c.term->SetFixParam(fixparam)) {
			return 2; // Failure
		}
		if (!c.term->SetLSQParam(lsqparam)) {
			return 2; // Failure
		}
		extraTerms.push_back(c);
		return 0; // This line doesn't give any point
	} else if (str[0] == 'S') {
		point.dataType = DATATYPE_S;
		str = str.Mid(1).Trim();
		wxStringTokenizer tokenizer(str, " \t");
		while ( tokenizer.HasMoreTokens() )
		{
			wxString token = tokenizer.GetNextToken();
			double val, coeff = 1.0;
			long int ival;
			switch (fieldId) {
			case 0: // T
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.x = val;
				break;
			case 1: // Cp
				if (token[token.size() - 1] == 'c') {
					token = token.Left(token.size() - 1);
					coeff = 4.184;
				}
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.y = val * coeff;
				break;
			case 2: // isRelDev
				if (!token.ToLong(&ival)) {
					isLineValid = false;
				}
				point.isRelDev = (ival) ? true : false;
				break;
			case 3: // w (weight)
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.w = val;
				break;
			case 4: // Data series ID
				point.setSeriesID(token);
				break;
			}
			fieldId++;
		}
		if (fieldId < 2) {
			isLineValid = false;
		}
	} else {
		point.dataType = DATATYPE_CP;
		if (str[0] == 'C') {
			str = str.Mid(1).Trim();
		}
		wxStringTokenizer tokenizer(str, " \t");
		while ( tokenizer.HasMoreTokens() )
		{
			wxString token = tokenizer.GetNextToken();
			double val, coeff = 1.0;
			long int ival;
			switch (fieldId) {
			case 0: // T
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.x = val;
				break;
			case 1: // Cp
				if (token[token.size() - 1] == 'c') {
					token = token.Left(token.size() - 1);
					coeff = 4.184;
				}
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.y = val * coeff;
				break;
			case 2: // isRelDev
				if (!token.ToLong(&ival)) {
					isLineValid = false;
				}
				point.isRelDev = (ival) ? true : false;
				break;
			case 3: // w (weight)
				if (!token.ToDouble(&val)) {
					isLineValid = false;
				}
				point.w = val;
				break;
			case 4: // Data series ID
				point.setSeriesID(token);
				break;
			}
			fieldId++;
		}
		if (fieldId < 2) {
			isLineValid = false;
		}
	}
	// Check the result
	if (!isLineValid) {
		return 2; // Failure (line is corrupted)
	} else {
		return 1; // Success
	}
}


/**
 * \brief Loads data file from disk
 * \param fileName  Full name of the loaded file
 * \param fmt  Reserved for future use
 */
bool XYDataFile::LoadFile(wxString &fileName, int fmt)
{
	wxTextFile file;
	long int numOfPoints = 0, numOfTruncPoints = 0;
	
	if (!file.Open(fileName))
		return false;

	Reset(); // Clear preloaded data from the previous file
	fileInfo.clear();
	int lineId = 0;
	for (wxString str = file.GetFirstLine(); !file.Eof(); str = file.GetNextLine(), lineId++) {
		if (lineId <= 2) { // First three lines: header
			fileInfo.push_back(str);
		} else if (lineId == 3) { // Line 4: number of points
			if (!str.ToLong(&numOfPoints) || numOfPoints < 0) {
				wxString strErrMsg = wxString::Format("Line 4 (%s) must be a positive integer number", str);
				MessageBoxStub(strErrMsg, "CpFit", wxICON_ERROR);
				numOfPoints = 0;
				numOfTruncPoints = 0;
				file.Close();
				return false;
			}
		} else if (lineId == 4) { // Line 5: reserved non-negative integer
			if (!str.ToLong(&numOfTruncPoints) || numOfTruncPoints < 0) {
				wxString strErrMsg = wxString::Format("Line 5 (%s) must be a non-negative integer number", str);
				MessageBoxStub(strErrMsg, "CpFit", wxICON_ERROR);
				numOfPoints = 0;
				numOfTruncPoints = 0;
				file.Close();
				return false;
			}
		} else if (lineId > 4) { // Lines 6-N: actual Cp(T), H(T2)-H(T1) and S(T) data
			Point point;
			int ans = ParsePoint(str, point);
			if (ans > 0) {
				if (ans == 1) {
					data.push_back(point);
				} else {
					wxString strErrMsg = wxString::Format("Line %d (%s) is corrupted", lineId + 1, str);
					MessageBoxStub(strErrMsg, "CpFit", wxICON_ERROR);
					file.Close();
					return false;
				}
			}
		}
	}
	file.Close();
	// Check validity of alpha and theta vectors (size consistency)
	if (params_0.alpha.size() != params_0.theta.size()) {
		MessageBoxStub("ALPHA and THETA vectors sizes are not consisent. They will be ignored.", "CpFit", wxICON_EXCLAMATION);
		params_0.alpha.clear();
		params_0.theta.clear();
	}
	// Check validity of numeric fields in the header
	if (numOfTruncPoints != 0) {
		if (!writeToStdio) {
			MessageBoxStub("Non-zero value in Line 5 will be ignored", "CpFit", wxICON_EXCLAMATION);
		} else {
			printf("Non-zero value in Line 5 will be ignored\n");
		}
	}
	if ((long int) data.size() != numOfPoints) {
		if (!writeToStdio) {
			MessageBoxStub(
				wxString::Format("Value in Line 4 (%d) and number of points (%d) are not matching",
					(int) numOfPoints, (int) data.size()),
				"CpFit", wxICON_EXCLAMATION);
		} else {
			printf("Value in Line 4 and number of points are not matching\n");
		}
	}
	// Save file name and return true (load of file was successful)
	this->fileName = fileName;
	return true;
}

/**
 * \brief Save data to the disk
 * \param fileName  Full name of the loaded file
 * \param fmt  Reserved for future use
 */
bool XYDataFile::SaveFile(wxString &fileName, int fmt)
{
	wxTextFile file;
	// Open file for writing (if it exists --
	// it will be erases)
	if (wxFileExists(fileName)) {
		if (!wxRemoveFile(fileName)) {
			return false;
		}
	}
	if (!file.Create(fileName)) {
		return false;
	}
	// Save metadata
	if (fileInfo.size() >= 3) {
		file.AddLine(fileInfo[0]);
		file.AddLine(fileInfo[1]);
		file.AddLine(fileInfo[2]);
	} else {
		MessageBoxStub("Internal error: fileInfo is empty. Empty file metadata will be written", "CpFit", wxICON_ERROR);
		file.AddLine("");
		file.AddLine("");
		file.AddLine("");
	}
	// Save data about the size
	file.AddLine(wxString::Format("%d", (int) data.size()));
	file.AddLine(wxString::Format("%d", 0));
	// Check if additional columns (with w and bool are required)
	bool addExtraRows = false;
	for (unsigned int i = 0; i < data.size(); i++) {
		if (data[i].isRelDev || data[i].w != 1.0 || data[i].dataType != DATATYPE_CP || data[i].getSeriesID().size() > 0) {
			addExtraRows = true;
			break;
		}
	}
	// Save extra terms
	for (unsigned int i = 0; i < extraTerms.size(); i++) {
		ExtraTerm *t = extraTerms[i].term;
		wxString line = wxString::Format("X %g %g %s ", t->Tmin, t->Tmax, t->GetType());
		std::vector <double> fixparam = t->GetFixParam();
		std::vector <double> lsqparam = t->GetLSQParam();
		for (unsigned int j = 0; j < fixparam.size(); j++) {
			line += wxString::Format("%g ", fixparam[j]);
		}
		line += "| ";
		for (unsigned int j = 0; j < lsqparam.size(); j++) {
			line += wxString::Format("%g ", lsqparam[j]);
		}
		file.AddLine(line);
	}
	// Save alpha and theta
	size_t nterms = params_0.alpha.size();
	if (nterms > 0) {
		wxString lineAlpha = "AL ";
		wxString lineTheta = "TH ";
		for (size_t i = 0; i < nterms; i++) {
			lineAlpha += wxString::Format("%.14g ", params_0.alpha[i]);
			lineTheta += wxString::Format("%.14g ", params_0.theta[i]);
		}
		file.AddLine(lineAlpha);
		file.AddLine(lineTheta);
	}
	// Save polynomial data
	size_t npolyterms = params_0.polyCoeff.size();
	if (npolyterms > 0) {
		wxString linePoly = "PO ";
		for (size_t i = 0; i < npolyterms; i++) {
			linePoly += wxString::Format("%.14g %.14g ",
				params_0.polyCoeff[i], params_0.polyDegree[i]);
		}
		file.AddLine(linePoly);
	}
	// Save raw data
	for (size_t i = 0; i < data.size(); i++) {
		if (addExtraRows) { // New format without w and reldev
			int reldev = (data[i].isRelDev) ? 1 : 0;
			if (data[i].dataType == DATATYPE_CP) {
				file.AddLine(wxString::Format("%.14g %.14g %d %.14g %s", 
					data[i].x, data[i].y, reldev, data[i].w, data[i].getSeriesID()));
			} else if (data[i].dataType == DATATYPE_H) {
				file.AddLine(wxString::Format("H %.14g %.14g %.14g %d %.14g %s", 
					data[i].x1, data[i].x2, data[i].y, reldev, data[i].w, data[i].getSeriesID()));
			} else if (data[i].dataType == DATATYPE_S) {
				file.AddLine(wxString::Format("S %.14g %.14g %d %.14g %s", 
					data[i].x, data[i].y, reldev, data[i].w, data[i].getSeriesID()));
			} else {
				file.AddLine("# UNKNOWN DATA TYPE");
			}
		} else { // Old format without w and reldev
			file.AddLine(wxString::Format("%.14g %.14g", data[i].x, data[i].y));
		}
	}
	// Save changes and close the file
	this->fileName = fileName;
	file.Write();
	file.Close();
	return true;
}


/**
 * \brief  Clears all data and file name but keeps files on disk intact
 */
void XYDataFile::Reset()
{
	data.clear();
	extraTerms.clear();
	params_0.alpha.clear();
	params_0.theta.clear();
	params_0.polyCoeff.clear();
	params_0.polyDegree.clear();
	if (fileInfo.size() >= 3) {
		fileInfo[0] = "";
		fileInfo[1] = "";
		fileInfo[2] = "";
	}
	fileName = "";
}
