/**
 * \file 
 * \brief Implementation of 2D plots
 *
 * Contains several classes for windows with Cp(T), Cp(T)/T, S(T)-S(0)
 * and H(T)-H(0) plots: calculated + experimental, absolute residuals,
 * relative residuals.
 *
 * \author Alexey L. Voskov (alvoskov@gmail.com)
 * \copyright (C) 2016-2020 Alexey L. Voskov
 *
 * Laboratory of Chemical Thermodynamics, Department of Chemistry,
 *
 * Lomonosov Moscow State University, Moscow, Russia, 119991
 *
 * This file is the part of the CpFit program
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/);
 */

#ifndef __XYPLOTS_H
#define __XYPLOTS_H

/**
 * \brief  Implementation of generic 2D plot for user data set
 */
class XYPlotWindow : public wxFrame
{
public:
	XYPlotWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes);
	virtual void InitPlot();
	void EnableLegend();
	void DisableLegend();

	std::vector <PlotMarker> seriesMarkers;
private:
	void InitSeriesMarkers();
	void OnSize(wxSizeEvent &event);
	void OnSaveAs(wxCommandEvent &event);
	void OnCopy(wxCommandEvent &event);
	void OnChkLegend(wxCommandEvent &event);
	void OnRbLegendPos(wxCommandEvent &event);
	void OnChkGrid(wxCommandEvent &event);
	void OnChkLogScaleX(wxCommandEvent &event);

	wxDECLARE_EVENT_TABLE();
protected:
	wxMenuBar *menuBar;
	wxMenuItem *mnuLegendPosTopLeft;
	wxMenuItem *mnuLegendPosTopRight;
	wxMenuItem *mnuLegendPosBottomLeft;
	wxMenuItem *mnuLegendPosBottomRight;
	LSQContext lsqCont; /*!< User-defined data */
	LSQResult lsqRes; /*!< User-defined data */
	PlotCanvas *pl;	/*!< Canvas with 2D plot */
};

/**
 * \brief Different X-Y coordinates types for YDataWindow and heat capacities.
 */
enum YDataWindowType
{
	YDATAWINDOW_NORMAL,
	YDATAWINDOW_LOWT_T_CP,
	YDATAWINDOW_LOWT_T2_CPT,
	YDATAWINDOW_LOWT_T2_CPT3
};

/**
 * \brief  Experimental + calculated data 2D plots (Cp(T), H(T)-H(0), S(T)-S(0));
 */
class YDataWindow : public XYPlotWindow
{
public:
	YDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes,
		TDDataType dataType, YDataWindowType wndType = YDATAWINDOW_NORMAL);
	virtual void InitPlot();
private:
	void InitAxesLabels();
	void InitXAxisLabel();
	void InitYAxisLabel();
	void InitCpAxisLabel();
	void DrawDataSeries();
	void DrawModel();
	void FindTLimits();
	void TransformXYCoordinates(double &x, double &y);

	TDDataType dataType;
	YDataWindowType wndType;

	double minT;
	double maxT;
};

/**
 * \brief  Experimental + calculated data for H(T) - H(T0)
 *
 * The key features of the plot:
 * - Normed (H(T) - H(T0))/(T - T0) and not normed (H(T) - H(T0)) plots are supported
 * - T0 interval can be set (as T0 +- DT0)
 */
class DHDataWindow : public XYPlotWindow
{
public:
	DHDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes, double T0 = 298.15, double DT0 = 0.5, bool isNormed = true);
	virtual void InitPlot();
private:
	bool isNormed;
	double T0;
	double DT0;
};

/**
 * \brief  Cp(T)/T 2D plots (experimental + calculated data). Also separate terms are shown.
 */
class CpTDataWindow : public XYPlotWindow
{
public:
	CpTDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes);
	virtual void InitPlot();
};

/**
 * \brief Residuals (absolute and relative) 2D plots
 */
class ResDataWindow : public XYPlotWindow
{
public:
	ResDataWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes, TDDataType dataType, bool isRel);
	virtual void InitPlot();
private:
	bool isRel;
	TDDataType dataType;
};

/**
 * \brief Window with phonon density spectrum estimation, i.e.
 * &alpha;-&theta; plot.
 */
class AlphaThetaSpectrumWindow : public XYPlotWindow
{
public:
	AlphaThetaSpectrumWindow(wxFrame *parent, const wxString &title, LSQContext &lsqCont, LSQResult &lsqRes);
	virtual void InitPlot();
};


#endif
